/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <sam9407.h>

#include "mixerctl.h"
#include "init.h"

#define NR_OF_ATOMS		256
#define ATOM_MASK		0xFFFF
#define ATOM_FIXED		0x10000
#define HASH_BUCKETS		211

typedef struct HashEntry {
  int atom;
  struct HashEntry *next;
} HashEntry;

typedef struct {
  int alloc, n;
  char **names;
  HashEntry *bucket[HASH_BUCKETS];
} Atoms;

static Atoms atoms;

/* mixer options that've been specified
 * in config file
 * if <0, no value was given
 */
static int mixerOptions[SAM_MIXER_ID_LAST];

/* command line options that can be specified in config
 * file as well
 */
static struct {
  char *name;
  char **destP;
  int atom;
} stringOptions[]={
  { "device",		&options.device },
  { "firmware", 	&options.firmware },
  { "soundbank",	&options.soundbanks[0] },
  { "soundbank0",	&options.soundbanks[0] },
  { "soundbank1",	&options.soundbanks[1] },
  { "soundbank2",	&options.soundbanks[2] },
  { "soundbank3",	&options.soundbanks[3] },
  { "soundbank4",	&options.soundbanks[4] },
  { "soundbank5",	&options.soundbanks[5] },
  { "soundbank6",	&options.soundbanks[6] },
  { "soundbank7",	&options.soundbanks[7] },
};

static int hash(char *str)
{
  char *p;
  unsigned h=0,g;
  for(p=str;*p;p++){
    h=(h<<4)+*p;
    if((g=h&0xf0000000)!=0){
      h^=g>>24;
      h^=g;
    }
  }
  return h%HASH_BUCKETS;
}

static int intern(char *name, int create)
{
  HashEntry *entry;
  int h=hash(name), alloc;
  for(entry=atoms.bucket[h]; entry; entry=entry->next) {
    if(strcmp(atoms.names[entry->atom], name)==0)
      return entry->atom;
  }
  if(!create)
    return -1;
  entry=(HashEntry *)xmalloc(sizeof(HashEntry));
  if(create&ATOM_FIXED)
    entry->atom=create&ATOM_MASK;
  else
    entry->atom=atoms.n++;
  entry->next=atoms.bucket[h];
  atoms.bucket[h]=entry;
  if(entry->atom>=atoms.alloc) {
    alloc=entry->atom*3/2;
    if(alloc<NR_OF_ATOMS)
      alloc=NR_OF_ATOMS;
    if(atoms.names)
      atoms.names=(char **)xrealloc(atoms.names, alloc*sizeof(char *));
    else
      atoms.names=(char **)xmalloc(alloc*sizeof(char *));
    memset(atoms.names+atoms.alloc, 0, (alloc-atoms.alloc)*sizeof(char *));
    atoms.alloc=alloc;
  }
  atoms.names[entry->atom]=xstrdup(name);
  return entry->atom;
}

void parseConfigFile(int cardNr)
{
  char buf[256], *cp, *name, *value;
  FILE *fp;
  int atomCard, atom, currentCard, i;

  memset(mixerOptions, ~0, sizeof(mixerOptions));

  if(!(fp=fopen(SAM_CONFIG_FILE, "r")))
    return;

  atoms.n=SAM_MIXER_ID_LAST;
  for(i=0; samMixerAssoc[i].name; i++)
    intern(samMixerAssoc[i].name, ATOM_FIXED|samMixerAssoc[i].id);

  for(i=0; i<NumberOf(stringOptions); i++)
    stringOptions[i].atom=intern(stringOptions[i].name, 1);

  atomCard=intern("card", 1);
  currentCard=0;

  while(fgets(buf, sizeof(buf), fp)) {
    cp=buf;
    while(isspace(*cp)) cp++;
    if(!*cp || *cp=='#')
      continue;
    name=cp;
    while(*cp && !isspace(*cp)) cp++;
    if(*cp)
      *cp++=0;
    if((atom=intern(name, 0))<0)
      continue;
    /* skip spaces at the beginning of value */
    while(isspace(*cp)) cp++;
    value=cp;
    /* find the end of value */
    while(*cp && !isspace(*cp)) cp++;
    *cp=0;
    
    if(atom==atomCard) {
      currentCard=strtol(value, NULL, 0);
      continue;
    }

    if(currentCard!=cardNr)
      continue;

    /* known keyword */
    if(atom>=SAM_MIXER_ID_LAST) {
      /* must be a string option */
      for(i=0; i<NumberOf(stringOptions); i++)
	if(stringOptions[i].atom==atom)
	  *stringOptions[i].destP=xstrdup(value);
    } else
      /* mixer adjustment */
      mixerOptions[atom]=strtol(value, NULL, 0);
  }

  fclose(fp);
}

void adjustMixer(void)
{
  int pass, n, i, j;
  SamMixerValue mixerValues[SAM_MIXER_ID_LAST+1], oldValues[SAM_MIXER_ID_LAST+1];

  for(pass=0; pass<2; pass++) {
    n=0;
    for(i=0; samMixerAssoc[i].name; i++) {
      if(pass==!samMixerAssoc[i].doesReset &&
	 mixerOptions[samMixerAssoc[i].id]>=0) {
	oldValues[n].id=samMixerAssoc[i].id;
	n++;
      }
    }

    if(n>0) {
      oldValues[n].id=SAM_MIXER_ID_SENTINEL;
      if(ioctl(mixerFd, SAM_IOC_MIXER_GET, oldValues)<0)
	fatal(1, "can't retrieve mixer values from: %s", options.device);
    }

    n=j=0;
    for(i=0; samMixerAssoc[i].name; i++) {
      if(pass==!samMixerAssoc[i].doesReset &&
	 mixerOptions[samMixerAssoc[i].id]>=0) {
	if(mixerOptions[samMixerAssoc[i].id]!=oldValues[j].value) {
	  mixerValues[n].id=samMixerAssoc[i].id;
	  mixerValues[n].value=mixerOptions[samMixerAssoc[i].id];
	  n++;
	}
	j++;
      }
    }

    if(n>0) {
      mixerValues[n].id=SAM_MIXER_ID_SENTINEL;
      
      if(ioctl(mixerFd, SAM_IOC_MIXER_SET, mixerValues)<0)
	fatal(1, "can't set mixer values on: %s", options.device);
    }
  }
}
