/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "init.h"

#ifndef MAP_FAILED
#define MAP_FAILED ((void *)-1)
#endif

char *prgName;
int mixerFd=-1;

Options options;

void fatal(int perr,char *format, ...)
{
  int errn=errno;
  va_list ap;
  va_start(ap, format);
  fprintf(stderr,"%s: ",prgName);
  vfprintf(stderr,format,ap);
  if(perr){
    fprintf(stderr,": ");
    errno=errn;
    perror(NULL);
  }else
    fprintf(stderr,"\n");
  va_end(ap);
  exit(1);
}

void *xmalloc(int size)
{
  void *result=malloc(size);
  if(!result)
    fatal(0, "virtual memory exhausted.");
  return result;
}

void *xrealloc(void *ptr, int size)
{
  void *result=realloc(ptr, size);
  if(!result)
    fatal(0, "virtual memory exhausted.");
  return result;
}

char *xstrdup(char *str)
{
  char *result=strdup(str);
  if(!result)
    fatal(0, "virtual memory exhausted.");
  return result;
}

static int openMixer(int cardNr)
{
  static char devName[256];
  struct stat st;

  if(options.device) {
    if(stat(options.device, &st)<0)
      return 0;
  } else {
    sprintf(devName, "/dev/sam%d_mixer", cardNr);
    if(stat(devName, &st)<0) {
      if(cardNr==0) {
	if(stat("/dev/mixer", &st)>=0)
	  options.device="/dev/mixer";
      }
    } else
      options.device=devName;
    if(!options.device)
      return 0;
  }

  if(!S_ISCHR(st.st_mode))
    return 0;

  return (mixerFd=open(options.device, O_RDWR))>=0;
}

static void loadFirmware(char *hardware)
{
  int fd, n, len;
  struct stat st;
  SamFirmware firmware;
  char path[256], *cp;

  /* replace %s elements in firmware path
   * by hardware identifier
   */
  len=strlen(hardware);
  cp=options.firmware;
  for(n=0; n<sizeof(path)-1 && *cp; cp++) {
    if(*cp=='%') {
      switch(*++cp) {
      case 's':
	if(n+len>=sizeof(path))
	  fatal(0, "firmware path too long: %s", options.firmware);
	memcpy(path+n, hardware, len);
	n+=len;
	break;
      case '%':
	path[n++]='%';
	break;
      default:
	fatal(0, "invalid firmware format: %s", options.firmware);
	break;
      }
    } else
      path[n++]=*cp;
  }
  if(*cp)
    fatal(0, "firmware path too long: %s", options.firmware);
  path[n]=0;

  if(options.verbose)
    fprintf(stderr, "%s: Initializing firmware from: %s\n", prgName, path);
  if((fd=open(path, O_RDONLY))<0)
    fatal(1, "can't open firmware: %s", path);
  if(fstat(fd, &st)<0)
    fatal(1, "can't stat firmware: %s", path);
  if((firmware.data=(unsigned short *)mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0))==MAP_FAILED)
    fatal(1, "can't mmap firmware: %s", path);
  firmware.size=st.st_size>>1;
  if(ioctl(mixerFd, SAM_IOC_FIRMWARE_LOAD, &firmware)<0)
    fatal(1, "can't install firmware: %s", path);
  munmap((void *)firmware.data, st.st_size);
  close(fd);
}

static void loadSoundBanks(void)
{
  int fd, bank, i;
  struct stat st;
  SamSoundBank soundBank;
  SamSoundBankInfo infos[SAM_NROF_SBANKS];
  char sbankName[SAM_SBANK_NAME_LEN+1], *cp;

  if(ioctl(mixerFd, SAM_IOC_BANK_INFO, infos)<0)
    fatal(1, "can't retrieve soundbank information");

  for(bank=0; bank<SAM_NROF_SBANKS; bank++) {
    if(!options.soundbanks[bank])
      continue;

    if((cp=strrchr(options.soundbanks[bank], '/')))
      cp++;
    else
      cp=options.soundbanks[bank];
    for(i=0; i<SAM_SBANK_NAME_LEN && *cp && *cp!='.'; i++, cp++)
      sbankName[i]=toupper(*cp);
    sbankName[i]=0;
    if(strcmp(sbankName, infos[bank].name)==0) {
      if(options.verbose)
	fprintf(stderr, "%s: soundbank %s already loaded at %d\n", prgName, sbankName, bank);
      continue;
    }

    if(*options.soundbanks[bank]) {
      if(options.verbose)
	fprintf(stderr, "%s: loading at #%d soundbank: %s\n", prgName, bank, options.soundbanks[bank]);

      if((fd=open(options.soundbanks[bank], O_RDONLY))<0)
	fatal(1, "can't open soundbank: %s", options.soundbanks[bank]);
      if(fstat(fd, &st)<0)
	fatal(1, "can't stat soundbank: %s", options.soundbanks[bank]);
      if((soundBank.data=(char *)mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0))==MAP_FAILED)
	fatal(1, "can't mmap soundbank: %s", options.soundbanks[bank]);
      soundBank.id=bank;
      soundBank.size=st.st_size;
      if(ioctl(mixerFd, SAM_IOC_BANK_LOAD, &soundBank)<0)
	fatal(1, "can't load soundbank: %s", options.soundbanks[bank]);
      munmap(soundBank.data, st.st_size);
      close(fd);
    } else {
      if(options.verbose)
	fprintf(stderr, "%s: unloading soundbank: %d\n", prgName, bank);
      if(ioctl(mixerFd, SAM_IOC_BANK_UNLOAD, &bank)<0)
	fatal(1, "can't unload soundbank: %d", bank);
    }
  }
}

typedef union
{
	unsigned short dbit;
	struct  bit_field
	{
	   unsigned short mute        : 1;
	   unsigned short wasel       : 1;
	   unsigned short midisel     : 1;
	   unsigned short daadsel     : 1;
	   unsigned short echo_vol    : 1;
	   unsigned short echo_on     : 1;
	   unsigned short dacsel      : 2;
	}  bit;
}DACBIT;

static void SetDaad( void )
{
  DACBIT db; 

  db.dbit = 0;

  db.bit.mute    = 1;
  db.bit.wasel   = 1;
  db.bit.midisel = 1;

  if(ioctl(mixerFd, SAM_IOC_SET_DAAD, db.dbit)<0)
      fatal(1, "can't query SAM9407 driver");
}

static void usage(void)
{
  fprintf(stderr,"usage: %s [-v] [-n cardNr] [-i] [-d <mixer_device>] [-f <firmware>] [-b <id>:<soundbank> ]\n", prgName);
  exit(1);
}

int main(int argc, char **argv)
{
  int cardNr, expectedCardNr, optionVerbose, c, bank;
  char *cp;
  SamDriverInfo driverInfo;

  prgName=argv[0];

  optionVerbose=0;
  for(expectedCardNr=cardNr=0;;cardNr++) {
    memset(&options, 0, sizeof(options));
    options.verbose=optionVerbose;

    parseConfigFile(cardNr);

    while(cardNr==expectedCardNr && (c=getopt(argc, argv, "b:d:f:in:v"))!=EOF) {
      switch(c) {
      case 'b':
	cp=NULL; /* make compiler happy */
	if((bank=strtol(optarg, NULL, 0))<0 || bank>=SAM_NROF_SBANKS ||
	   !(cp=strchr(optarg, ':')))
	  fatal(0, "invalid soundbank parameter: %s", optarg);
	options.soundbanks[bank]=cp+1;
	break;
      case 'd':
	options.device=optarg;
	break;
      case 'f':
	options.firmware=optarg;
	break;
      case 'i':
	options.initialize=1;
	break;
      case 'n':
	expectedCardNr=strtol(optarg, NULL, 0);
	break;
      case 'v':
	options.verbose=optionVerbose=1;
	break;
      default:
	usage();
      }
    }

    if(!openMixer(cardNr))
      break;

    if(ioctl(mixerFd, SAM_IOC_DRIVER_INFO, &driverInfo)<0)
      fatal(1, "can't query SAM9407 driver");
    
    if(options.verbose)
      fprintf(stderr, "%s: found hardware=%s, firmware=%s\n", prgName,
	      driverInfo.hardware,
	      *driverInfo.firmware ? driverInfo.firmware : "n/a");
    
    if(options.initialize || !*driverInfo.firmware) {
      if(!options.firmware)
		fatal(0, "No firmware file specified");
      
      loadFirmware(driverInfo.hardware);
    }

    SetDaad();
	
    adjustMixer();

    loadSoundBanks();

    close(mixerFd);
  }

  if(options.verbose)
    fprintf(stderr, "%s: %d cards handled\n", prgName, cardNr);

  return 0;
}
