/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/soundcard.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "driver.h"
#include "io.h"
#include "alloc.h"
#include "seq.h"
#include "ulaw.h"
#include "audio.h"

//#define _DEBUG
//#define SAM_BC_WRITE_DELIVERS_ALL


#define SAM_MIN_RATE		100
#define SAM_MAX_PLAYBACK_RATE	48000
#define SAM_MAX_RECORD_RATE	32755
#define SAM_DEFAULT_VOLUME	0xFFFF
#define SAM_AUDIO_BUF_OVERHEAD	16

#define SAM_FORMAT_16BIT(format)\
				((format)==AFMT_S16_LE ||\
				 (format)==AFMT_S16_BE ||\
				 (format)==AFMT_U16_LE ||\
				 (format)==AFMT_U16_BE)

typedef struct {
  struct wait_queue *feedSam;
  int shutdown;
} SamAudioInfo;

SamAudioFeedThreadInfo samFeedThreadInfo;

extern inline unsigned samCvtVolToEWS64(unsigned vol)
{
  unsigned result=(vol+1)/2;
  if(result>0x8000)
    result=0x8000;
  return result;
}

extern inline unsigned samCvtVolFromEWS64(unsigned vol)
{
  unsigned result=vol*2;
  if(result>0xFFFF)
    result=0xFFFF;
  return result;
}

static int adjustVolume(SamCard *card, int channel)
{
  SamChannelInfo *chp=card->channels+channel;

#ifdef _DEBUG
    printk( "Adjust Vol(%d) : %x, %x, %x, %x\n", 
	  channel,
          chp->volume.left,
          chp->volume.right,
          chp->volume.auxLeft,
          chp->volume.auxRight );
#endif

  return samTalk(card,
		 SamSendCtrl, 4, SAM_CMD_W_VOLLEFT, channel, chp->volume.left, chp->volume.left>>8,
		 SamSendCtrl, 4, SAM_CMD_W_VOLRIGHT, channel, chp->volume.right, chp->volume.right>>8,
		 SamSendCtrl, 4, SAM_CMD_W_VOLAUXLEFT, channel, chp->volume.auxLeft, chp->volume.auxLeft>>8,
		 SamSendCtrl, 4, SAM_CMD_W_VOLAUXRIGHT, channel, chp->volume.auxRight, chp->volume.auxRight>>8,
		 SamEndOfScript);
}

/* calculate words needed for 250 msec worth of sound */
extern inline unsigned samWordsNeeded(SamChannelInfo *chp)
{
  unsigned needed;
  needed=chp->rate/4;
//  needed=chp->rate/2;
  if(chp->stereo)
    needed<<=1;

#ifdef _DEBUG
//  printk( "WordNeed = %d %d %d\n", chp->rate, chp->alloced, needed );
#endif

  return needed;
}

static int allocBuffers(SamCard *card, int channel)
{
  SamChannelInfo *chp=card->channels+channel;
  unsigned needed;
  int error;

  needed=samWordsNeeded(chp);

  if(needed==chp->needed && chp->data && needed<=chp->alloced)
    /* nothing changed */
    return 0;

  chp->needed=needed;
  chp->count=0;

  if(chp->data && chp->needed<=chp->alloced)
    /* enough space already */
    return 0;

  if(chp->data) {
    /* release used resources */
    if((error=samFree(card, (channel<<8)|SAM_MMT_ID_STR_WAVE, ~SAM_MMT_ROM_MASK))<0)
      return error;
    kfree(chp->data);
    atomic_dec(&card->usageCount);
  }

  chp->alloced=chp->needed;
  chp->data=(short *)kmalloc(chp->alloced<<1, GFP_KERNEL);

  if(!chp->data) {
    chp->alloced=0;
    return -ENOMEM;
  }
  
  down(&card->usageCountLock);
  atomic_inc(&card->usageCount);
  up(&card->usageCountLock);

  /* allocated twice the size needed in SAM9407
   * plus some magic overhead
   */
  if((error=samAlloc(card, (channel<<8)|SAM_MMT_ID_STR_WAVE, (chp->needed<<1)+SAM_AUDIO_BUF_OVERHEAD, NULL, SAM_BOUNDARY_64K, NULL, NULL))<0) {
    kfree(chp->data);
    chp->data=NULL;
    chp->alloced=0;
    atomic_dec(&card->usageCount);
    return error;
  }

  return 0;
}

/* release both kernel and SAM9407 buffers */
static int releaseBuffers(SamCard *card, int channel)
{
  SamChannelInfo *chp=card->channels+channel;
  int error;
  if(!chp->data)
    return 0;

  chp->counter.finishedBytes+=chp->count<<SAM_FORMAT_16BIT(chp->format);

  kfree(chp->data);
  chp->data=NULL;
  chp->alloced=0;
  chp->count=0;
  error=samFree(card, (channel<<8)|SAM_MMT_ID_STR_WAVE, ~SAM_MMT_ROM_MASK);

  atomic_dec(&card->usageCount);

  return error;
}

static void newCounterFrame(SamChannelInfo *chp, int recording, int count)
{
  SamAudioFrameCounter *newFrame, *oldFrame;

  /* playback requires 2 counter frames,
   * while recording can do with just one
   */
  if(!recording) {
    oldFrame=chp->counter.frames+chp->counter.frameIdx;
    newFrame=chp->counter.frames+!chp->counter.frameIdx;
  } else
    oldFrame=newFrame=chp->counter.frames;

  if(newFrame->used)
    chp->counter.finishedBytes+=newFrame->bytes;

  if(oldFrame->used && oldFrame->bytes)
    chp->counter.transitions++;

  if(count>=0) {
    /* channel still open */
    newFrame->bytes=count<<SAM_FORMAT_16BIT(chp->format);
    newFrame->startJiffies=jiffies;
    newFrame->first=!oldFrame->used;
    newFrame->used=1;
  } else {
    /* channel is closing,
     * create an unused frame
     */
    newFrame->bytes=0;
    newFrame->used=0;
  }

  if(!recording)
    chp->counter.frameIdx=!chp->counter.frameIdx;
}

/* calculate the pointer in output/input
 * stream, based on the current time
 */
static unsigned calculateCounterPtr(SamChannelInfo *chp, int recording)
{
  SamAudioFrameCounter *frame;
  int intervalTime, deltaTime, bytes, max;

  /* playback requires 2 counter frames,
   * while recording can do with just one
   */
  if(!recording)
    frame=chp->counter.frames+!chp->counter.frameIdx;
  else
    frame=chp->counter.frames;
    
  if(!frame->used)
    return 0;

  intervalTime=HZ*chp->needed/chp->rate>>!!chp->stereo;
  deltaTime=jiffies-frame->startJiffies;
  /* for playback, starting with the second frame,
   * we're one interval behind
   */
  if(!recording && !frame->first)
    deltaTime-=intervalTime;
  if(deltaTime<0)
    deltaTime=0;

  bytes=(chp->rate<<(SAM_FORMAT_16BIT(chp->format)+!!chp->stereo))*deltaTime/HZ;
  max=frame->bytes;

  if(bytes>>SAM_FORMAT_16BIT(chp->format)>=samWordsNeeded(chp)<<2) {
    /* if bytecount exceeds frame buffer size by 4 times
     * we can safely assume that data has reached it's destination
     * and append 2 unused frames
     */
    newCounterFrame(chp, recording, -1);
    newCounterFrame(chp, recording, -1);
  }

  return bytes<max ? bytes : max;
}

static int openChannel(SamCard *card, int channel)
{
  SamChannelInfo *chp=card->channels+channel;
  unsigned char format;
  int error;

  if(chp->state!=SamChannelStateClosed && chp->state!=SamChannelStateClosing)
    return 0;

  /* when using IRQ's
   * start by filling up half stream
   * to give write(2) system call some time
   * to fill up buffers
   */
  if(card->irq)
    chp->justOpened=1;

  if(chp->stereo)
    format=0x02;
  else
    format=0x00;
  if(card->firmwareInfo->openAck) {
    /* W_OPEN will send acknowledge */
    error=samTalk(card,
		  SamSendCtrl, 5, SAM_CMD_W_OPEN, channel, format, chp->rate, chp->rate>>8,
		  SamEndOfScript);
    chp->state=SamChannelStateNotStarted;
  } else {
    error=samTalk(card,
		  SamSendCtrl, 5, SAM_CMD_W_OPEN, channel, format, chp->rate, chp->rate>>8,
		  SamSendCtrl, 2, SAM_CMD_W_START, channel,
		  SamEndOfScript);
    chp->state=SamChannelStateOpen;
  }

  if(error>=0 && channel>=card->firmwareInfo->playBackChannels)
    newCounterFrame(chp, 1, chp->needed);

  if(error>=0)
  {
    error=adjustVolume(card, channel);

#ifdef _DEBUG
//    printk( "AdjustVol = %x\n", error ); 
#endif	
  }
  return error;
}

/* All audio streaming is done in a separate
 * thread, so we're able to close channels
 * when receiving an asynchronous notification
 * with no caller thread waiting
 */
int samAudioFeedThread(void *arg)
{
  SamCard *card;
  SamChannelInfo *chp;
  unsigned flags;
  int error, haveJob, cardNr, channel, channels;

  current->session=1;
  current->pgrp=1;
  if(current->mm)
    current->mm->arg_end=current->mm->arg_start;
  strcpy(current->comm, "sam9407d");

  save_flags(flags); cli();
  for(;;) {
#if LINUX_VERSION_CODE>=0x20200
    spin_lock_irq(&current->sigmask_lock);
    sigfillset(&current->blocked); /* block all signals */
    recalc_sigpending(current);
    spin_unlock_irq(&current->sigmask_lock);
#else /* Linux 2.0 */
    current->signal=0;
#endif
    interruptible_sleep_on(&samFeedThreadInfo.feedSam);
    if(samFeedThreadInfo.shutdown) {
      restore_flags(flags);
      break;
    }
    do {
      haveJob=0;
      for(cardNr=0; cardNr<samCardsNr; cardNr++) {
	card=samCards[cardNr];

	down(&card->usageCountLock); cli();
	if(!card->firmwareLoaded) {
	  up(&card->usageCountLock);
	  continue;
	}
	atomic_inc(&card->usageCount);
	up(&card->usageCountLock);

	channels=card->firmwareInfo->playBackChannels+card->firmwareInfo->recordChannels;

	for(channel=0; channel<channels; channel++) {
	  chp=card->channels+channel;
	  if(chp->state==SamChannelStateSendStart || \
	     chp->state==SamChannelStateUnderflow || \
	     chp->state==SamChannelStateOverflow) {
	    haveJob++;

	    down(&chp->accessLock);
	    restore_flags(flags);

	    switch(chp->state) {
	    
	    case SamChannelStateSendStart:
	      error=samTalkPickLock(card,
				    SamSendCtrl, 2, SAM_CMD_W_START, channel,
				    SamEndOfScript);
	      if(error<0)

#ifdef _DEBUG
		printk(KERN_ERR "sam9407: error starting audio channel\n");
#endif		

	      chp->state=SamChannelStateOpen;
	      break;
	    
	    case SamChannelStateUnderflow:
	      if(chp->justOpened) {
		/* when using IRQ's
		 * start by filling up half stream
		 * to give write(2) system call some time
		 * to fill up buffers
		 */
		newCounterFrame(chp, 0, 0);

		error=samTalkPickLock(card,
				      SamBurstWriteConst, 0, chp->needed,
				      SamSendCtrl, 2, SAM_CMD_END_XFER, channel,
				      SamEndOfScript);
		if(error<0)
		  printk(KERN_ERR "sam9407: error writing to audio channel\n");

		chp->justOpened=0;
		chp->state=SamChannelStateOpen;
	      } else if(chp->data && chp->count) {
		newCounterFrame(chp, 0, chp->count);

		/* fill up missing data with zeroes */
		if(chp->count<chp->needed)
		  memset(chp->data+chp->count, 0, (chp->needed-chp->count)*2);

		error=samTalkPickLock(card,
				      SamBurstWrite, chp->data, chp->needed,
				      SamSendCtrl, 2, SAM_CMD_END_XFER, channel,
				      SamEndOfScript);
		if(error<0)
		  printk(KERN_ERR "sam9407: error writing to audio channel\n");

		chp->state=SamChannelStateOpen;

		chp->count=0;
		wake_up(&chp->feedMe);
	      } else {
		/* no data, close channel */
		newCounterFrame(chp, 0, -1);

		error=samTalkPickLock(card,
				      SamBurstWriteConst, 0, chp->needed,
				      SamSendCtrl, 2, SAM_CMD_END_XFER, channel,
				      SamSendCtrl, 2, SAM_CMD_W_CLOSE, channel,
				      SamEndOfScript);
		if(error<0)
		  printk(KERN_ERR "sam9407: can't close audio channel\n");
	      
		chp->state=SamChannelStateClosing;

		chp->count=0;
	      
		if(!chp->inUse)
		  /* not even in use anymore, release the channel */
		  if(releaseBuffers(card, channel)<0)
		    printk(KERN_ERR "sam9407: can't release streaming wave buffer\n");
	      }
	      break;

	    case SamChannelStateOverflow:
	      if(chp->data && chp->count==0) {
		newCounterFrame(chp, 1, chp->needed);

		error=samTalkPickLock(card,
				      SamBurstRead, chp->data, chp->needed,
				      SamSendCtrl, 2, SAM_CMD_END_XFER, channel,
				      SamEndOfScript);
		if(error<0)
		  printk(KERN_ERR "sam9407: error reading from audio channel\n");

		chp->state=SamChannelStateOpen;

		chp->count=chp->needed;
		wake_up(&chp->feedMe);
	      } else {
		/* no room to store data, close channel */
		newCounterFrame(chp, 1, -1);

		error=samTalkPickLock(card,
				      SamSendCtrl, 2, SAM_CMD_END_XFER, channel,
				      SamSendCtrl, 2, SAM_CMD_W_CLOSE, channel,
				      SamEndOfScript);
		if(error<0)
		  printk(KERN_ERR "sam9407: can't close audio channel\n");

		if(card->firmwareInfo->closeAckRecChannel)
		  chp->state=SamChannelStateClosing;
		else
		  chp->state=SamChannelStateClosed;

		if(!chp->inUse)
		  /* not even in use anymore, release the channel */
		  if(releaseBuffers(card, channel)<0)
		    printk(KERN_ERR "sam9407: can't release streaming wave buffer\n");
	      }
	      break;
	    default: /* make compiler happy */
	    }

	    up(&chp->accessLock);
	    save_flags(flags); cli();
	  }
	}

	/* samAudioFeedThread is shared with the sequencer
	 * to implement asynchronous playback
	 * of sequencer events
	 */
	samCheckPlaybackSeq(card);

	atomic_dec(&card->usageCount);
      }
    } while(haveJob);
  }
  samFeedThreadInfo.shutdown=0;
  return 0;
}

int samStartFeedThread(void)
{
  int error;

  /* start audio streaming thread */
  if((error=kernel_thread(samAudioFeedThread, NULL, 0))<0) {
    printk(KERN_ERR "sam9407: can't fork off audio streaming thread.\n");
    return error;
  }

  return 0;
}

void samKillFeedThread(void)
{
  unsigned long start;

  samFeedThreadInfo.shutdown=1;
  start=jiffies;
  while(samFeedThreadInfo.shutdown && jiffies-start<5*HZ) {
    wake_up(&samFeedThreadInfo.feedSam);
#if LINUX_VERSION_CODE>=0x20200
    if(current->need_resched)
      schedule();
#else /* Linux 2.0 */
    if(need_resched)
      schedule();
#endif /* Linux 2.0 */
  }
  if(samFeedThreadInfo.shutdown)
    panic("sam9407: oops, can't kill audio streaming thread.\n");
}

int samAllocateChannels(SamCard *card)
{
  int size, nChannels, channel;
  SamChannelInfo *channels;

  nChannels=card->firmwareInfo->playBackChannels+card->firmwareInfo->recordChannels;
  size=nChannels*sizeof(SamChannelInfo);

  channels=(SamChannelInfo *)kmalloc(size, GFP_KERNEL);
  if(!channels)
    return -ENOMEM;

  memset(channels, 0, size);
  for(channel=0; channel<nChannels; channel++)
    channels[channel].accessLock=MUTEX;

  card->channels=channels;

  return 0;
}

void samReleaseChannels(SamCard *card)
{
  int nChannels, channel;
  SamChannelInfo *channels;

  if(!(channels=card->channels))
    return;

  card->channels=NULL;

  /* release any pending audio buffers,
   * as we know nobody is working on them anymore
   */
  nChannels=card->firmwareInfo->playBackChannels+card->firmwareInfo->recordChannels;
  for(channel=0; channel<nChannels; channel++)
    if(channels[channel].data)
      kfree(channels[channel].data);
  
  kfree(channels);
}

void samAudioNotification(SamCard *card, unsigned v)
{
  unsigned flags;
  int nChannels;
  save_flags(flags); cli();
  if(!card->firmwareLoaded) {
    restore_flags(flags);
    return;
  }
  nChannels=card->firmwareInfo->playBackChannels+card->firmwareInfo->recordChannels;
  if(v<card->firmwareInfo->playBackChannels) {
    card->channels[v].state=SamChannelStateUnderflow;
    wake_up(&samFeedThreadInfo.feedSam);
  } else if(v>=card->firmwareInfo->playBackChannels && v<nChannels) {
    card->channels[v].state=SamChannelStateOverflow;
    wake_up(&samFeedThreadInfo.feedSam);
  } else if(card->firmwareInfo->openAck && v>=card->firmwareInfo->openAck && v<card->firmwareInfo->openAck+nChannels) {
    card->channels[v-card->firmwareInfo->openAck].state=SamChannelStateSendStart;
    wake_up(&samFeedThreadInfo.feedSam);
  } else if(v>=card->firmwareInfo->closeAck && v<card->firmwareInfo->closeAck+nChannels) {
    card->channels[v-card->firmwareInfo->closeAck].state=SamChannelStateClosed;
    wake_up(&card->channels[v-card->firmwareInfo->closeAck].waitClosed);
  }
  restore_flags(flags);
}

int samOpenAudio(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  SamChannelInfo *chp;
  int channel, start, end;

  if(mode&SAM_MODE_READ_MASK) {
    start=card->firmwareInfo->playBackChannels;
    end=start+card->firmwareInfo->recordChannels;
  } else {
    start=0;
    end=card->firmwareInfo->playBackChannels;
  }

  /* check for a free channel */
  for(channel=start; channel<end; channel++) {
    chp=card->channels+channel;
    down(&chp->accessLock);
    if(!chp->inUse && chp->state==SamChannelStateClosed) {
      chp->inUse=1;
      *closure=channel;
      up(&chp->accessLock);
      break;
    }
    up(&chp->accessLock);
  }

  if(channel>=end)
    return -EBUSY;

  down(&chp->accessLock);
  switch(minor) {
  case SoundMinorAudio:
    chp->format=AFMT_MU_LAW;
    chp->rate=8000;
    chp->stereo=0;
    break;
  default:
    chp->format=AFMT_S16_LE;
    if(mode&SAM_MODE_READ_MASK)
      /* record mode */
      chp->rate=22050;
    else
      /* playback mode */
      chp->rate=44100;
#ifdef SAM_BC_DSP_IS_MONO
    chp->stereo=0;
#else
    chp->stereo=1;
#endif
  }

  if(card->firmwareInfo->volFormatEWS64) {
    chp->volume.left=samCvtVolToEWS64(SAM_DEFAULT_VOLUME);
    chp->volume.right=samCvtVolToEWS64(SAM_DEFAULT_VOLUME);
    chp->volume.auxLeft=samCvtVolToEWS64(SAM_DEFAULT_VOLUME);
    chp->volume.auxRight=samCvtVolToEWS64(SAM_DEFAULT_VOLUME);
  } else {
    chp->volume.left=SAM_DEFAULT_VOLUME;
    chp->volume.right=SAM_DEFAULT_VOLUME;
    chp->volume.auxLeft=SAM_DEFAULT_VOLUME;
    chp->volume.auxRight=SAM_DEFAULT_VOLUME;
  }

  memset(&chp->counter, 0, sizeof(chp->counter));

  up(&chp->accessLock);
  return 0;
}

void samReleaseAudio(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  int channel=*closure;
  SamChannelInfo *chp=card->channels+channel;
  unsigned flags;

  if(mode&SAM_MODE_WRITE_MASK) {
    /* finish a pending buffer */
    down(&chp->accessLock);
    if(chp->count>0 && chp->count<chp->needed) {
      memset(chp->data+chp->count, 0, (chp->needed-chp->count)<<1);
      chp->count=chp->needed;
    }
    up(&chp->accessLock);
  }

  /* linux-2.2.0-pre4 requires current->sigpending to be recalculated
   * whenever a signal was received, that signal wasn't available
   * in release anymore.
   */
#if defined(SAM_LINUX22_SIGPENDING_BUG) && LINUX_VERSION_CODE>=0x20200
  spin_lock_irq(&current->sigmask_lock);
  recalc_sigpending(current);
  spin_unlock_irq(&current->sigmask_lock);
#endif /* SAM_LINUX22_SIGPENDING_BUG */

  save_flags(flags); cli();
  while(chp->state!=SamChannelStateClosed) {
    interruptible_sleep_on(&chp->waitClosed);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current))
      break;
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked)
      break;
#endif /* Linux 2.0 */
  }
  restore_flags(flags);
    
  down(&chp->accessLock);
  if(chp->state==SamChannelStateClosed || chp->state==SamChannelStateClosing)
    /* release channel if it's not open anymore */
    if(releaseBuffers(card, channel)<0)
      printk(KERN_ERR "sam9407: can't release streaming wave buffer\n");
  chp->inUse=0;
  up(&chp->accessLock);
}

int samReadAudio(struct inode *inode, struct file *file, char *buf, int count, SamCard *card, char *closure)
{
  int channel=*closure;
  SamChannelInfo *chp=card->channels+channel;
  unsigned flags;
  short *sp;
  unsigned char *cp;
  int error, atomicSize, avail, n, i, v;

#ifdef _DEBUG
  printk( "samReadAudio = %d\n", count );
#endif 
  
  if(count<0)
    return -EINVAL;

  if((error=verify_area(VERIFY_WRITE, buf, count))<0)
    return error;

  down(&chp->accessLock);

  atomicSize=1<<(SAM_FORMAT_16BIT(chp->format)+!!chp->stereo);

#ifndef SAM_BC_NO_ATOMIC_CHECK
  if(count%atomicSize) {
    up(&chp->accessLock);
    return -EINVAL;
  }
#endif

  cp=(char *)buf;
  n=count;

  /* allocate buffers and open recording channel
   */
  if((error=allocBuffers(card, channel))<0 ||
     (error=openChannel(card, channel))<0) {
    up(&chp->accessLock);
    return error;
  }
  
  if(chp->count==0 && (file->f_flags & O_NONBLOCK)) {
    up(&chp->accessLock);
    return -EAGAIN;
  }

  /* wait for incoming data */
  while(chp->count==0) {
    save_flags(flags); cli();
    up(&chp->accessLock);
    interruptible_sleep_on(&chp->feedMe);
    restore_flags(flags);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current))
      return -ERESTARTSYS;
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked)
      return -ERESTARTSYS;
#endif /* Linux 2.0 */
    down(&chp->accessLock);
  }

  /* convert the data */
  sp=chp->data+(chp->needed-chp->count);
  avail=chp->count;

  if(SAM_FORMAT_16BIT(chp->format)) {
    if(avail>n>>1)
      avail=n>>1;
  } else {
    if(avail>n)
      avail=n;
  }

  switch(chp->format) {
  case AFMT_MU_LAW:
    for(i=0; i<avail; i++)
      SAM_PUT_USER(samShort2Ulaw(sp[i]), cp+i);
    break;
  case AFMT_S8:
    for(i=0; i<avail; i++)
      SAM_PUT_USER(sp[i]>>8, (char *)cp+i);
    break;
  case AFMT_U8:
    for(i=0; i<avail; i++)
      SAM_PUT_USER((sp[i]+0x8000)>>8, cp+i);
    break;
  case AFMT_U16_LE:
    for(i=0; i<avail; i++)
      SAM_PUT_USER(sp[i]+0x8000, (unsigned short *)cp+i);
    break;
  case AFMT_S16_BE:
    for(i=0; i<avail; i++)
      SAM_PUT_USER(((unsigned short)sp[i]>>8)|((unsigned short)sp[i]<<8), (short *)cp+i);
    break;
  case AFMT_U16_BE:
    for(i=0; i<avail; i++) {
      v=sp[i]+0x8000;
      SAM_PUT_USER((v>>8)|(v<<8), (unsigned short *)cp+i);
    }
    break;
  default: /* AFMT_S16_LE, our native format */
    SAM_MEMCPY_TOFS(cp, sp, avail<<1);
  }

  chp->count-=avail;
  if(SAM_FORMAT_16BIT(chp->format))
    cp+=avail<<1;
  else
    cp+=avail;

  up(&chp->accessLock);

  return cp-(unsigned char *)buf;
}

int samWriteAudio(struct inode *inode, struct file *file, const char *buf, int count, SamCard *card, char *closure)
{
  int channel=*closure;
  SamChannelInfo *chp=card->channels+channel;
  unsigned flags;
  short *sp;
  unsigned char *cp;
  int error, atomicSize, avail, n, i, v;

#ifdef _DEBUG
  printk( "samWriteAudio = %d\n", count );
#endif  
  
  if(count<0)
    return -EINVAL;

  if((error=verify_area(VERIFY_READ, buf, count))<0)
    return error;

  down(&chp->accessLock);

  atomicSize=1<<(SAM_FORMAT_16BIT(chp->format)+!!chp->stereo);

#ifndef SAM_BC_NO_ATOMIC_CHECK
  if(count%atomicSize) {
    up(&chp->accessLock);
    return -EINVAL;
  }
#endif

  cp=(char *)buf;
  n=count;

  if(chp->data && chp->count>=chp->needed && (file->f_flags & O_NONBLOCK)) {
    up(&chp->accessLock);
    return -EAGAIN;
  }

#ifdef SAM_BC_WRITE_DELIVERS_ALL
 deliver_again:
#endif

  /* wait until pending request finished
   */
  while(chp->data && chp->count>=chp->needed) {
    save_flags(flags); cli();
    up(&chp->accessLock);
    /* signal inode semaphore to give other
     * channels a chance to receive data as well
     */
    up(&inode->i_sem);
    interruptible_sleep_on(&chp->feedMe);
    restore_flags(flags);
    down(&inode->i_sem);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current))
      return -ERESTARTSYS;
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked)
      return -ERESTARTSYS;
#endif /* Linux 2.0 */
    down(&chp->accessLock);
  }

  /* okay, the stage is ours
   * allocate buffers first
   */
  if((error=allocBuffers(card, channel))<0) {
    up(&chp->accessLock);
    return error;
  }

#ifdef _DEBUG
//  printk( "allocBuffers ok = %d\n", channel );
#endif  
  
  /* convert the data */
  sp=chp->data+chp->count;
  avail=chp->needed-chp->count;

  if(SAM_FORMAT_16BIT(chp->format)) {
    if(avail>n>>1)
      avail=n>>1;
  } else {
    if(avail>n)
      avail=n;
  }

  switch(chp->format) {
  case AFMT_MU_LAW:
    for(i=0; i<avail; i++)
      sp[i]=samUlaw2Short(SAM_GET_USER(cp+i));
    break;
  case AFMT_S8:
    for(i=0; i<avail; i++)
      sp[i]=SAM_GET_USER((char *)cp+i)<<8;
    break;
  case AFMT_U8:
    for(i=0; i<avail; i++)
      sp[i]=(SAM_GET_USER(cp+i)<<8)-0x8000;
    break;
  case AFMT_U16_LE:
    for(i=0; i<avail; i++)
      sp[i]=(int)SAM_GET_USER((unsigned short *)cp+i)-0x8000;
    break;
  case AFMT_S16_BE:
    for(i=0; i<avail; i++) {
      v=SAM_GET_USER((unsigned short *)cp+i);
      sp[i]=(v>>8)|(v<<8);
    }
    break;
  case AFMT_U16_BE:
    for(i=0; i<avail; i++) {
      v=SAM_GET_USER((unsigned short *)cp+i);
      sp[i]=(((v>>8)|(v<<8))&0xFFFF)-0x8000;
    }
    break;
  default: /* AFMT_S16_LE, our native format */
    SAM_MEMCPY_FROMFS(sp, cp, avail<<1);
  }

  chp->count+=avail;
  if(SAM_FORMAT_16BIT(chp->format)) {
    cp+=avail<<1;
    n-=avail<<1;
  } else {
    cp+=avail;
    n-=avail;
  }
  
  /* if enough data is available, get the sound started if not done yet */
  if(chp->count>=chp->needed) {
    if((error=openChannel(card, channel))<0) {
      up(&chp->accessLock);
      return error;
    }
  }
#ifdef _DEBUG
//  printk( "openChannel = %d\n", channel );
#endif  

#ifdef SAM_BC_WRITE_DELIVERS_ALL
  if(n>0)
    goto deliver_again;
#endif

  up(&chp->accessLock);

  return cp-(unsigned char *)buf;
//  return count;
}

#if LINUX_VERSION_CODE>=0x20200
unsigned samPollAudio(struct inode *inode, struct file *file, poll_table *wait, SamCard *card, mode_t mode, char *closure)
{
  int channel=*closure;
  SamChannelInfo *chp=card->channels+channel;
  unsigned mask;

  poll_wait(file, &chp->feedMe, wait);

  down(&chp->accessLock);
  mask=0;
  if(mode & SAM_MODE_READ_MASK) {
    if(chp->count)
      mask|= POLLIN | POLLRDNORM;
  } else {
    if(chp->count<chp->needed)
      mask|= POLLOUT | POLLWRNORM;
  }
  up(&chp->accessLock);

  return mask;
}
#else /* Linux 2.0 */
int samSelectAudio(struct inode *inode, struct file *file, int type, select_table *wait, SamCard *card, char *closure)
{
  int channel=*closure;
  SamChannelInfo *chp=card->channels+channel;
  unsigned flags;

  switch(type) {
  case SEL_IN:
    down(&chp->accessLock);
    if(chp->count) {
      up(&chp->accessLock);
      return 1;
    }
    save_flags(flags); cli();
    up(&chp->accessLock);
    select_wait(&chp->feedMe, wait);
    restore_flags(flags);
    break;
  case SEL_OUT:
    down(&chp->accessLock);
    if(chp->count<chp->needed) {
      up(&chp->accessLock);
      return 1;
    }
    save_flags(flags); cli();
    up(&chp->accessLock);
    select_wait(&chp->feedMe, wait);
    restore_flags(flags);
    break;
  }
  return 0;
}
#endif /* Linux 2.0 */

int samIoctlAudio(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg, SamCard *card, mode_t mode, char *closure, int callNr)
{
  int channel=*closure;
  SamChannelInfo *chp=card->channels+channel;
  unsigned flags, v;
  int error=0;
  SamVolume *volume;
  audio_buf_info bufInfo;
  count_info countInfo;

  switch(cmd) {
  case SNDCTL_DSP_RESET:
    down(&chp->accessLock);
    
#ifdef _DEBUG
    printk( "SNDCTL_DSP_RESET : %d\n", channel );
#endif	

    error=releaseBuffers(card, channel);
    up(&chp->accessLock);
    break;
  case SNDCTL_DSP_SYNC:
  case SNDCTL_DSP_POST:
    down(&chp->accessLock);

#ifdef _DEBUG
    printk( "SNDCTL_DSP_SYNC, POST : %d\n", channel );
#endif	
    
    if((mode & SAM_MODE_WRITE_MASK) && chp->data && chp->count) {
      save_flags(flags); cli();
      up(&chp->accessLock);
      interruptible_sleep_on(&chp->feedMe);
      restore_flags(flags);
#if LINUX_VERSION_CODE>=0x20200
      if(signal_pending(current))
	error=-ERESTARTSYS;
#else /* Linux 2.0 */
      if(current->signal & ~current->blocked)
	error=-ERESTARTSYS;
#endif /* Linux 2.0 */
    } else
      up(&chp->accessLock);
    break;
  case SNDCTL_DSP_SPEED:
    if((error=verify_area(VERIFY_WRITE, (unsigned *)arg, sizeof(unsigned)))<0)
      return error;
    v=SAM_GET_USER((unsigned *)arg);
    if(mode&SAM_MODE_READ_MASK) {
      /* record mode */
      if(v<SAM_MIN_RATE)
	v=SAM_MIN_RATE;
      if(v>SAM_MAX_RECORD_RATE)
	v=SAM_MAX_RECORD_RATE;
    } else {
      /* play mode */
      if(v<SAM_MIN_RATE)
	v=SAM_MIN_RATE;
      if(v>SAM_MAX_PLAYBACK_RATE)
	v=SAM_MAX_PLAYBACK_RATE;
    }
#ifdef _DEBUG
    printk( "SNDCTL_DSP_SEEP : %d=%d\n", channel, v );
#endif	
    
    SAM_PUT_USER(v, (unsigned *)arg);
    down(&chp->accessLock);
    if(chp->rate!=v)
      if((error=releaseBuffers(card, channel))>=0)
	chp->rate=v;
    up(&chp->accessLock);
    break;
  case SNDCTL_DSP_STEREO:
    if((error=verify_area(VERIFY_READ, (unsigned *)arg, sizeof(unsigned)))<0)
      return error;
    down(&chp->accessLock);
    if(chp->stereo!=!!SAM_GET_USER((unsigned *)arg))
      if((error=releaseBuffers(card, channel))>=0)
	chp->stereo=!!SAM_GET_USER((unsigned *)arg);

#ifdef _DEBUG
    printk( "SNDCTL_DSP_STEREO : %d=%d\n", channel, chp->stereo );
#endif	
	
    up(&chp->accessLock);
    break;
  case SNDCTL_DSP_GETBLKSIZE:
    if(callNr==0) {
      if((error=verify_area(VERIFY_WRITE, (unsigned *)arg, sizeof(unsigned)))<0)
	return error;
      SAM_PUT_USER(samWordsNeeded(chp)<<1, (unsigned *)arg);
    }
    break;
  case SOUND_PCM_WRITE_CHANNELS:
#ifdef _DEBUG
	printk( "SOUND_PCM_WRITE_CHANNELS \n" );
#endif	
	
    if((error=verify_area(VERIFY_WRITE, (int *)arg, sizeof(int)))<0)
      return error;
    if(SAM_GET_USER((int *)arg)==1 || SAM_GET_USER((int *)arg)==2) {
      down(&chp->accessLock);
      if(chp->stereo!=SAM_GET_USER((int *)arg)==2) {
	if((error=releaseBuffers(card, channel))>=0)
	  chp->stereo=SAM_GET_USER((int *)arg)==2;
      } else
	SAM_PUT_USER(!!chp->stereo+1, (int *)arg);
      up(&chp->accessLock);
    }
    break;
  case SOUND_PCM_WRITE_FILTER:
    return -EINVAL;
    break;
  case SNDCTL_DSP_SUBDIVIDE:
    /* ignored */
    break;
  case SNDCTL_DSP_SETFRAGMENT:
    /* ignore fragments, we keep our own sizes */
    break;
  case SNDCTL_DSP_GETFMTS:
    if(callNr==0) {
      if((error=verify_area(VERIFY_WRITE, (unsigned *)arg, sizeof(unsigned)))<0)
	return error;
      SAM_PUT_USER(AFMT_MU_LAW |
		   AFMT_U8 |
		   AFMT_S16_LE |
		   AFMT_S16_BE |
		   AFMT_S8 |
		   AFMT_U16_LE |
		   AFMT_U16_BE, (unsigned *)arg);
    }
    break;
  case SNDCTL_DSP_SETFMT:
    if((error=verify_area(VERIFY_READ, (unsigned *)arg, sizeof(unsigned)))<0)
      return error;
    v=SAM_GET_USER((unsigned *)arg);
    if(v!=AFMT_MU_LAW &&
       v!=AFMT_U8 &&
       v!=AFMT_S16_LE &&
       v!=AFMT_S16_BE &&
       v!=AFMT_S8 &&
       v!=AFMT_U16_LE &&
       v!=AFMT_U16_BE)
      return -EINVAL;

#ifdef _DEBUG
    printk( "SNDCTL_DSP_SETFMT : %d=%d\n", channel, v );
#endif	
      
    down(&chp->accessLock);
    if(chp->format!=v)
      if((error=releaseBuffers(card, channel))>=0)
	chp->format=v;
    up(&chp->accessLock);
    break;
  case SNDCTL_DSP_GETOSPACE:
  case SNDCTL_DSP_GETISPACE:
#ifdef _DEBUG
	printk( "SOUND_DSP_GETOSPACE \n" );
#endif	
	
    if((cmd==SNDCTL_DSP_GETOSPACE && (mode & SAM_MODE_WRITE_MASK)) ||
       (cmd==SNDCTL_DSP_GETISPACE && (mode & SAM_MODE_READ_MASK))) {
      if((error=verify_area(VERIFY_WRITE, (audio_buf_info *)arg, sizeof(audio_buf_info)))<0)
		return error;
      memset(&bufInfo, 0, sizeof(audio_buf_info));
      down(&chp->accessLock);
      bufInfo.fragments=chp->count ? 0 : 1;
      bufInfo.fragstotal=1;
      bufInfo.fragsize=samWordsNeeded(chp)<<1;
      bufInfo.bytes=(samWordsNeeded(chp)-chp->count)<<1;
      up(&chp->accessLock);
      SAM_MEMCPY_TOFS((audio_buf_info *)arg, &bufInfo, sizeof(audio_buf_info));
    }
    break;
  case SNDCTL_DSP_NONBLOCK:
    /* why does OSS carry a separate ioctl(2) for nonblocking I/O ? */
    break;
  case SNDCTL_DSP_GETCAPS:
    if(callNr==0) {
      if((error=verify_area(VERIFY_WRITE, (unsigned *)arg, sizeof(unsigned)))<0)
	return error;
      SAM_PUT_USER(DSP_CAP_DUPLEX|DSP_CAP_BATCH|DSP_CAP_COPROC, (unsigned *)arg);
    }
    break;
  case SNDCTL_DSP_SETTRIGGER:
    /* we don't support SETTRIGGER,
     * so fall just through next case
     * and return results from GETTRIGGER
     */
  case SNDCTL_DSP_GETTRIGGER:
    if((error=verify_area(VERIFY_WRITE, (unsigned *)arg, sizeof(unsigned)))<0)
      return error;
    v=0;
    if(mode & SAM_MODE_WRITE_MASK)
      v=PCM_ENABLE_OUTPUT;
    if(mode & SAM_MODE_READ_MASK)
      v=PCM_ENABLE_INPUT;
    if(callNr)
      v|=SAM_GET_USER((unsigned *)arg);
    SAM_PUT_USER(v, (unsigned *)arg);
    break;
  case SNDCTL_DSP_GETIPTR:
  case SNDCTL_DSP_GETOPTR:
    if((cmd==SNDCTL_DSP_GETOPTR && (mode & SAM_MODE_WRITE_MASK)) ||
       (cmd==SNDCTL_DSP_GETIPTR && (mode & SAM_MODE_READ_MASK))) {
      if((error=verify_area(VERIFY_WRITE, (count_info *)arg, sizeof(count_info)))<0)
	return error;
      memset(&countInfo, 0, sizeof(countInfo));
      down(&chp->accessLock);
      countInfo.ptr=calculateCounterPtr(chp, mode & SAM_MODE_READ_MASK);
      countInfo.bytes=chp->counter.finishedBytes+countInfo.ptr;
      countInfo.blocks=chp->counter.transitions;
      chp->counter.transitions=0;
      up(&chp->accessLock);
      SAM_MEMCPY_TOFS((count_info *)arg, &countInfo, sizeof(count_info));
    }
    break;
  case SNDCTL_DSP_MAPINBUF:
    return -EINVAL;
    break;
  case SNDCTL_DSP_MAPOUTBUF:
    return -EINVAL;
    break;
  case SNDCTL_DSP_SETSYNCRO:
    return -EINVAL;
    break;
  case SNDCTL_DSP_SETDUPLEX:
    /* we're always duplex */
    break;
  case SAM_IOC_VOLUME_GET:
    if(callNr==0) {
      volume=(SamVolume *)arg;
      if((error=verify_area(VERIFY_WRITE, volume, sizeof(SamVolume)))<0)
	return error;
      down(&chp->accessLock);
      if(card->firmwareInfo->volFormatEWS64) {
	SAM_PUT_USER(samCvtVolFromEWS64(chp->volume.left), &volume->left);
	SAM_PUT_USER(samCvtVolFromEWS64(chp->volume.right), &volume->right);
	SAM_PUT_USER(samCvtVolFromEWS64(chp->volume.auxLeft), &volume->auxLeft);
	SAM_PUT_USER(samCvtVolFromEWS64(chp->volume.auxRight), &volume->auxRight);
      } else {
	SAM_MEMCPY_TOFS((SamVolume *)arg, &chp->volume, sizeof(SamVolume));
      }
	
#ifdef _DEBUG
    printk( "IOC Vol %x, %x, %x, %x\n", 
          SAM_GET_USER(&volume->left),
          SAM_GET_USER(&volume->right),
          SAM_GET_USER(&volume->auxLeft),
          SAM_GET_USER(&volume->auxRight));
#endif
	
      up(&chp->accessLock);
    }
    break;
  case SAM_IOC_VOLUME_SET:
    volume=(SamVolume *)arg;
    if((error=verify_area(VERIFY_READ, volume, sizeof(SamVolume)))<0)
      return error;
    down(&chp->accessLock);

#ifdef _DEBUG
    printk( "IOC Vol %x, %x, %x, %x\n", 
          SAM_GET_USER(&volume->left),
          SAM_GET_USER(&volume->right),
          SAM_GET_USER(&volume->auxLeft),
          SAM_GET_USER(&volume->auxRight));
#endif

    if(card->firmwareInfo->volFormatEWS64) {
      chp->volume.left=samCvtVolToEWS64(SAM_GET_USER(&volume->left));
      chp->volume.right=samCvtVolToEWS64(SAM_GET_USER(&volume->right));
      chp->volume.auxLeft=samCvtVolToEWS64(SAM_GET_USER(&volume->auxLeft));
      chp->volume.auxRight=samCvtVolToEWS64(SAM_GET_USER(&volume->auxRight));
    } else {
      SAM_MEMCPY_FROMFS(&chp->volume, (SamVolume *)arg, sizeof(SamVolume));
    }
    if(chp->state!=SamChannelStateClosed && chp->state!=SamChannelStateClosing)
      error=adjustVolume(card, channel);
    up(&chp->accessLock);
    break;
  default:
#ifdef _DEBUG
    printk( "default : %x\n", cmd );
#endif	
  
    return -ENOSYS;
  }
  return error;
}

void samGetAudioInfo(SamCard *card, SamDriverInfo *driverInfo)
{
  int max, i;

  driverInfo->playChannelsUsed=driverInfo->recordChannelsUsed=0;
  driverInfo->playChannelsTotal=card->firmwareInfo->playBackChannels;
  driverInfo->recordChannelsTotal=card->firmwareInfo->recordChannels;

  max=card->firmwareInfo->playBackChannels+card->firmwareInfo->recordChannels;
  for(i=0; i<max; i++) {
    if(card->channels[i].inUse || card->channels[i].state!=SamChannelStateClosed)
      if(i<card->firmwareInfo->playBackChannels)
	driverInfo->playChannelsUsed++;
      else
	driverInfo->recordChannelsUsed++;
  }
}
