/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE<0x20200
#include <linux/bios32.h>
#endif
#include <linux/pci.h>
#include <asm/io.h>

#include "driver.h"
#include "cards.h"

#define SAM_PNP_ADDRPORT	0x279
#define SAM_PNP_WRITEPORT	0xA79
#define SAM_PNP_READPORT	pnpReadPort

#define SAM_PNP_REG_RDPORT	0x00
#define SAM_PNP_REG_ISOLATE	0x01
#define SAM_PNP_REG_CC		0x02
#define SAM_PNP_REG_WAKE	0x03
#define SAM_PNP_REG_CSN		0x06
#define SAM_PNP_REG_LDN		0x07
#define SAM_PNP_REG_PORT_BASE	0x60
#define SAM_PNP_REG_IRQ_BASE	0x70

#define SAM_PNP_REG_CC_WAIT_KEY	0x02

#define SAM_PNP_CC_RESET_CSN	0x04
#define SAM_PNP_CC_RESET_DRV	0x07

static int pnpReadPort;

static void pnpInitKey(void)
{
  int i;
  unsigned char lfsr;

  outb(0x00, SAM_PNP_ADDRPORT);
  outb(0x00, SAM_PNP_ADDRPORT);

  lfsr=0x6A;

  for(i=0; i<32; i++) {
    outb(lfsr, SAM_PNP_ADDRPORT);
    lfsr=(((lfsr&1)^((lfsr>>1)&1))<<7)|(lfsr>>1);
  }
}

extern inline void samPnpOut(int value, int reg)
{
  outb(reg, SAM_PNP_ADDRPORT);
  outb(value, SAM_PNP_WRITEPORT);
}

extern inline unsigned samPnpIn(int reg)
{
  outb(reg, SAM_PNP_ADDRPORT);
  return inb(SAM_PNP_READPORT);
}

static int pnpIsolateCard(int csn)
{
  int i, j;
  int isConfigured;
  SamHardwareResource res[SamHardwareResSentinel];
  unsigned v;
  unsigned char lfsr, id[9], c1, c2, b;
  SamHardwareInfo *hardwareInfo;
  SamPnPInfo *pnpInfo;
  SamHardwareResType resType;
  SamPnpResource *resource;

  samPnpOut(0x00, SAM_PNP_REG_WAKE);
  lfsr=0x6A;
  
  outb(SAM_PNP_REG_ISOLATE, SAM_PNP_ADDRPORT);
  udelay(750);
  for(i=0; i<9; i++) {
    id[i]=0;
    for(j=0; j<8; j++) {
      udelay(250);
      b=0;
      c1=inb(SAM_PNP_READPORT);
      c2=inb(SAM_PNP_READPORT);
      if(c1==0x55 && c2==0xAA)
	b=1;
      id[i]|=b<<j;
      if(i<8)
	lfsr=(((lfsr&1)^((lfsr>>1)&1)^b)<<7)+(lfsr>>1);
    }
  }
  if(id[8]!=lfsr)
    return 0;
  samPnpOut(csn, SAM_PNP_REG_CSN);

  /* we've found an ISA PnP board
   * check if the ID# is known
   */
  for(pnpInfo=samPnPInfos; pnpInfo->id; pnpInfo++) {
    if(pnpInfo->id==*(unsigned *)id) {
      hardwareInfo=samHardwareInfos+pnpInfo->hardwareType;
      memset(res, 0, sizeof(res));
      isConfigured=1;
      samPnpOut(csn, SAM_PNP_REG_WAKE);
      for(resType=0; resType<SamHardwareResSentinel; resType++) {
	resource=pnpInfo->resources+resType;

	v=0;
	if(resource->id) {
	  samPnpOut(resource->ldn, SAM_PNP_REG_LDN);
	  switch(resType) {
	  case SamHardwareResPort:
	  case SamHardwareResAuxPort0:
	  case SamHardwareResAuxPort1:
	    v=samPnpIn(SAM_PNP_REG_PORT_BASE+2*resource->idx)<<8 |
	      samPnpIn(SAM_PNP_REG_PORT_BASE+2*resource->idx+1);
	    break;
	  case SamHardwareResIrq:
	    v=samPnpIn(SAM_PNP_REG_IRQ_BASE+2*resource->idx);
	    break;
	  default:
	    /* make compiler happy */
	  }
	}

	if(!v && !resource->isOptional) {
	  isConfigured=0;
	  break;
	}

	res[resType].value=v;
	res[resType].id=resource->id;
      }
      if(isConfigured) {
	printk(KERN_INFO "%s port=0x%x, irq=%d\n",
	       hardwareInfo->name,
	       res[SamHardwareResPort].value,
	       res[SamHardwareResIrq].value);
	if(samRegisterCard(pnpInfo->hardwareType, res)<0)
	  printk(KERN_ERR "sam9407: couldn't register card %s\n", hardwareInfo->name);
      } else
	printk(KERN_INFO "sam9407: board %s is not PnP configured\n", hardwareInfo->name);
      break;
    }
  }

  return 1;
}

static void findPNPCards(void)
{
  int csn;
  samPnpOut(SAM_PNP_CC_RESET_DRV, SAM_PNP_REG_CC);

  pnpInitKey();

  /* Reset all CSNs */
  samPnpOut(SAM_PNP_CC_RESET_CSN, SAM_PNP_REG_CC);

  csn=1;
  for(pnpReadPort=0x203; pnpReadPort<0x400; pnpReadPort+=8) {
    if(check_region(pnpReadPort, 1)<0)
      continue;

    samPnpOut(0x00, SAM_PNP_REG_WAKE);
    samPnpOut(pnpReadPort>>2, SAM_PNP_REG_RDPORT);

    if(pnpIsolateCard(csn)) {
      csn++;
      break;
    }
  }
  if(pnpReadPort>=0x400) {
    samPnpOut(SAM_PNP_REG_CC_WAIT_KEY, SAM_PNP_REG_CC);
    return;
  }

  while(pnpIsolateCard(csn++));

  samPnpOut(SAM_PNP_REG_CC_WAIT_KEY, SAM_PNP_REG_CC);
}

#ifdef CONFIG_PCI
static void findPCICards(void)
{
  SamHardwareInfo *hardwareInfo;
  SamPCIInfo *pciInfo;
  int status, idx, isConfigured;
#if LINUX_VERSION_CODE>=0x20200
  struct pci_dev *pciDev;
#else /* Linux 2.0 */
  unsigned char bus, devFn;
#endif
  unsigned char b;
  unsigned v;
  SamHardwareResource res[SamHardwareResSentinel];
  SamHardwareResType resType;
  SamPCIResource *resource;

  for(pciInfo=samPCIInfos; pciInfo->vendor; pciInfo++) {
    hardwareInfo=samHardwareInfos+pciInfo->hardwareType;
#if LINUX_VERSION_CODE>=0x20200
    pciDev=NULL;
#endif
    for(idx=0;;idx++) {
#if LINUX_VERSION_CODE>=0x20200
      if(!(pciDev=pci_find_device(pciInfo->vendor, pciInfo->devId, pciDev)))
	break;
#else /* Linux 2.0 */
      if(pcibios_find_device(pciInfo->vendor, pciInfo->devId, idx, &bus, &devFn)!=PCIBIOS_SUCCESSFUL)
	break;
#endif /* Linux 2.0 */
      memset(res, 0, sizeof(res));
      isConfigured=1;
      for(resType=0; resType<SamHardwareResSentinel; resType++) {
	resource=pciInfo->resources+resType;
	v=0;
	if(resource->id) {
	  switch(resType) {
	  case SamHardwareResPort:
	  case SamHardwareResAuxPort0:
	  case SamHardwareResAuxPort1:
#if LINUX_VERSION_CODE>=0x20200
	    status=pci_read_config_dword(pciDev, resource->where, &v);
#else /* Linux 2.0 */
	    status=pcibios_read_config_dword(bus, devFn, resource->where, &v);
#endif /* Linux 2.0 */
	    if(status==PCIBIOS_SUCCESSFUL)
	      v&=PCI_BASE_ADDRESS_IO_MASK;
	    else
	      v=0;
	    break;
	  case SamHardwareResIrq:
#if LINUX_VERSION_CODE>=0x20200
	    status=pci_read_config_byte(pciDev, resource->where, &b);
#else /* Linux 2.0 */
	    status=pcibios_read_config_byte(bus, devFn, resource->where, &b);
#endif /* Linux 2.0 */
	    if(status==PCIBIOS_SUCCESSFUL && b!=0xFF)
	      v=b;
	    break;
	  default:
	    /* make compiler happy */
	  }
	}

	if(!v && !resource->isOptional) {
	  isConfigured=0;
	  break;
	}

	res[resType].value=v;
	res[resType].id=resource->id;
      }
      if(isConfigured) {
	printk(KERN_INFO "%s port=0x%x, irq=%d\n",
	       hardwareInfo->name,
	       res[SamHardwareResPort].value,
	       res[SamHardwareResIrq].value);
	if(samRegisterCard(pciInfo->hardwareType, res)<0)
	  printk(KERN_ERR "sam9407: couldn't register card %s\n", hardwareInfo->name);
      } else
	printk(KERN_INFO "sam9407: board %s is not PCI configured\n", hardwareInfo->name);
    }
  }
}
#endif

void samFindCards(void)
{
  findPNPCards();
#ifdef CONFIG_PCI
  findPCICards();
#endif
}
