#ifndef _DRIVER_H
#define _DRIVER_H

#include <asm/atomic.h>
#include <linux/sched.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE>=0x20200
#include <linux/poll.h>
#endif

#include "firmware.h"
#include "vendor.h"

#define SAM_MODE_READ_MASK		0x01
#define SAM_MODE_WRITE_MASK		0x02

#define SAM_IO_TIMEOUT			(2*HZ)

#define SAM_MAX_CARDS			4
#define SAM_STORED_MIDI_EVENTS		256
#define SAM_SEQ_EVENT_SIZE		8

#define SAM_NUMBER_OF(vec)		(sizeof(vec)/sizeof(*(vec)))
#define SAM_OFFSET_OF(type,attr)	((char *)(&((type *)0)->attr)-(char *)0)

#define SAM_CARD_NR(inode)		(MINOR((inode)->i_rdev)>>6)
#define SAM_CARD(inode)			(samCards[SAM_CARD_NR(inode)])
#define SAM_MINOR_DEVICE(inode)		((SoundMinorDevice)(MINOR((inode)->i_rdev)&0xF))
#define SAM_PORT_IDX(inode)		((MINOR((inode)->i_rdev)>>4)&3)

#if LINUX_VERSION_CODE>=0x20200
#define SAM_GET_USER(ptr)		(*(ptr))
#define SAM_PUT_USER(v, ptr)		(*(ptr)=(v))
#define SAM_MEMCPY_FROMFS(d, s, n)	memcpy(d, s, n)
#define SAM_MEMCPY_TOFS(d, s, n)	memcpy(d, s, n)
#define SAM_ATOMIC_READ(v)		atomic_read(v)
#define SAM_ATOMIC_SET(v,i)		atomic_set(v,i)
#else
#define SAM_GET_USER(ptr)		get_user(ptr)
#define SAM_PUT_USER(v, ptr)		put_user(v, ptr)
#define SAM_MEMCPY_FROMFS(d, s, n)	memcpy_fromfs(d, s, n)
#define SAM_MEMCPY_TOFS(d, s, n)	memcpy_tofs(d, s, n)
#define SAM_ATOMIC_READ(v)		(*(v))
#define SAM_ATOMIC_SET(v,i)		(*(v)=i)
#endif

typedef enum {
  SoundMinorCtl,
  SoundMinorSeq,
  SoundMinorMidi,
  SoundMinorDsp,
  SoundMinorAudio,
  SoundMinorDsp16,
  SoundMinorStatus,
  SoundMinorAwfm,
  SoundMinorSeq2,
  SoundMinorSndproc,
  SoundMinorPss=SoundMinorSndproc
} SoundMinorDevice;

extern int samMajor;

/* audio feed thread (aka sam9407d) */

typedef struct {
  struct wait_queue *feedSam;
  int shutdown;
} SamAudioFeedThreadInfo;

extern SamAudioFeedThreadInfo samFeedThreadInfo;

/* I/O */

typedef struct {
  struct semaphore ioLock, stillTryingLock;
  struct wait_queue *somethingHappened;
  struct wait_queue *pickLockScreaming;
  struct wait_queue *pickLockHappy;
  int ioTimeout;		/* timeout for I/O operations */
  int ioExpired;		/* ioTimeout exceeded */
  int pickLock;			/* samAudioFeedThread wants to talk */
  int waitForUARTAck;
} SamIOInfo;

/* MMT */

typedef struct {
  struct semaphore accessLock;
  unsigned mmtSize;
  SamMMTEntry *mmt, *tmpMMT;
  SamMMTSBank *sbank, *tmpSBank;
} SamMemInfo;

/* /dev/mixer */

typedef struct {
  struct semaphore accessLock;
  struct wait_queue *waitDirty;

  /* each process that opens /dev/mixer
   * gets an index into usedMask
   * which is used to mark controls that
   * have been modified in idDirty
   *
   * firmwareDirty, sbankDirty will be triggered, whenever
   * changes have been made to the firmware or soundbank
   * they will be returned as values <0 in samReadMixer
   */
  unsigned usedMask;
  unsigned idDirty[SAM_MIXER_ID_LAST];
  unsigned firmwareDirty, sbankDirty;
} SamMixerInfo;

/* /dev/dsp & /dev/audio */

/* As all channel operations are handled
 * asynchronously, we keep some states here
 */
typedef enum {
  SamChannelStateClosed,
  SamChannelStateNotStarted,
  SamChannelStateSendStart,
  SamChannelStateOpen,
  SamChannelStateUnderflow,
  SamChannelStateOverflow,
  SamChannelStateClosing,
} SamChannelState;

typedef struct {
  char used, first;
  unsigned long startJiffies;
  unsigned bytes;
} SamAudioFrameCounter;

typedef struct {
  int frameIdx, transitions; /* points to the frame, we filled last */
  SamAudioFrameCounter frames[2];
  unsigned finishedBytes;
} SamAudioCounter;

typedef struct {
  struct semaphore accessLock;
  char inUse, justOpened;
  SamChannelState state;
  unsigned format;
  unsigned short rate;
  char stereo;
  short *data;
  unsigned alloced, needed, count;
  SamVolume volume;
  SamAudioCounter counter;
  struct wait_queue *feedMe, *waitClosed;
} SamChannelInfo;

/* /dev/midi  */
typedef enum {
  SamMidiPortInternal,
  SamMidiPortExternal,
  SamMidiPortSentinel /* == last_port + 1 */
} SamMidiPortType;

typedef enum {
  SamMidiPortOwnerNone,
  SamMidiPortOwnerMidi,
  SamMidiPortOwnerSeq
} SamMidiPortOwner;

typedef struct {
  unsigned char ringBuffer[SAM_STORED_MIDI_EVENTS];
  int usedIdx, usedCount;
} SamMidiBuffer;

typedef struct {
  SamMidiPortOwner owner;
} SamMidiPort;

typedef struct {
  struct semaphore accessLock;
  SamMidiPort ports[SamMidiPortSentinel]; /* indexed by SamMidiPortType */
  struct wait_queue *dataAvail;
  int recording, recOverflow;
  SamMidiBuffer rec;
} SamMidiInfo;

/* /dev/sequencer */

typedef struct {
  unsigned char ringBuffer[SAM_STORED_MIDI_EVENTS*SAM_SEQ_EVENT_SIZE];
  int usedIdx, usedCount;
  int timerRunning;
  atomic_t timerTicks;
} SamSeqBuffer;

typedef struct {
  SamSeqBuffer play, rec;
  unsigned timeBase, tempo;
  struct wait_queue *moreSpace, *dataAvail;
  int recording, recOverflow, followupRecEvent, lastRecTimer;
  int sleeping;
  atomic_t wakeupTicks, lastWaitTimerTicks;
} SamSeqInfo;

/* /dev/sndstat */

typedef struct {
  int n, i;
  char s[0];
} SamSndStatMsg;

typedef struct {
  struct semaphore accessLock;
} SamSndStatInfo;

/* Soundcard */
typedef struct SamCard {
  int port, irq, auxPort0, auxPort1;
  SamHardwareType hardwareType;
  SamHardwareInfo *hardwareInfo;
  SamFirmwareType firmwareType;
  SamFirmwareInfo *firmwareInfo;

  /* keep usageCountLock down while
   * checking for usageCount==0
   * in vsamTalkScheduled, when uploading
   * new firmware, or whenever
   * incrementing usageCount
   */
  struct semaphore usageCountLock;
  int firmwareLoaded;
  atomic_t usageCount;

  SamIOInfo io;
  SamMemInfo mem;
  SamMixerInfo mixer;
  SamChannelInfo *channels;
  SamMidiInfo midi;
  SamSeqInfo seq;
  SamSndStatInfo sndStat;

  union {
    SamEWS64CtrlValues ews64;
  } ctrlValues;
} SamCard;

extern int samCardsNr;
extern SamCard *samCards[SAM_MAX_CARDS];

int samRegisterCard(SamHardwareType hardwareType, SamHardwareResource *res);

#endif /* _DRIVER_H */
