/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/module.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <asm/io.h>

#include "driver.h"
#include "cards.h"
#include "vendor.h"
#include "io.h"
#include "alloc.h"
#include "audio.h"
#include "mixer.h"
#include "midi.h"
#include "seq.h"
#include "sndstat.h"

typedef struct {
  char needsFirmware, separateReadWrite;
  int (*open) (struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure);
  void (*release) (struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure);
  int (*read) (struct inode *inode, struct file *file, char *buf, int count, SamCard *card, char *closure);
  int (*write) (struct inode *inode, struct file *file, const char *buf, int count, SamCard *card, char *closure);
#if LINUX_VERSION_CODE>=0x20200
  unsigned (*poll)(struct inode *inode, struct file *file, poll_table *wait, SamCard *card, mode_t mode, char *closure);
#else /* Linux 2.0 */
  int (*select) (struct inode *inode, struct file *file, int, select_table *wait, SamCard *card, char *closure);
#endif /* Linux 2.0 */
  int (*ioctl) (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg, SamCard *card, mode_t mode, char *closure, int callNr);
} SamMinorOps;

/* SAM file operations */
#if LINUX_VERSION_CODE>=0x20200
int samOpen(struct inode *inode, struct file *file);
int samRelease(struct inode *inode, struct file *file);
ssize_t samRead(struct file *file, char *buf, size_t count, loff_t *pos);
ssize_t samWrite(struct file *file, const char *buf, size_t count, loff_t *pos);
unsigned samPoll(struct file *file, poll_table *wait);
int samIoctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
#else /* Linux 2.0 */
int samOpen(struct inode *inode, struct file *file);
void samRelease(struct inode *inode, struct file *file);
int samRead(struct inode *inode, struct file *file, char *buf, int count);
int samWrite(struct inode *inode, struct file *file, const char *buf, int count);
int samSelect(struct inode *inode, struct file *file, int type, select_table *wait);
int samIoctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
#endif /* Linux 2.0 */

int samMajor=SOUND_MAJOR;
int samCardsNr;
SamCard *samCards[SAM_MAX_CARDS];

#if LINUX_VERSION_CODE>=0x20200
struct file_operations samFileOps={
  NULL,				/* lseek */
  samRead,			/* read */
  samWrite,			/* write */
  NULL,				/* readdir */
  samPoll,			/* select */
  samIoctl,			/* ioctl */
  NULL,				/* mmap */
  samOpen,			/* open */
  NULL,				/* flush */
  samRelease,			/* release */
  NULL,				/* fsync */
  NULL,				/* fasync */
  NULL,				/* check_media_change */
  NULL,				/* revalidate */
  NULL				/* lock */
};
#else /* Linux 2.0 */
struct file_operations samFileOps={
  NULL,				/* lseek */
  samRead,			/* read */
  samWrite,			/* write */
  NULL,				/* readdir */
  samSelect,			/* select */
  samIoctl,			/* ioctl */
  NULL,				/* mmap */
  samOpen,			/* open */
  samRelease,			/* release */
  NULL,				/* fsync */
  NULL,				/* fasync */
  NULL,				/* check_media_change */
  NULL				/* revalidate */
};
#endif /* Linux 2.0 */

SamMinorOps samMinorOps[]={
  {				/* SoundMinorCtl */
    0,				/* needsFirmware */
    0,				/* separateReadWrite */
    samOpenMixer,		/* open */
    samReleaseMixer,		/* release */
    samReadMixer,		/* read */
    NULL,			/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollMixer,		/* poll */
#else /* Linux 2.0 */
    samSelectMixer,		/* select */
#endif /* Linux 2.0 */
    samIoctlMixer		/* ioctl */
  },
  {				/* SoundMinorSeq */
    1,				/* needsFirmware */
    0,				/* separateReadWrite */
    samOpenSeq,			/* open */
    samReleaseSeq,		/* release */
    samReadSeq,			/* read */
    samWriteSeq,		/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollSeq,			/* poll */
#else /* Linux 2.0 */
    samSelectSeq,		/* select */
#endif /* Linux 2.0 */
    samIoctlSeq			/* ioctl */
  },
  {				/* SoundMinorMidi */
    1,				/* needsFirmware */
    0,				/* separateReadWrite */
    samOpenMidi,		/* open */
    samReleaseMidi,		/* release */
    samReadMidi,		/* read */
    samWriteMidi,		/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollMidi,		/* poll */
#else /* Linux 2.0 */
    samSelectMidi,		/* select */
#endif /* Linux 2.0 */
    samIoctlMidi		/* ioctl */
  },
  {				/* SoundMinorDsp */
    1,				/* needsFirmware */
    1,				/* separateReadWrite */
    samOpenAudio,		/* open */
    samReleaseAudio,		/* release */
    samReadAudio,		/* read */
    samWriteAudio,		/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollAudio,		/* poll */
#else /* Linux 2.0 */
    samSelectAudio,		/* select */
#endif /* Linux 2.0 */
    samIoctlAudio		/* ioctl */
  },
  {				/* SoundMinorAudio */
    1,				/* needsFirmware */
    1,				/* separateReadWrite */
    samOpenAudio,		/* open */
    samReleaseAudio,		/* release */
    samReadAudio,		/* read */
    samWriteAudio,		/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollAudio,		/* poll */
#else /* Linux 2.0 */
    samSelectAudio,		/* select */
#endif /* Linux 2.0 */
    samIoctlAudio		/* ioctl */
  },
  {				/* SoundMinorDsp16 */
    1,				/* needsFirmware */
    1,				/* separateReadWrite */
    samOpenAudio,		/* open */
    samReleaseAudio,		/* release */
    samReadAudio,		/* read */
    samWriteAudio,		/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollAudio,		/* poll */
#else /* Linux 2.0 */
    samSelectAudio,		/* select */
#endif /* Linux 2.0 */
    samIoctlAudio		/* ioctl */
  },
  {				/* SoundMinorStatus */
    0,				/* needsFirmware */
    0,				/* separateReadWrite */
    samOpenSndStat,		/* open */
    samReleaseSndStat,		/* release */
    samReadSndStat,		/* read */
    NULL,			/* write */
#if LINUX_VERSION_CODE>=0x20200
    samPollSndStat,		/* poll */
#else /* Linux 2.0 */
    samSelectSndStat,		/* select */
#endif /* Linux 2.0 */
    NULL			/* ioctl */
  }
};

/* SAM file operations */
int samOpen(struct inode *inode, struct file *file)
{
  int cardNr=SAM_CARD_NR(inode);
  SamCard *card;
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  int error;
  char *closure;

  if(cardNr>=samCardsNr || minor<0 || minor>=SAM_NUMBER_OF(samMinorOps) ||
     !samMinorOps[minor].open)
    return -ENODEV;

  card=samCards[cardNr];
  if(samMinorOps[minor].needsFirmware) {
    down(&card->usageCountLock);
    if(!card->firmwareLoaded) {
      up(&card->usageCountLock);
      return -ENOPKG;
    }
    atomic_inc(&card->usageCount);
    up(&card->usageCountLock);
  }

  if(samMinorOps[minor].separateReadWrite) {
    closure=(char *)&file->private_data;
    if(file->f_mode & SAM_MODE_READ_MASK)
      if((error=(*samMinorOps[minor].open)(inode, file, card, SAM_MODE_READ_MASK, closure))<0)
	return error;
    if(file->f_mode & SAM_MODE_WRITE_MASK)
      if((error=(*samMinorOps[minor].open)(inode, file, card, SAM_MODE_WRITE_MASK, closure+1))<0) {
	if(samMinorOps[minor].release)
	  (*samMinorOps[minor].release)(inode, file, card, SAM_MODE_READ_MASK, closure);
	return error;
      }
  } else
    if((error=(*samMinorOps[minor].open)(inode, file, card, 0, NULL))<0)
      return error;

  MOD_INC_USE_COUNT;
  return 0;
}

#if LINUX_VERSION_CODE>=0x20200
int samRelease(struct inode *inode, struct file *file)
#else /* Linux 2.0 */
void samRelease(struct inode *inode, struct file *file)
#endif /* Linux 2.0 */
{
  SamCard *card=SAM_CARD(inode);
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  char *closure;

  if(samMinorOps[minor].release) {
    if(samMinorOps[minor].separateReadWrite) {
      closure=(char *)&file->private_data;
      if(file->f_mode & SAM_MODE_READ_MASK)
	(*samMinorOps[minor].release)(inode, file, card, SAM_MODE_READ_MASK, closure);
      if(file->f_mode & SAM_MODE_WRITE_MASK)
	(*samMinorOps[minor].release)(inode, file, card, SAM_MODE_WRITE_MASK, closure+1);
    } else
      (*samMinorOps[minor].release)(inode, file, card, 0, NULL);
  }
  if(samMinorOps[minor].needsFirmware)
    atomic_dec(&card->usageCount);
  MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE>=0x20200
  return 0;
#endif /* Linux 2.0 */
}

#if LINUX_VERSION_CODE>=0x20200
ssize_t samRead(struct file *file, char *buf, size_t count, loff_t *pos)
#else /* Linux 2.0 */
int samRead(struct inode *inode, struct file *file, char *buf, int count)
#endif /* Linux 2.0 */
{
#if LINUX_VERSION_CODE>=0x20200
  struct inode *inode=file->f_dentry->d_inode;
#endif /* Linux 2.0 */
  SamCard *card=SAM_CARD(inode);
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  char *closure;

  if(samMinorOps[minor].read) {
    if(samMinorOps[minor].separateReadWrite)
      closure=(char *)&file->private_data;
    else
      closure=NULL;
    return (*samMinorOps[minor].read)(inode, file, buf, count, card, closure);
  } else
    return -ENOSYS;
}

#if LINUX_VERSION_CODE>=0x20200
ssize_t samWrite(struct file *file, const char *buf, size_t count, loff_t *pos)
#else /* Linux 2.0 */
int samWrite(struct inode *inode, struct file *file, const char *buf, int count)
#endif /* Linux 2.0 */
{
#if LINUX_VERSION_CODE>=0x20200
  struct inode *inode=file->f_dentry->d_inode;
#endif /* Linux 2.0 */
  SamCard *card=SAM_CARD(inode);
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  char *closure;

  if(samMinorOps[minor].write) {
    if(samMinorOps[minor].separateReadWrite)
      closure=(char *)&file->private_data+1;
    else
      closure=NULL;
    return (*samMinorOps[minor].write)(inode, file, buf, count, card, closure);
  } else
    return -ENOSYS;
}

#if LINUX_VERSION_CODE>=0x20200
unsigned samPoll(struct file *file, poll_table *wait)
{
  struct inode *inode=file->f_dentry->d_inode;
  SamCard *card=SAM_CARD(inode);
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  char *closure;
  unsigned mask;

  if(samMinorOps[minor].poll) {
    if(samMinorOps[minor].separateReadWrite) {
      closure=(char *)&file->private_data;
      mask=0;
      if(file->f_mode & SAM_MODE_READ_MASK)
	mask|=(*samMinorOps[minor].poll)(inode, file, wait, card, SAM_MODE_READ_MASK, closure);
      if(file->f_mode & SAM_MODE_WRITE_MASK)
	mask|=(*samMinorOps[minor].poll)(inode, file, wait, card, SAM_MODE_READ_MASK, closure+1);
    } else
      mask=(*samMinorOps[minor].poll)(inode, file, wait, card, 0, NULL);
    return mask;
  } else
    return 0;
}
#else /* Linux 2.0 */
int samSelect(struct inode *inode, struct file *file, int type, select_table *wait)
{
  SamCard *card=SAM_CARD(inode);
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  char *closure;
  if(samMinorOps[minor].select) {
    if(samMinorOps[minor].separateReadWrite) {
      closure=(char *)&file->private_data;
      if(type==SEL_OUT)
	closure++;
    } else
      closure=NULL;
    return (*samMinorOps[minor].select)(inode, file, type, wait, card, closure);
  } else
    return -ENOSYS;
}
#endif /* Linux 2.0 */

int samIoctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
  SamCard *card=SAM_CARD(inode);
  SoundMinorDevice minor=SAM_MINOR_DEVICE(inode);
  int result, callNr;
  char *closure;
  if(samMinorOps[minor].ioctl) {
    closure=(char *)&file->private_data;
    result=0;
    if(samMinorOps[minor].separateReadWrite) {
      callNr=0;
      if(file->f_mode & SAM_MODE_READ_MASK)
	if((result=(*samMinorOps[minor].ioctl)(inode, file, cmd, arg, card, SAM_MODE_READ_MASK, closure, callNr++))<0)
	  return result;
      if(file->f_mode & SAM_MODE_WRITE_MASK)
	if((result=(*samMinorOps[minor].ioctl)(inode, file, cmd, arg, card, SAM_MODE_WRITE_MASK, closure+1, callNr++))<0)
	  return result;
    } else
      result=(*samMinorOps[minor].ioctl)(inode, file, cmd, arg, card, 0, NULL, 0);
    return result;
  } else
    return -ENOSYS;
}

/* Initialization */

int samRegisterCard(SamHardwareType hardwareType, SamHardwareResource *res)
{
  int error, len, i;
  SamCard *card;
  char buf[32], *cp;

  if(samCardsNr>=SAM_MAX_CARDS) {
    printk(KERN_ERR "sam9407: too many cards\n");
    return -EIO;
  }
    
  if(check_region(res[SamHardwareResPort].value, 4)<0) {
    printk(KERN_ERR "sam9407: I/O port 0x%x already in use\n", res[SamHardwareResPort].value);
    return -EIO;
  }
  for(i=SamHardwareResAuxPort0; i<=SamHardwareResAuxPort1; i++) {
    if(res[i].value && check_region(res[i].value, 1)<0) {
      printk(KERN_ERR "sam9407: I/O port 0x%x already in use\n", res[i].value);
      return -EIO;
    }
  }

  request_region(res[SamHardwareResPort].value, 4, res[SamHardwareResPort].id);
  for(i=SamHardwareResAuxPort0; i<=SamHardwareResAuxPort1; i++) {
    if(res[i].value)
      request_region(res[i].value, 1, res[i].id);
  }

  card=(SamCard *)kmalloc(sizeof(SamCard), GFP_KERNEL);
  if(!card) {
    printk(KERN_ERR "sam9407: out of memory\n");
    return -ENOMEM;
  }
  memset(card, 0, sizeof(SamCard));

  card->port=res[SamHardwareResPort].value;
  card->auxPort0=res[SamHardwareResAuxPort0].value;
  card->auxPort1=res[SamHardwareResAuxPort1].value;
  card->hardwareType=hardwareType;
  card->hardwareInfo=samHardwareInfos+hardwareType;
  card->firmwareType=SamFirmwareTypeGenuine;
  card->firmwareInfo=samFirmwareInfos+SamFirmwareTypeGenuine;

  card->usageCountLock=MUTEX;
  card->io.ioLock=MUTEX;
  card->io.stillTryingLock=MUTEX;
  card->io.ioTimeout=SAM_IO_TIMEOUT;
  card->mem.accessLock=MUTEX;
  card->mixer.accessLock=MUTEX;
  card->midi.accessLock=MUTEX;
  card->sndStat.accessLock=MUTEX;

  samCards[samCardsNr]=card;
  samCardsNr++;

  /* try to detect firmware */
  if(samFirmwareDetect(card)) {
    error=0;
	
    for(i=0; i<SamFirmwareTypeSentinel; i++) {
      if(samFirmwareInfos[i].probeStr) {
		len=strlen(samFirmwareInfos[i].probeStr);
		if(len<sizeof(buf)-2) {
		  if((error=samTalk(card,
			        SamReadMem, samFirmwareInfos[i].probeAddr>>1, buf, (len+1)>>1,
					SamEndOfScript))<0)
	  		break;
		  cp=buf+(samFirmwareInfos[i].probeAddr&1);
		  cp[len]=0;
  
		  if(strcmp(cp, samFirmwareInfos[i].probeStr)==0) {
	  	    card->firmwareType=i;
	  	    card->firmwareInfo=samFirmwareInfos+i;
	  	    break;
		  }
		}
      }
    }
    if(error>=0)
      error=samInitializeMMT(card);
    if(error>=0)
      error=samAllocateChannels(card);
    if(error<0) {
      /* firmware seems to be broken
       */
      card->firmwareType=SamFirmwareTypeGenuine;
      card->firmwareInfo=samFirmwareInfos+SamFirmwareTypeGenuine;
      samReleaseChannels(card);
      samReleaseMMT(card);
    }

    if(error>=0)
      card->firmwareLoaded=1;
  }

  if(res[SamHardwareResIrq].value)
    if(request_irq(res[SamHardwareResIrq].value, samInterrupt, SA_SHIRQ, res[SamHardwareResIrq].id, card)<0)
      printk(KERN_WARNING "sam9407: Unable to register IRQ %d\n", res[SamHardwareResIrq].value);
    else
      card->irq=res[SamHardwareResIrq].value;

  return 0;
}

int sam9407_init(void)
{
  if(!samMajor) {
    printk(KERN_ERR "sam9407: samMajor not specified.\n");
    return -EIO;
  }

  if(register_chrdev(samMajor, "sam9407", &samFileOps)<0) {
    printk(KERN_ERR "sam9407: can't register sound dev, sound module already loaded ?\n");
    return -EIO;
  }

  if(samStartFeedThread()<0) {
    if(unregister_chrdev(samMajor, "sam9407")<0)
      panic("sam9407: Unable to unregister device\n");
    return -EIO;
  }

  samStartPollNotifications();

  samFindCards();

  return 0;
}

#ifdef MODULE
#if LINUX_VERSION_CODE>=0x20200
MODULE_PARM(samMajor,"i");
#endif

int init_module(void)
{
  return sam9407_init();
}

void cleanup_module(void)
{
  SamCard *card;
  int i;

  if(unregister_chrdev(samMajor, "sam9407")<0)
    panic("sam9407: Unable to unregister device\n");

  /* disable IRQ's first
   * to avoid bad surprises
   */
  for(i=0; i<samCardsNr; i++) {
    card=samCards[i];
    if(card->irq) {
      free_irq(card->irq, card);
      card->irq=0;
    }
  }

  samStopPollNotifications();
  samKillFeedThread();

  for(i=0; i<samCardsNr; i++) {
    card=samCards[i];

    card->firmwareLoaded=0;
    samReleaseChannels(card);
    samReleaseMMT(card);

    release_region(card->port, 4);
    if(card->auxPort0)
      release_region(card->auxPort0, 1);
    if(card->auxPort1)
      release_region(card->auxPort1, 1);

    samCards[i]=NULL;
    kfree(card);
  }
}
#endif
