/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdarg.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <asm/io.h>

#include "driver.h"
#include "audio.h"
#include "midi.h"
#include "io.h"
#include "vendor.h"

#define SAM_CONTROL		(card->port+1)
#define SAM_STATUS		(card->port+1)
#define SAM_DATA8		(card->port)
#define SAM_DATA16		(card->port+2)
#define SAM_DAAD		(card->port+7)

#define SAM_IO_DETECT_TIMEOUT	HZ/10
#define SAM_POLL_INTERVAL_IRQ	2*HZ
#define SAM_POLL_INTERVAL_POLL	HZ/20
#define SAM_POLL_INTERVAL_MIDI	1
#define SAM_EAGAIN_TIMEOUT	1
#define SAM_MAX_ASYNCS_INAROW	32
#define SAM_MAX_TALK_RETRIES	16
#define SAM_MAX_DISCARD_BYTES	0x1000

typedef struct {
  struct timer_list pollTimer;
  int pollAsync;		/* poll for asynchronous notifications */
} IOPollInfo;

static void pollNotification(unsigned long data);
static int samTalkNoLock(SamCard *card, SamTalkCmd cmd, ...);

static IOPollInfo pollInfo={
  {
    NULL, NULL, 0, 0, pollNotification
  }
};

#if LINUX_VERSION_CODE<0x20200
#ifdef __i386__
extern inline void outswFS(unsigned short port, unsigned short *sp, unsigned count)
{
  __asm__ __volatile__
    ("push %%ds\n\t"
     "push %%fs\n\t"
     "pop %%ds\n\t"
     "cld\n\t"
     "rep\n\t"
     "outsw\n\t"
     "pop %%ds": : "d" (port), "S" (sp), "c" (count));
}
#else
extern inline void outswFS(unsigned short port, unsigned short *sp, unsigned count)
{
  int i;
  for(i=0; i<count; i++)
    outw(SAM_GET_USER(sp+i), port);
}
#endif
#endif /* !Linux 2.0 */

static void checkAsync(SamCard *card)
{
  unsigned flags;
  int loops, status, v;

  for(loops=0;
      (inb(SAM_STATUS)&(SAM_STATUS_TE|SAM_STATUS_ID_SYNC))==0 &&
	loops<SAM_MAX_ASYNCS_INAROW;
      loops++) {
    /* received a notification outside a samTalk script
     * check if condition is still valid after disabling IRQ's
     */
    save_flags(flags); cli();
    if(((status=inb(SAM_STATUS))&(SAM_STATUS_TE|SAM_STATUS_ID_SYNC))==0) {
      v=inb(SAM_DATA8);
      restore_flags(flags);
      switch(status&SAM_STATUS_ID_MASK) {
      case SAM_STATUS_ID_AUDIO:
	samAudioNotification(card, v);
	break;
      case SAM_STATUS_ID_MIDI:
	if(v==SAM_ACK_UART_MOD) {
	  card->io.waitForUARTAck=0;
	  wake_up(&card->io.somethingHappened);
	} else
	  samMidiInput(card, v);
	break;
      }
      restore_flags(flags);
    }
  }
}

void samInterrupt(int irq, void *devId, struct pt_regs *regs)
{
  SamCard *card=(SamCard *)devId;

  checkAsync(card);
  wake_up(&card->io.somethingHappened);
}

static void ioTimedOut(unsigned long data)
{
  SamCard *card=(SamCard *)data;
  card->io.ioExpired=1;
  wake_up(&card->io.somethingHappened);
}

static void pollNotification(unsigned long data)
{
  unsigned flags;
  int needsPolling, midiInput, i;

  needsPolling=midiInput=0;
  for(i=0; i<samCardsNr; i++) {
    if(!samCards[i]->irq)
      needsPolling=1;
    if(samCards[i]->midi.ports[SamMidiPortExternal].owner!=SamMidiPortOwnerNone)
      midiInput=1;
    checkAsync(samCards[i]);
  }

  save_flags(flags); cli();
  if(pollInfo.pollAsync) {
    if(needsPolling) {
      /* If midi input is attached, we need to poll even more
       * often, as the usual 20Hz are not accurate enough
       */
      if(midiInput)
	pollInfo.pollTimer.expires=jiffies + SAM_POLL_INTERVAL_MIDI;
      else
	pollInfo.pollTimer.expires=jiffies + SAM_POLL_INTERVAL_POLL;
    } else
      pollInfo.pollTimer.expires=jiffies + SAM_POLL_INTERVAL_IRQ;
    add_timer(&pollInfo.pollTimer);
  }
  restore_flags(flags);
}

void samStartPollNotifications(void)
{
  pollInfo.pollAsync=1;
  pollInfo.pollTimer.expires=jiffies+SAM_POLL_INTERVAL_POLL;
  add_timer(&pollInfo.pollTimer);
}

void samStopPollNotifications(void)
{
  pollInfo.pollAsync=0;
  del_timer(&pollInfo.pollTimer);
}

static int samWaitReady(SamCard *card, int forcePolling, int (*condition)(SamCard *))
{
  unsigned flags;
  struct timer_list ioTimer;

  card->io.ioExpired=0;
  init_timer(&ioTimer);
  ioTimer.expires=jiffies+card->io.ioTimeout;
  ioTimer.data=(unsigned long)card;
  ioTimer.function=ioTimedOut;
  add_timer(&ioTimer);

  while((*condition)(card)!=0) {
    if(!forcePolling && card->irq) {
      /* using IRQ handler */
      save_flags(flags); cli();
      if((*condition)(card)!=0)
	sleep_on(&card->io.somethingHappened);
      restore_flags(flags);
    } else {
      /* polling mode */
      checkAsync(card);
#if LINUX_VERSION_CODE>=0x20200
      if(current->need_resched)
	schedule();
#else /* Linux 2.0 */
      if(need_resched)
	schedule();
#endif /* Linux 2.0 */
    }
    if(card->io.ioExpired) {
      del_timer(&ioTimer);
      if(!forcePolling && card->irq) {
	printk(KERN_WARNING "sam9407: IRQ %d not working properly, using poll mode.\n", card->irq);
	free_irq(card->irq, card);
	card->irq=0;
	/* reschedule timer for polling mode */
	card->io.ioExpired=0;
	ioTimer.expires=jiffies+card->io.ioTimeout;
	add_timer(&ioTimer);
      } else
	/* not much to do, if we're already in polling mode */
	return -EIO;
    }
  }
  del_timer(&ioTimer);
  return 0;
}

static int samWritableCondition(SamCard *card)
{
  return inb(SAM_STATUS)&SAM_STATUS_RF;
}

extern inline int samOutb(SamCard *card, int value, int port)
{
  int error;
  if(inb(SAM_STATUS)&SAM_STATUS_RF)
    if((error=samWaitReady(card, 1, samWritableCondition))<0)
      return error;
  outb(value, port);
  return 0;
}

//+ kjk
void samSetDaad(SamCard *card, unsigned short value )
{
  if( card->hardwareType == SamHardwareTypeST128 ) {
	  outb(value, SAM_DAAD);
  } else if ( card->hardwareType == SamHardwareTypeST128DDMA ) {
	  outb(1,     card->port + 0xb );
	  outb(value, card->port + 0xa );
  }
}

static int samReadableCondition(SamCard *card)
{
  return (inb(SAM_STATUS)&(SAM_STATUS_TE|SAM_STATUS_ID_SYNC))!=SAM_STATUS_ID_SYNC;
}

extern inline int samInb(SamCard *card)
{
  unsigned flags;
  int error, v;

  save_flags(flags); cli();
  while((inb(SAM_STATUS)&(SAM_STATUS_TE|SAM_STATUS_ID_SYNC))!=SAM_STATUS_ID_SYNC) {
    restore_flags(flags);
    if((error=samWaitReady(card, !card->firmwareLoaded, samReadableCondition))<0)
      return error;
    cli();
  }
  v=inb(SAM_DATA8);
  restore_flags(flags);
  return v;
}

static void discardInput(SamCard *card)
{
  unsigned flags;
  int n;

  save_flags(flags); cli();
  for(n=0; n<SAM_MAX_DISCARD_BYTES && (inb(SAM_STATUS)&(SAM_STATUS_TE|SAM_STATUS_ID_SYNC))==SAM_STATUS_ID_SYNC; n++)
    inb(SAM_DATA8);
  restore_flags(flags);
}

static int samUARTAckCondition(SamCard *card)
{
  return card->io.waitForUARTAck;
}

static int switchUARTMode(SamCard *card)
{
  int error;
  
  card->io.waitForUARTAck=1;
  if((error=samOutb(card, SAM_CMD_UART_MOD, SAM_CONTROL))<0)
    return error;

  while(card->io.waitForUARTAck) {
    if((error=samWaitReady(card, !card->firmwareLoaded, samUARTAckCondition))<0)
      return error;
  }
  return 0;
}

static int memOp(SamCard *card, int ctrl, SamTalkCmd burstOp, unsigned aAddr, unsigned short *aSp, int words)
{
  int result, chunk, max, i;
  unsigned addr;
  unsigned short *sp;

  addr=aAddr;
  sp=aSp;
  result=0;

  for(i=0; result>=0 && i<words; i+=chunk) {
    chunk=words-i<SAM_MAX_TRANSFER_SIZE ? words-i : SAM_MAX_TRANSFER_SIZE;
    /* if chunk would exceed page boundary, limit it */
    max=((~addr)&0xFFFF)+1;
    if(chunk>max)
      chunk=max;
    
    result=samTalkNoLock(card,
			 SamSendCtrl, 7, ctrl,
			 addr, addr>>8, addr>>16, addr>>24,
			 chunk, chunk>>8,
			 SamExpectRedo, SAM_ACK_OP, SAM_NAK_OP,
			 burstOp, sp, chunk,
			 SamEndOfScript);
    addr+=chunk;
    sp+=chunk;
  }

  return result;
}

static int vsamTalk(SamCard *card, SamTalkCmd aCmd, va_list ap, va_list *restart)
{
  SamTalkCmd cmd;
  int done, result, n, i, cond;
  unsigned addr, *addrP, v, v2;
  unsigned short *sp;
  unsigned char *cp;

  cmd=aCmd;
  done=result=0;

  while(!done && result>=0) {
    switch(cmd) {
    case SamDiscardInput:
      /* if we're being picklocked by audio streaming
       * this is a good place to restart
       */
      if(restart)
	*restart=ap;

      udelay(1000);
      discardInput(card);
      break;
    case SamUARTMode:
      if(restart)
	*restart=ap;

      result=switchUARTMode(card);
      break;
    case SamSendCtrlInit:
      if(restart)
	*restart=ap;

      n=va_arg(ap, int);
      for(i=0; i<n && result>=0; i++) {
	udelay(200);
	result=samOutb(card, va_arg(ap, unsigned char), SAM_CONTROL);
      }
      break;
    case SamSendCtrl:
      if(restart)
	*restart=ap;

      n=va_arg(ap, int);
//      for(i=0; i<n && result>=0; i++)
      for(i=0; i<n; i++)
//	result=samOutb(card, va_arg(ap, unsigned char), i>0 ? SAM_DATA8 : SAM_CONTROL);
	result+=samOutb(card, va_arg(ap, unsigned char), i>0 ? SAM_DATA8 : SAM_CONTROL);
      break;
    case SamSendCtrlIf:
      if(restart)
	*restart=ap;

      cond=va_arg(ap, int);
      n=va_arg(ap, int);
      if(cond)
	for(i=0; i<n && result>=0; i++)
	  result=samOutb(card, va_arg(ap, unsigned char), i>0 ? SAM_DATA8 : SAM_CONTROL);
      else
	for(i=0; i<n && result>=0; i++)
	  (void)va_arg(ap, unsigned char);
      break;
    case SamSendCtrlP:
      if(restart)
	*restart=ap;

      n=va_arg(ap, int);
      cp=va_arg(ap, unsigned char *);
      for(i=0; i<n && result>=0; i++)
	result=samOutb(card, cp[i], i>0 ? SAM_DATA8 : SAM_CONTROL);
      break;
    case SamSendMidi:
      cp=va_arg(ap, unsigned char *);
      n=va_arg(ap, int);
      result=samOutb(card, SAM_CMD_EN_MIDOUT, SAM_CONTROL);
      for(i=0; i<n && result>=0; i++)
	result=samOutb(card, cp[i], SAM_DATA8);
      break;
    case SamSendMidiFS:
      cp=va_arg(ap, unsigned char *);
      n=va_arg(ap, int);
      result=samOutb(card, SAM_CMD_EN_MIDOUT, SAM_CONTROL);
      for(i=0; i<n && result>=0; i++)
	result=samOutb(card, SAM_GET_USER(cp+i), SAM_DATA8);
      break;
    case SamExpect:
      if((result=samInb(card))>=0)
	if(result!=(v=va_arg(ap, unsigned char))) {
	  if(card->firmwareLoaded)
	    printk(KERN_ERR "sam9407: invalid response, expected=0x%02x, received=0x%02x\n", v, result);
	  result=-EIO;
	}
      break;
    case SamExpectRedo:
      v=va_arg(ap, unsigned char);
      v2=va_arg(ap, unsigned char);
      if((result=samInb(card))>=0) {
	if(result!=v)
	  if(result!=v2) {
	    if(card->firmwareLoaded)
	      printk(KERN_ERR "sam9407: invalid response, expected=0x%02x or 0x%02x, received=0x%02x\n", v, v2, result);
	    result=-EIO;
	  } else
	    result=-EAGAIN;
      }
      break;
    case SamGetByte:
      if((result=samInb(card))>=0)
	*va_arg(ap, unsigned char *)=result;
      break;
    case SamGetWord:
      v=0;
      for(i=0; i<2 && (result=samInb(card))>=0; i++)
	v|=result<<(i<<3);
      if(result>=0)
	*va_arg(ap, unsigned short *)=v;
      break;
    case SamGetDWord:
      v=0;
      for(i=0; i<4 && (result=samInb(card))>=0; i++)
	v|=result<<(i<<3);
      if(result>=0)
	*va_arg(ap, unsigned *)=v;
      break;
    case SamBurstRead:
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      insw(SAM_DATA16, sp, n);
      break;
    case SamBurstWrite:
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      outsw(SAM_DATA16, sp, n);
      break;
    case SamBurstWriteFS:
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
#if LINUX_VERSION_CODE>=0x20200
      outsw(SAM_DATA16, sp, n);
#else /* Linux 2.0 */
      outswFS(SAM_DATA16, sp, n);
#endif /* Linux 2.0 */
      break;
    case SamBurstWriteConst:
      v=va_arg(ap, unsigned short);
      n=va_arg(ap, int);
      for(i=0; i<n; i++)
	outw(v, SAM_DATA16);
      break;
    case SamReadMem:
      addr=va_arg(ap, unsigned);
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      result=memOp(card, SAM_CMD_RD_MEM, SamBurstRead, addr, sp, n);
      break;
    case SamReadMemP:
      addrP=va_arg(ap, unsigned *);
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      result=memOp(card, SAM_CMD_RD_MEM, SamBurstRead, *addrP, sp, n);
      break;
    case SamWriteMem:
      addr=va_arg(ap, unsigned);
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      result=memOp(card, SAM_CMD_WRT_MEM, SamBurstWrite, addr, sp, n);
      break;
    case SamWriteMemP:
      addrP=va_arg(ap, unsigned *);
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      result=memOp(card, SAM_CMD_WRT_MEM, SamBurstWrite, *addrP, sp, n);
      break;
    case SamWriteMemFS:
      addr=va_arg(ap, unsigned);
      sp=va_arg(ap, unsigned short *);
      n=va_arg(ap, int);
      result=memOp(card, SAM_CMD_WRT_MEM, SamBurstWriteFS, addr, sp, n);
      break;
    case SamGetMMT:
      addrP=va_arg(ap, unsigned *);
      if(card->firmwareInfo->getMMTAck)
	result=samTalkNoLock(card,
			     SamSendCtrl, 2, SAM_CMD_GET_MMT, 0x00,
			     SamExpect, card->firmwareInfo->getMMTAck,
			     SamGetDWord, addrP,
			     SamEndOfScript);
      else
	result=samTalkNoLock(card,
			     SamSendCtrl, 2, SAM_CMD_GET_MMT, 0x00,
			     SamGetDWord, addrP,
			     SamEndOfScript);
      break;
    case SamSetMMT:
      addr=va_arg(ap, unsigned);
      result=samTalkNoLock(card,
			   SamSendCtrl, 5, SAM_CMD_SET_MMT,
			   addr, addr>>8, addr>>16, addr>>24,
			   SamEndOfScript);
      break;
    case SamEndOfScript:
      result=0;
      done=1;
      break;
    default:
      result=-EINVAL;
    }
    if(!done && result>=0)
      cmd=va_arg(ap, SamTalkCmd);
  }

  return result;
}

static int samTalkNoLock(SamCard *card, SamTalkCmd cmd, ...)
{
  va_list ap;
  int result;

  va_start(ap, cmd);
  result=vsamTalk(card, cmd, ap, NULL);
  va_end(ap);

  return result;
}

static void eagainTimedOut(unsigned long data)
{
  wake_up((struct wait_queue **)data);
}

static int vsamTalkScheduled(SamCard *card, int checkBusy, SamTalkCmd cmd, va_list *ap)
{
  unsigned flags;
  int result, tries;
  struct timer_list eagainTimer;

  down(&card->io.stillTryingLock);
  down(&card->io.ioLock);
  tries=0;
  do {
    if(tries>0 && !card->io.pickLock) {
      /* if we've received an -EAGAIN,
       * wait until samTalkPickLock started
       * or SAM_EAGAIN_TIMEOUT timed out
       */
      init_timer(&eagainTimer);
      eagainTimer.expires=jiffies+SAM_EAGAIN_TIMEOUT;
      eagainTimer.data=(unsigned long)&card->io.pickLockScreaming;
      eagainTimer.function=eagainTimedOut;
      save_flags(flags); cli();
      if(!card->io.pickLock) {
	add_timer(&eagainTimer);
	sleep_on(&card->io.pickLockScreaming);
      }
      restore_flags(flags);
    }

    save_flags(flags); cli();
    while(card->io.pickLock) {
      /* samTalkPickLock is trying to get access */
      up(&card->io.ioLock);
      sleep_on(&card->io.pickLockHappy);
      down(&card->io.ioLock);
      cli();
    }
    restore_flags(flags);

    if(checkBusy) {
      down(&card->usageCountLock);
      if(SAM_ATOMIC_READ(&card->usageCount)) {
	up(&card->usageCountLock);
	up(&card->io.ioLock);
	up(&card->io.stillTryingLock);
	return -EBUSY;
      }
    }
    
    result=vsamTalk(card, cmd, *ap, ap);

    if(checkBusy)
      up(&card->usageCountLock);

  } while(result==-EAGAIN && ++tries<SAM_MAX_TALK_RETRIES);
  up(&card->io.ioLock);
  up(&card->io.stillTryingLock);

  if(result==-EAGAIN) {
    if(card->firmwareLoaded)
      printk(KERN_ERR "sam9407: script failed after %d retries.\n", tries);
    result=-EIO;
  }

  return result;
}

/* public functions */
int samFirmwareDetect(SamCard *card)
{
  int result, oldTimeout;
  down(&card->io.ioLock);
  oldTimeout=card->io.ioTimeout;
  card->io.ioTimeout=SAM_IO_DETECT_TIMEOUT;
  result=samTalkNoLock(card,
		       SamDiscardInput,
		       SamSendCtrl, 1, SAM_CMD_RESET,
		       SamUARTMode,
		       SamDiscardInput,
		       SamEndOfScript)>=0;
  card->io.ioTimeout=oldTimeout;
  up(&card->io.ioLock);
  return result;
}

int samTalkPickLock(SamCard *card, SamTalkCmd cmd, ...)
{
  va_list ap;
  int result;

  card->io.pickLock=1;
  wake_up(&card->io.pickLockScreaming);

  down(&card->io.ioLock);
  card->io.pickLock=0;
  wake_up(&card->io.pickLockHappy);

  va_start(ap, cmd);
  result=vsamTalk(card, cmd, ap, NULL);
  va_end(ap);

  up(&card->io.ioLock);

  return result;
}

int samTalk(SamCard *card, SamTalkCmd cmd, ...)
{
  int result;
  va_list ap;

  va_start(ap, cmd);
  result=vsamTalkScheduled(card, 0, cmd, &ap);
  va_end(ap);

  return result;
}

int samTalkNotBusy(SamCard *card, SamTalkCmd cmd, ...)
{
  int result;
  va_list ap;

  va_start(ap, cmd);
  result=vsamTalkScheduled(card, 1, cmd, &ap);
  va_end(ap);

  return result;
}
