/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/soundcard.h>

#include "driver.h"
#include "io.h"
#include "alloc.h"
#include "seq.h"
#include "midi.h"

unsigned char samMidiSelCtrls[]={
  SAM_CMD_MID_PORT0,
  SAM_CMD_MID_PORT1
};

int samResetMidiPort(SamCard *card, SamMidiPortType portType)
{
  char reset=0xFF;
  return samTalk(card,
		 SamSendCtrl, 1, samMidiSelCtrls[portType],
		 SamSendMidi, &reset, 1,
		 SamEndOfScript);
}

/* handle extern midi events */
void samMidiInput(SamCard *card, unsigned v)
{
  unsigned flags;
  int i;

  save_flags(flags); cli();
  switch(card->midi.ports[SamMidiPortExternal].owner) {
  case SamMidiPortOwnerSeq:
    samMidiInputSeq(card, v);
    break;
  case SamMidiPortOwnerMidi:
    if(card->midi.recording) {
      if(card->midi.rec.usedCount<SAM_STORED_MIDI_EVENTS) {
	i=(card->midi.rec.usedIdx+card->midi.rec.usedCount)%SAM_STORED_MIDI_EVENTS;
	card->midi.rec.ringBuffer[i]=v;
	card->midi.rec.usedCount++;
      } else
	card->midi.recOverflow=1;
      wake_up(&card->midi.dataAvail);
    }
    break;
  default:
    /* SamMidiPortOwnerNone */
  }
  restore_flags(flags);
}

static int resetMidi(SamCard *card, SamMidiPortType portType, int recording)
{
  card->midi.recording=0;
  card->midi.rec.usedIdx=card->midi.rec.usedCount=0;
  card->midi.recOverflow=0;

  card->midi.recording=!!recording;

  return samResetMidiPort(card, portType);
}

int samOpenMidi(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  unsigned portIdx=SAM_PORT_IDX(inode);
  SamMidiPort *port;
  int error;

  printk( "%d\n", portIdx );
  
  if(portIdx>=SamMidiPortSentinel)
    return -ENODEV;

  down(&card->midi.accessLock);
  
  port=card->midi.ports+portIdx;
  if(port->owner!=SamMidiPortOwnerNone) {
    up(&card->midi.accessLock);
    return -EBUSY;
  }

  file->private_data=(void *)portIdx;
  card->midi.ports[portIdx].owner=SamMidiPortOwnerMidi;

  if((error=resetMidi(card, portIdx, file->f_mode & SAM_MODE_READ_MASK))<0) {
    card->midi.ports[portIdx].owner=SamMidiPortOwnerNone;
    up(&card->midi.accessLock);
    return error;
  }

  up(&card->midi.accessLock);
  return 0;
}

void samReleaseMidi(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  int portIdx=(int)file->private_data;
  SamMidiPort *port=card->midi.ports+portIdx;

  down(&card->midi.accessLock);
  if(resetMidi(card, portIdx, 0)<0)
    printk(KERN_ERR "sam9407: couldn't reset midi channel\n");
  port->owner=SamMidiPortOwnerNone;
  up(&card->midi.accessLock);
}

int samReadMidi(struct inode *inode, struct file *file, char *buf, int count, SamCard *card, char *closure)
{
  int portIdx=(int)file->private_data;
  unsigned flags;
  int error, n;
  unsigned char *cp;

  if(count<0)
    return -EINVAL;

  if((error=verify_area(VERIFY_WRITE, buf, count))<0)
    return error;

  /* only SamMidiPortExternal can return useful data */
  if(portIdx!=SamMidiPortExternal)
    return 0;

  cp=(unsigned char *)buf;
  n=count;

  down(&card->midi.accessLock);

  if(!card->midi.recOverflow && !card->midi.rec.usedCount &&
     (file->f_flags & O_NONBLOCK)) {
    up(&card->midi.accessLock);
    return -EAGAIN;
  }

  /* wait for incoming data */
  save_flags(flags); cli();
  while(!card->midi.recOverflow && !card->midi.rec.usedCount) {
    up(&card->midi.accessLock);
    interruptible_sleep_on(&card->midi.dataAvail);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current)) {
      restore_flags(flags);
      return -ERESTARTSYS;
    }
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked) {
      restore_flags(flags);
      return -ERESTARTSYS;
    }
#endif /* Linux 2.0 */
    down(&card->midi.accessLock);
    cli();
  }
  restore_flags(flags);

  if(card->midi.recOverflow) {
    save_flags(flags); cli();
    card->midi.recOverflow=0;
    card->midi.rec.usedIdx=card->midi.rec.usedCount=0;
    restore_flags(flags);
    up(&card->midi.accessLock);
    return -EIO;
  }

  while(n>0) {
    save_flags(flags); cli();
    if(!card->midi.rec.usedCount) {
      restore_flags(flags);
      break;
    }
    put_user(card->midi.rec.ringBuffer[card->midi.rec.usedIdx], cp);
    card->midi.rec.usedIdx=(card->midi.rec.usedIdx+1)%SAM_STORED_MIDI_EVENTS;
    card->midi.rec.usedCount--;
    restore_flags(flags);
    cp++;
    n--;
  }

  up(&card->midi.accessLock);

  return cp-(unsigned char *)buf;
}

int samWriteMidi(struct inode *inode, struct file *file, const char *buf, int count, SamCard *card, char *closure)
{ 
  int portIdx=(int)file->private_data;
  unsigned char *cp;
  int error, sigPending, n, size;

  if(count<0)
    return -EINVAL;

  if((error=verify_area(VERIFY_READ, buf, count))<0)
    return error;

  n=count;
  cp=(unsigned char *)buf;

  down(&card->midi.accessLock);

  while(n>0) {
    /* check for pending signals */
#if LINUX_VERSION_CODE>=0x20200
    sigPending=signal_pending(current);
#else /* Linux 2.0 */
    sigPending=current->signal & ~current->blocked;
#endif /* Linux 2.0 */
    if(sigPending) {
      up(&card->midi.accessLock);
      if(cp>(unsigned char *)buf)
	return cp-(unsigned char *)buf;
      else
	return -ERESTARTSYS;
    }

    /* limit transfers to SAM_STORED_MIDI_EVENTS
     * so we won't block signals for too long a time
     */
    size=n;
    if(size>SAM_STORED_MIDI_EVENTS)
      size=SAM_STORED_MIDI_EVENTS;
    if((error=samTalk(card,
		      SamSendCtrl, 1, samMidiSelCtrls[portIdx],
		      SamSendMidiFS, cp, size,
		      SamEndOfScript))<0)
      return error;
    cp+=size;
    n-=size;
  }
 
  up(&card->midi.accessLock);

  return cp-(unsigned char *)buf;
}

#if LINUX_VERSION_CODE>=0x20200
unsigned samPollMidi(struct inode *inode, struct file *file, poll_table *wait, SamCard *card, mode_t mode, char *closure)
{
  unsigned portIdx=SAM_PORT_IDX(inode);
  unsigned flags, mask;

  poll_wait(file, &card->midi.dataAvail, wait);

  mask=POLLOUT|POLLWRNORM;

  if(portIdx!=SamMidiPortExternal) {
    save_flags(flags); cli();
    if(card->midi.recOverflow || card->midi.rec.usedCount)
      mask|=POLLIN | POLLRDNORM;
    restore_flags(flags);
  } else
    /* non-external ports always return EOF */
    mask|=POLLIN | POLLRDNORM;

  return mask;
}
#else /* Linux 2.0 */
int samSelectMidi(struct inode *inode, struct file *file, int type, select_table *wait, SamCard *card, char *closure)
{
  unsigned portIdx=SAM_PORT_IDX(inode);
  unsigned flags;

  switch(type) {
  case SEL_IN:
    /* non-external ports always return EOF */
    if(portIdx!=SamMidiPortExternal)
      return 1;

    save_flags(flags); cli();
    if(card->midi.recOverflow || card->midi.rec.usedCount) {
      restore_flags(flags);
      return 1;
    }
    select_wait(&card->midi.dataAvail, wait);
    restore_flags(flags);
    break;
  case SEL_OUT:
    return 1;
    break;
  }
  return 0;
}
#endif /* Linux 2.0 */

int samIoctlMidi(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg, SamCard *card, mode_t mode, char *closure, int callNr)
{
  switch(cmd) {
  default:
    return -ENOSYS;
  }
}
