/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/soundcard.h>
#include <asm/io.h>

#include "driver.h"
#include "io.h"
#include "alloc.h"
#include "audio.h"
#include "mixer.h"

#define ABS(v) ((v)>=0?(v):-(v))
unsigned SBPort = 0x220;

static void SB_Mixer_Out(unsigned reg, unsigned val)
{
  int v = ((val<<4)|(val&0xf));
  outb((char)reg, SBPort+0x4);
  outb((char)v,   SBPort+0x5);
}

static int SB_Mixer_In(int reg)
{
  outb( (char)reg, SBPort+0x4 );
  return inb(SBPort+0x5);
}

static int putInt(unsigned long arg, int value)
{
  int error;
  if(value<0)
    return value;
  if((error=verify_area(VERIFY_WRITE, (void *)arg, sizeof(int)))<0)
    return error;
  SAM_PUT_USER(value, (int *)arg);
  return 0;
}

static int inputAvailable(SamCard *card, unsigned mask)
{
  int i;

  if((card->mixer.firmwareDirty&mask) ||
     (card->mixer.sbankDirty&mask))
    return 1;

  for(i=0; i<SAM_MIXER_ID_LAST; i++)
    if((card->mixer.idDirty[i]&mask))
      return 1;

  return 0;
}

static int driverInfo(SamCard *card, SamDriverInfo *info)
{
  int error;
  SamDriverInfo tmpInfo;

  if((error=verify_area(VERIFY_WRITE, info, sizeof(SamDriverInfo)))<0)
    return error;

  memset(&tmpInfo, 0, sizeof(tmpInfo));
  strncpy(tmpInfo.hardware, card->hardwareInfo->name, sizeof(tmpInfo.hardware)-1);

  if(card->firmwareLoaded) {
    strncpy(tmpInfo.firmware, card->firmwareInfo->name, sizeof(tmpInfo.firmware)-1);
    samGetMemInfo(card, &tmpInfo);
    samGetAudioInfo(card, &tmpInfo);
  }

  SAM_MEMCPY_TOFS(info, &tmpInfo, sizeof(SamDriverInfo));

  return 0;
}

static int getMixer(SamCard *card, SamMixerValue *values)
{
  unsigned char ctrlTbl[0x80];
  int error;
  unsigned addr, v=0, nch;
  SamMixerValue *mp;
  SamMixerVendorInfo *info;
  SamChannelInfo *chp;

  if((error=samFindMMT(card, SAM_MMT_ID_CTRL_TBL, SAM_MMT_ID_MASK, &addr, NULL))<0)
    return error;
  if(error==0)
    return -EIO;
  if((error=samTalk(card,
		    SamReadMem, addr, &ctrlTbl, sizeof(ctrlTbl)>>1,
		    SamEndOfScript))<0)
    return error;
  
  for(mp=values;;mp++) {
    if((error=verify_area(VERIFY_WRITE, mp, sizeof(SamMixerValue)))<0)
      return error;
    if(SAM_GET_USER(&mp->id)==SAM_MIXER_ID_SENTINEL)
      return 0;
    if(SAM_GET_USER(&mp->id)<0 || SAM_GET_USER(&mp->id)>=SAM_MIXER_ID_LAST)
      return -EINVAL;
    info=card->firmwareInfo->mixerInfo+SAM_GET_USER(&mp->id);
    if(info->type==SamMixerTypeUnsupported)
      continue;

    switch(info->type) {
    case SamMixerTypeBoolean:
	  v=ctrlTbl[info->ctrl];
      v=!!v;
      break;
    case SamMixerTypeByteEWS64:
    case SamMixerTypeByteEWS64Twice:
      /* EWS64 is using ranges
       * from 0x00-0x80, translate them to 0x00-0xFF,
       * so we won't end up having different ioctl API's
       */
	  v=ctrlTbl[info->ctrl];
      v=ABS((char)v)*2;
      if(v>=0xFF)
	  v=0xFF;
      break;
    case SamMixerTypeByteEWS647F:
	  v=ctrlTbl[info->ctrl];
      v=(v&0x7F)<<1;
      break;
    case SamMixerType7Bit:
	  v=ctrlTbl[info->ctrl];
      v&=0x7F;
      break;
    case SamMixerType3Bit:
	  v=ctrlTbl[info->ctrl];
      v&=0x7;
      break;
    case SamMixerType0To2:
	  v=ctrlTbl[info->ctrl];
      v&=0x3;
      break;
    case SamMixerTypeWav:
      if( SAM_GET_USER(&mp->id)>=SAM_MIXER_ID_WAV_VOL0 && 
		  SAM_GET_USER(&mp->id)<=SAM_MIXER_ID_WAV_VOL8 )
      {
	    nch = SAM_GET_USER(&mp->id) - SAM_MIXER_ID_WAV_VOL0;
		chp = card->channels+nch;
		v = chp->volume.left;
      }
	  else
	  {
		nch = SAM_GET_USER(&mp->id) - SAM_MIXER_ID_WAV_AUX_VOL0;
		chp = card->channels+nch;
		v = chp->volume.auxLeft;
	  }
      break;
    case SamMixerTypeSB:
      v = SB_Mixer_In( info->ctrl );
      break;

    default:
	  v=ctrlTbl[info->ctrl];
	  break;
      /* make compiler happy */
    }
    SAM_PUT_USER(v, &mp->value);
  }
  return 0;
}

static int infoMixer(SamCard *card, SamMixerValue *values)
{
  int error;
  SamMixerValue *mp;
  SamMixerVendorInfo *info;

  for(mp=values;;mp++) {
    if((error=verify_area(VERIFY_WRITE, mp, sizeof(SamMixerValue)))<0)
      return error;
    if(SAM_GET_USER(&mp->id)==SAM_MIXER_ID_SENTINEL)
      return 0;
    if(SAM_GET_USER(&mp->id)<0 || SAM_GET_USER(&mp->id)>=SAM_MIXER_ID_LAST)
      return -EINVAL;
    info=card->firmwareInfo->mixerInfo+SAM_GET_USER(&mp->id);
    SAM_PUT_USER(info->type!=SamMixerTypeUnsupported, &mp->value);
  }
  return 0;
}

//+kjk
static int setMixerDaad(SamCard *card, unsigned short value )
{
  samSetDaad( card, value );  
  return 0;
}

static int setMixer(SamCard *card, SamMixerValue *values)
{
  int error, isWord, i, isWav, nch=0;
  unsigned addr, v, ctrlValue;
  SamMixerValue *mp;
  SamMixerVendorInfo *info;
  SamChannelInfo *chp;

  for(mp=values;;mp++) {
    if((error=verify_area(VERIFY_READ, mp, sizeof(SamMixerValue)))<0)
      return error;
    if(SAM_GET_USER(&mp->id)==SAM_MIXER_ID_SENTINEL)
      return 0;
    if(SAM_GET_USER(&mp->id)<0 || SAM_GET_USER(&mp->id)>=SAM_MIXER_ID_LAST)
      return -EINVAL;
    info=card->firmwareInfo->mixerInfo+SAM_GET_USER(&mp->id);
    if(info->type==SamMixerTypeUnsupported)
      continue;

    v=0;
    isWord=0;
	isWav =0;
    switch(info->type) {
    case SamMixerTypeBoolean:
      v=SAM_GET_USER(&mp->value)?0x7F:0x00;
      break;
    case SamMixerTypeByte:
      v=SAM_GET_USER(&mp->value);
      if(v>0xFF)
		v=0xFF;
      break;
    case SamMixerTypeByteEWS64:
      v=(SAM_GET_USER(&mp->value)+1)/2;
      if(v>0x80)
		v=0x80;
      break;
    case SamMixerTypeByteEWS64Twice:
      v=(SAM_GET_USER(&mp->value)+1)/2;
      if(v>0x80)
		v=0x80;
      v|=v<<8;
      isWord=1;
      break;
    case SamMixerTypeByteEWS647F:
      v=SAM_GET_USER(&mp->value)/2;
      if(v>0x7F)
		v=0x7F;
      break;
    case SamMixerType7Bit:
      v=SAM_GET_USER(&mp->value);
      if(v>0x7F)
		v=0x7F;
      break;
    case SamMixerType3Bit:
      v=SAM_GET_USER(&mp->value);
      if(v>0x07)
		v=0x07;
      break;
    case SamMixerType0To2:
      v=SAM_GET_USER(&mp->value);
      if(v>2)
		v=2;
      break;
    case SamMixerTypeWav:
	  isWav = 1;
      v=SAM_GET_USER(&mp->value);
      if( SAM_GET_USER(&mp->id)>=SAM_MIXER_ID_WAV_VOL0 && 
		  SAM_GET_USER(&mp->id)<=SAM_MIXER_ID_WAV_VOL8 )
	  {
		  nch = SAM_GET_USER(&mp->id) - SAM_MIXER_ID_WAV_VOL0;
		  chp = card->channels+nch;
		  chp->volume.left  = v;
		  chp->volume.right = v;
	  }
	  else
	  {
		  nch = SAM_GET_USER(&mp->id) - SAM_MIXER_ID_WAV_AUX_VOL0;
		  chp = card->channels+nch;
		  chp->volume.auxLeft  = v;
		  chp->volume.auxRight = v;
	  }
      break;
    case SamMixerTypeSB:
      v=SAM_GET_USER(&mp->value);
      SB_Mixer_Out( info->ctrl, v );
	  return 0;
	  
    default:
      /* make compiler happy */
    }

    if(info->doesReset) {
      if(card->firmwareInfo->resetRestoreCtrl) {
	  /* retrieve the current value to send along with
	   * reset hook control
	   */
	  if((error=samFindMMT(card, SAM_MMT_ID_CTRL_TBL, SAM_MMT_ID_MASK, &addr, NULL))<0)
		return error;
	  if(error==0)
		return -EIO;
	  /* read word aligned word boundary from control table */
	  if((error=samTalk(card,
				SamReadMem, addr+(card->firmwareInfo->resetRestoreCtrl>>1), &ctrlValue, 1,
				SamEndOfScript))<0)
		return error;
	  /* if offset is odd, take the MSB, otherwise LSB */
	  if(card->firmwareInfo->resetRestoreCtrl&1)
		ctrlValue>>=8;
		ctrlValue&=0xFF;
      }

      if(isWord)
		   error=samTalkNotBusy(card,
			     SamSendCtrl, 3, info->ctrl, v, v>>8,
			     SamExpect, 0,
			     SamUARTMode,
			     SamSendCtrlIf, card->firmwareInfo->resetRestoreCtrl, 2, card->firmwareInfo->resetRestoreCtrl, ctrlValue,
			     SamSendCtrlP, card->firmwareInfo->nResetReadjust, card->firmwareInfo->resetReadjust,
			     SamEndOfScript);
      else
		   error=samTalkNotBusy(card,
			     SamSendCtrl, 2, info->ctrl, v,
			     SamExpect, 0,
			     SamUARTMode,
			     SamSendCtrlIf, card->firmwareInfo->resetRestoreCtrl, 2, card->firmwareInfo->resetRestoreCtrl, ctrlValue,
			     SamSendCtrlP, card->firmwareInfo->nResetReadjust, card->firmwareInfo->resetReadjust,
			     SamEndOfScript);

      /* whenever the card received a reset
       * assume that all controls have become dirty
       */
      for(i=0; i<SAM_MIXER_ID_LAST; i++)
		if(card->firmwareInfo->mixerInfo[i].type!=SamMixerTypeUnsupported)
		  card->mixer.idDirty[i]=~0;
      wake_up(&card->mixer.waitDirty);
    } else {
      if(isWord)
	  {
		error=samTalk(card,
		      SamSendCtrl, 3, info->ctrl, v, v>>8,
		      SamEndOfScript);
	  }
      else if( isWav )
	  {
		error=samTalk(card, // left volume
		      SamSendCtrl, 4, info->ctrl,   nch, v, v>>8,  //left
		      SamSendCtrl, 4, info->ctrl+1, nch, v, v>>8, //right
		      SamEndOfScript);
	  } else {
		error=samTalk(card,
		      SamSendCtrl, 2, info->ctrl, v,
		      SamEndOfScript);
	  }
    }

    if(error<0)
      return error;

    if(info->storeAt>=0)
      *(unsigned *)((char *)&card->ctrlValues+info->storeAt)=v;

    card->mixer.idDirty[SAM_GET_USER(&mp->id)]=~0;
    wake_up(&card->mixer.waitDirty);
  }
  return 0;
}

int samOpenMixer(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  int i, slot;
  unsigned mask;
  down(&card->mixer.accessLock);

  /* find a free slot in usedMask */
  for(slot=0; slot<sizeof(unsigned)*8; slot++)
    if(!(card->mixer.usedMask & (1<<slot))) {
      card->mixer.usedMask|=(1<<slot);
      file->private_data=(void *)slot;
      break;
    }

  if(slot>=sizeof(unsigned)*8) {
    /* all slots used */
    up(&card->mixer.accessLock);
    return -EBUSY;
  }

  /* reset dirty flags */
  mask=~(1<<slot);
  card->mixer.firmwareDirty&=mask;
  card->mixer.sbankDirty&=mask;
  for(i=0; i<SAM_MIXER_ID_LAST; i++)
    card->mixer.idDirty[i]&=mask;

  up(&card->mixer.accessLock);
  return 0;
}

void samReleaseMixer(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  int slot=(int)file->private_data;
  
  down(&card->mixer.accessLock);
  card->mixer.usedMask&=~(1<<slot);
  up(&card->mixer.accessLock);
}

int samReadMixer(struct inode *inode, struct file *file, char *buf, int count, SamCard *card, char *closure)
{
  int slot=(int)file->private_data;
  unsigned flags, mask;
  int error, n, i;

  if(count<0)
    return -EINVAL;

  if((error=verify_area(VERIFY_WRITE, buf, count))<0)
    return error;

  down(&card->mixer.accessLock);

  mask=1<<slot;

  if(!inputAvailable(card, mask) && (file->f_flags & O_NONBLOCK)) {
    up(&card->mixer.accessLock);
    return -EAGAIN;
  }

  /* wait for dirty flags */
  while(!inputAvailable(card, mask)) {
    save_flags(flags); cli();
    up(&card->mixer.accessLock);
    interruptible_sleep_on(&card->mixer.waitDirty);
    restore_flags(flags);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current))
      return -ERESTARTSYS;
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked)
      return -ERESTARTSYS;
#endif /* Linux 2.0 */
    down(&card->mixer.accessLock);
  }

  n=0;

  if(n<count && (card->mixer.firmwareDirty&mask)) {
    SAM_PUT_USER(SAM_MIXER_FIRMWARE_CHANGED, buf+n++);
    card->mixer.firmwareDirty&=~mask;
  }

  if(n<count && (card->mixer.sbankDirty&mask)) {
    SAM_PUT_USER(SAM_MIXER_SBANK_CHANGED, buf+n++);
    card->mixer.sbankDirty&=~mask;
  }

  for(i=0; i<SAM_MIXER_ID_LAST && n<count; i++)
    if(card->mixer.idDirty[i]&mask) {
      SAM_PUT_USER(i, buf+n++);
      card->mixer.idDirty[i]&=~mask;
    }

  up(&card->mixer.accessLock);
  return n;
}

#if LINUX_VERSION_CODE>=0x20200
unsigned samPollMixer(struct inode *inode, struct file *file, poll_table *wait, SamCard *card, mode_t mode, char *closure)
{
  int slot=(int)file->private_data;
  unsigned mask;

  poll_wait(file, &card->mixer.waitDirty, wait);

  down(&card->mixer.accessLock);
  mask=0;
  if(inputAvailable(card, 1<<slot))
    mask|= POLLIN | POLLRDNORM;
  up(&card->mixer.accessLock);

  return mask;
}
#else /* Linux 2.0 */
int samSelectMixer(struct inode *inode, struct file *file, int type, select_table *wait, SamCard *card, char *closure)
{
  int slot=(int)file->private_data;
  unsigned flags;

  switch(type) {
  case SEL_IN:
    down(&card->mixer.accessLock);
    if(inputAvailable(card, 1<<slot)) {
      up(&card->mixer.accessLock);
      return 1;
    }
    save_flags(flags); cli();
    up(&card->mixer.accessLock);
    select_wait(&card->mixer.waitDirty, wait);
    restore_flags(flags);
    break;
  case SEL_OUT:
    /* write(2) not used for mixer */
    break;
  }
  return 0;
}
#endif /* Linux 2.0 */

int samIoctlMixer(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg, SamCard *card, mode_t mode, char *closure, int callNr)
{
  int error;

  down(&card->mixer.accessLock);
  switch(cmd) {
  case SOUND_MIXER_READ_DEVMASK:
  case SAM_IOC_DRIVER_INFO:
  case SAM_IOC_FIRMWARE_LOAD:
    break;
  default:
    if(!card->firmwareLoaded) {
      up(&card->mixer.accessLock);
      return -ENOPKG;
    }
  }
  
  switch(cmd) {
  case SOUND_MIXER_READ_DEVMASK:
    error=putInt(arg, 0);
    break;
  case SAM_IOC_DRIVER_INFO:
    error=driverInfo(card, (SamDriverInfo *)arg);
    break;
  case SAM_IOC_FIRMWARE_LOAD:
    error=samFirmwareLoad(card, (SamFirmware *)arg);
    break;
  case SAM_IOC_MIXER_GET:
    error=getMixer(card, (SamMixerValue *)arg);
    break;
  case SAM_IOC_MIXER_SET:
    error=setMixer(card, (SamMixerValue *)arg);
    break;
  case SAM_IOC_MIXER_INFO:
    error=infoMixer(card, (SamMixerValue *)arg);
    break;
  case SAM_IOC_BANK_LOAD:
    error=samSBankLoad(card, (SamSoundBank *)arg);
    card->mixer.sbankDirty=~0;
    wake_up(&card->mixer.waitDirty);
    break;
  case SAM_IOC_BANK_UNLOAD:
    error=samSBankUnload(card, (unsigned char *)arg);
    card->mixer.sbankDirty=~0;
    wake_up(&card->mixer.waitDirty);
    break;
  case SAM_IOC_BANK_PRIO_CHG:
    error=samSBankPrioChange(card, (SamSoundBankPrio *)arg);
    card->mixer.sbankDirty=~0;
    wake_up(&card->mixer.waitDirty);
    break;
  case SAM_IOC_BANK_INFO:
    error=samSBankInfo(card, (SamSoundBankInfo *)arg);
    break;

//+kjk	
  case SAM_IOC_SET_DAAD:
	error = setMixerDaad( card, (unsigned short)arg );
    break;
	
  default:
    error=-ENOSYS;
  }
  up(&card->mixer.accessLock);

  return error;
}
