/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE>=0x20200
#include <linux/vmalloc.h>
#endif
#include <asm/byteorder.h>

#include "driver.h"
#include "io.h"
#include "alloc.h"
#include "mixer.h"

#define SAM_MMT_BANK_MASK		0xF7F
#define SAM_MAX_BANK_SIZE		(256<<20)
#define SAM_MAX_MT_WORDS		(64<<10)
#define SAM_PARAM_PAGE_WORDS		(64<<10)
#define SAM_PARAM_BLOCKS		2
#define SAM_SAMPLE_PAGE_WORDS		(256<<10)
#define SAM_SAMPLE_BLOCKS		16
#define SAM_INSTR_PROGRAMS		128

#ifdef __LITTLE_ENDIAN
#define SAM_CKID_VALUE(c1,c2,c3,c4)	((c1)|(c2)<<8|(c3)<<16|(c4)<<24)
#else
#define SAM_CKID_VALUE(c1,c2,c3,c4)	((c1)<<24|(c2)<<16|(c3)<<8|(c4))
#endif
#define SAM_CKID_DRCP			SAM_CKID_VALUE('D','R','C','P')
#define SAM_CKID_PRIO			SAM_CKID_VALUE('P','R','I','O')
#define SAM_CKID_IMAP			SAM_CKID_VALUE('I','M','A','P')
#define SAM_CKID_RLOC			SAM_CKID_VALUE('R','L','O','C')
#define SAM_CKID_PARA			SAM_CKID_VALUE('P','A','R','A')
#define SAM_CKID_SAMP			SAM_CKID_VALUE('S','A','M','P')

#define SAM_RELOC_TYPE_PARAM		0
#define SAM_RELOC_TYPE_MA2_PG		1
#define SAM_RELOC_TYPE_MA2_LOOP		2
#define SAM_RELOC_TYPE_MB_END		3

typedef struct {
  unsigned id;
  unsigned size;
} RiffChunk;

typedef struct {
  unsigned addr;
  unsigned char type, block;
} __attribute__((packed)) RelocEntry;

typedef struct {
  unsigned short program;
  unsigned char variation;
  unsigned char page;
  unsigned short paramPtr;
} ImapEntry;

typedef struct {
  unsigned short length;
  unsigned short paramPages[SAM_NROF_SBANKS*SAM_PARAM_BLOCKS];
  unsigned short instrPtrs[SAM_INSTR_PROGRAMS];
} MidiTable;

typedef struct {
  unsigned char variation;
  unsigned char bank;
  unsigned short paramPtr;
} InstrDef;

typedef struct {
  unsigned short program;
  unsigned char variation;
  unsigned char bank;
  unsigned short paramPtr;
} DrumDef;

typedef enum {
  PriorityItemImapInstr,
  PriorityItemImapDrum,
  PriorityItemInstrDef,
  PriorityItemDrumDef,
  PriorityItemIgnore,
} PriorityItemType;

typedef struct {
  PriorityItemType type;
  union {
    struct {
      ImapEntry *ptr;
    } imap;
    struct {
      InstrDef *ptr;
      unsigned short program;
    } instr;
    struct {
      DrumDef *ptr;
    } drum;
  } u;
} PriorityItem;

static int allocBlocks(SamCard *card,
		       int id, unsigned short *source, unsigned words, unsigned boundary,
		       unsigned blockWords, int maxBlocks, unsigned *blockAddrs, int *nBlocks)
{
  unsigned n, i;
  int error;
  *nBlocks=0;
  for(i=0; i<words; i+=blockWords) {
    if(*nBlocks>=maxBlocks) {
#ifdef DEBUG
      printk(KERN_DEBUG "sam9407: sounbank alloc: too many blocks\n");
#endif
      return -EINVAL;
    }
    n=words-i<blockWords ? words-i : blockWords;
    if((error=samAlloc(card, id|(*nBlocks<<12), n, blockAddrs+*nBlocks, boundary, NULL, NULL))<0)
      return error;
    if(source)
      if((error=samTalk(card, SamWriteMemFS, blockAddrs[*nBlocks], source+i, n, SamEndOfScript))<0)
	return error;
    (*nBlocks)++;
  }
  return 0;
}

static int loadMidiTable(SamCard *card, MidiTable **midiTableP)
{
  int error, i;
  unsigned addr, len, addr2, len2;
  unsigned char *cp;
  MidiTable *midiTable;

  *midiTableP=NULL;

  if((error=samFindMMT(card, SAM_MMT_ID_MIDI_MT, SAM_MMT_ID_MASK, &addr, &len))<0)
    return error;
  if(error==0)
    /* no Midi Mapping Table available */
    return 0;
  if((error=samFindMMT(card, SAM_MMT_ID_MIDI_MT_EXT, SAM_MMT_ID_MASK, &addr2, &len2))<0)
    return error;
  if(error>0 && addr2==addr+len)
    len+=len2;
  if((len<<1)<sizeof(MidiTable)) {
    /* table is too short */
    return 0;
  }
  if(!(cp=(unsigned char *)kmalloc(len<<1, GFP_KERNEL)))
    return -ENOMEM;
  midiTable=(MidiTable *)cp;
  if((error=samTalk(card, SamReadMem, addr, cp, len, SamEndOfScript))<0) {
    kfree(cp);
    return error;
  }
  if(midiTable->length>len) {
    /* truncated Midi Mapping Table */
    kfree(cp);
    return 0;
  }
  /* adjust midi mapping table offsets downwards */
  if(addr&0xFFFF)
    for(i=0; i<SAM_INSTR_PROGRAMS; i++)
      midiTable->instrPtrs[i]-=addr&0xFFFF;

  *midiTableP=midiTable;
  return 0;
}

static int relocateWriteMidiTable(SamCard *card, unsigned addr, void *closure)
{
  MidiTable *midiTable=(MidiTable *)closure;
  int error, i;
  /* relocate instrument pointers */
  if(addr&0xFFFF)
    for(i=0; i<SAM_INSTR_PROGRAMS; i++)
      midiTable->instrPtrs[i]+=addr&0xFFFF;

  error=samTalk(card,
		SamWriteMem, addr, midiTable, midiTable->length,
		SamEndOfScript);

  /* change instrument pointers back */
  if(addr&0xFFFF)
    for(i=0; i<SAM_INSTR_PROGRAMS; i++)
      midiTable->instrPtrs[i]-=addr&0xFFFF;

  return error;
}

static int countInstruments(SamCard *card,
			    unsigned char *imapPtr, unsigned imapSize,
			    MidiTable *midiTable,
			    PriorityItem *prioItems,
			    int *nInstrs, int *nDrums)
{
  PriorityItem *prioItem;
  ImapEntry *imapEntry;
  InstrDef *instrDef;
  DrumDef *drumDef;
  short *limit;
  int nPrioItems, isBackup, isDrums, bank, program, i;

  if(nInstrs)
    *nInstrs=0;
  if(nDrums)
    *nDrums=0;

  nPrioItems=0;

  /* count instruments & drums from IMAP */
  if(imapPtr) {
    i=0;
    for(isDrums=0; isDrums<2; isDrums++) {
      for(;;) {
	if(i+2>imapSize) {
	  /* instrument set not finished correctly */
#ifdef DEBUG
	  printk(KERN_DEBUG "sam9407: instrument set not finished correctly\n");
#endif
	  return -EINVAL;
	}
	if(SAM_GET_USER((unsigned short *)(imapPtr+i))==0xFFFF) {
	  i+=2;
	  break;
	}
	imapEntry=(ImapEntry *)(imapPtr+i);
	program=SAM_GET_USER(&imapEntry->program);
	if(SAM_GET_USER(&imapEntry->page)>=SAM_PARAM_BLOCKS) {
#ifdef DEBUG
	  printk(KERN_DEBUG "sam9407: too many parameter block pages\n");
#endif
	  return -EINVAL;
	}
	if(isDrums) {
	  if(nDrums)
	    (*nDrums)++;
	  if(prioItems) {
	    prioItem=prioItems+nPrioItems++;
	    prioItem->type=PriorityItemImapDrum;
	    prioItem->u.imap.ptr=imapEntry;
	  }
	} else {
	  if(program>=SAM_INSTR_PROGRAMS) {
#ifdef DEBUG
	    printk(KERN_DEBUG "sam9407: instrument program out of range\n");
#endif
	    return -EINVAL;
	  }
	  if(nInstrs)
	    (*nInstrs)++;
	  if(prioItems) {
	    prioItem=prioItems+nPrioItems++;
	    prioItem->type=PriorityItemImapInstr;
	    prioItem->u.imap.ptr=imapEntry;
	  }
	}
	i+=sizeof(ImapEntry);
      }
    }
  }

  /* count instruments & drums from MidiTable */
  if(midiTable) {
    /* count instruments */
    limit=(short *)midiTable+midiTable->length;

    for(i=0; i<SAM_INSTR_PROGRAMS; i++) {
      instrDef=(InstrDef *)((short *)midiTable+midiTable->instrPtrs[i]);
      for(isBackup=0; isBackup<2; isBackup++) {
	while((short *)(instrDef+1)<=limit &&
	      *(unsigned short *)instrDef!=0xFFFF) {
	  bank=instrDef->bank/SAM_PARAM_BLOCKS;
	  if(bank<SAM_NROF_SBANKS &&
	     *card->mem.sbank[bank].name) {
	    if(nInstrs)
	      (*nInstrs)++;
	    if(prioItems) {
	      prioItem=prioItems+nPrioItems++;
	      prioItem->type=PriorityItemInstrDef;
	      prioItem->u.instr.ptr=instrDef;
	      prioItem->u.instr.program=i;
	    }
	  }
	  instrDef++;
	}
	instrDef=(InstrDef *)((short *)instrDef+1);
      }
    }

    /* count drums */
    limit=(short *)midiTable+midiTable->length;

    drumDef=(DrumDef *)&midiTable[1];
    for(isBackup=0; isBackup<2; isBackup++) {
      while((short *)(drumDef+1)<=limit &&
	    *(unsigned short *)drumDef!=0xFFFF) {
	bank=drumDef->bank/SAM_PARAM_BLOCKS;
	if(bank<SAM_NROF_SBANKS &&
	   *card->mem.sbank[bank].name) {
	  if(nDrums)
	    (*nDrums)++;
	  if(prioItems) {
	    prioItem=prioItems+nPrioItems++;
	    prioItem->type=PriorityItemDrumDef;
	    prioItem->u.drum.ptr=drumDef;
	  }
	}
	drumDef++;
      }
      drumDef=(DrumDef *)((short *)drumDef+1);
    }
  }

  return 0;
}

/* sorting helper functions */
extern inline int samPrioIsInstr(PriorityItem *item)
{
  switch(item->type) {
  case PriorityItemImapInstr:
  case PriorityItemInstrDef:
    return 1;
  default:
    return 0;
  }
}

extern inline int samPrioProgramNr(PriorityItem *item)
{
  switch(item->type) {
  case PriorityItemImapInstr:
  case PriorityItemImapDrum:
    return SAM_GET_USER(&item->u.imap.ptr->program);
  case PriorityItemInstrDef:
    return item->u.instr.program;
  case PriorityItemDrumDef:
    return item->u.drum.ptr->program;
  default:
    /* PriorityItemIgnore:
     * will never be called that way,
     * but we want to keep the compiler happy
     */
    return 0;
  }
}

extern inline int samPrioVariationNr(PriorityItem *item)
{
  switch(item->type) {
  case PriorityItemImapInstr:
  case PriorityItemImapDrum:
    return SAM_GET_USER(&item->u.imap.ptr->variation);
  case PriorityItemInstrDef:
    return item->u.instr.ptr->variation;
  case PriorityItemDrumDef:
    return item->u.drum.ptr->variation;
  default:
    /* PriorityItemIgnore:
     * will never be called that way,
     * but we want to keep the compiler happy
     */
    return 0;
  }
}

extern inline int samPrioSBank(SamCard *card, int imapId, PriorityItem *item)
{
  switch(item->type) {
  case PriorityItemImapInstr:
  case PriorityItemImapDrum:
    return imapId;
    break;
  case PriorityItemInstrDef:
    return item->u.instr.ptr->bank/SAM_PARAM_BLOCKS;
  case PriorityItemDrumDef:
    return item->u.drum.ptr->bank/SAM_PARAM_BLOCKS;
  default:
    /* PriorityItemIgnore:
     * will never be called that way,
     * but we want to keep the compiler happy
     */
    return 0;
  }
}

extern inline int samPrioPriority(SamCard *card, int imapId, PriorityItem *item)
{
  return card->mem.sbank[samPrioSBank(card, imapId, item)].priority;
}

extern inline int samComparePrioItem(SamCard *card, int imapId, PriorityItem *a, PriorityItem *b)
{
  int cmp;
  /* we sort according to:
   * (drum|instr) / program / variation / priority / sbank
   */
  if((cmp=samPrioIsInstr(a)-samPrioIsInstr(b)))
    return cmp;

  if((cmp=samPrioProgramNr(a)-samPrioProgramNr(b)))
    return cmp;

  if((cmp=samPrioVariationNr(a)-samPrioVariationNr(b)))
    return cmp;

  if((cmp=samPrioPriority(card, imapId, b)-samPrioPriority(card, imapId, a)))
    return cmp;

  if((cmp=samPrioSBank(card, imapId, b)-samPrioSBank(card, imapId, a)))
    return cmp;

  return 0;
}

extern inline void samExchangePrioItems(PriorityItem *a, PriorityItem *b)
{
  PriorityItem tmp;

  memcpy(&tmp, a, sizeof(PriorityItem));
  memcpy(a, b, sizeof(PriorityItem));
  memcpy(b, &tmp, sizeof(PriorityItem));
}

/* sort prioItems vector according to priorities */
static void qsortPrioItems(SamCard *card, int imapId, PriorityItem *low, PriorityItem *high)
{
  PriorityItem *i, *j, m;

  i=low;
  j=high;

  memcpy(&m, low+(high-low)/2, sizeof(PriorityItem));

  while(i<=j) {
    while(i<high && samComparePrioItem(card, imapId, i, &m)<0)
      i++;
    while(j>low && samComparePrioItem(card, imapId, j, &m)>0)
      j--;
    if(i<=j) {
      if(i!=j)
	samExchangePrioItems(i, j);
      i++;
      j--;
    }
  }

  if(low<j)
    qsortPrioItems(card, imapId, low, j);

  if(i<high)
    qsortPrioItems(card, imapId, i, high);
}

static void fillupMidiTable(SamCard *card, int imapId,
			    MidiTable *midiTable, unsigned *paramAddrs,
			    PriorityItem *prioItems, int nPrioItems)
{
  PriorityItem *prioItem;
  InstrDef *instrDef;
  DrumDef *drumDef;
  short *dest;
  int expectInstr, expectProgram, expectBackup;
  int lastProgram, lastVariation;
  int maxProgram, prioIdx, prioIdxSaved;

  dest=(short *)&midiTable[1];
  prioIdx=0;

  for(expectInstr=0; expectInstr<2; expectInstr++) {
    maxProgram=expectInstr ? SAM_INSTR_PROGRAMS : 1;
    for(expectProgram=0; expectProgram<maxProgram; expectProgram++) {
      /* setup instrument pointers in midiTable */
      if(expectInstr)
	midiTable->instrPtrs[expectProgram]=dest-(short *)midiTable;

      /* PriorityItem doesn't distinguish between
       * current/backup instrument/drum definitions
       * so we restart from here for both current and
       * backup definitions
       * We mark already handled items in prioItems
       */
      prioIdxSaved=prioIdx;
      lastProgram=lastVariation=-1;
      for(expectBackup=0; expectBackup<2; expectBackup++) {
	prioIdx=prioIdxSaved;
	while(prioIdx<nPrioItems) {
	  prioItem=prioItems+prioIdx;

	  if(prioItem->type==PriorityItemIgnore) {
	    prioIdx++;
	    continue;
	  }

	  /* check if the prioItem points to an element we expect */
	  if(expectInstr!=samPrioIsInstr(prioItem))
	    break;
	  if(expectInstr && expectProgram!=samPrioProgramNr(prioItem))
	    break;

	  /* in current instrument/drum set
	   * we skip ambiguous program/variation combinations
	   * to be handled in backup set
	   */
	  if(!expectBackup &&
	     lastProgram==samPrioProgramNr(prioItem) &&
	     lastVariation==samPrioVariationNr(prioItem)) {
	    prioIdx++;
	    continue;
	  }

	  /* all checks succeeded, append this entry */
	  switch(prioItem->type) {
	  case PriorityItemImapInstr:
	    instrDef=(InstrDef *)dest;
	    instrDef->variation=SAM_GET_USER(&prioItem->u.imap.ptr->variation);
	    instrDef->bank=imapId*SAM_PARAM_BLOCKS+SAM_GET_USER(&prioItem->u.imap.ptr->page);
	    instrDef->paramPtr=SAM_GET_USER(&prioItem->u.imap.ptr->paramPtr)+
	      (paramAddrs[SAM_GET_USER(&prioItem->u.imap.ptr->page)]&0xFFFF);
	    *(char **)&dest+=sizeof(InstrDef);
	    break;
	  case PriorityItemImapDrum:
	    drumDef=(DrumDef *)dest;
	    drumDef->program=SAM_GET_USER(&prioItem->u.imap.ptr->program);
	    drumDef->variation=SAM_GET_USER(&prioItem->u.imap.ptr->variation);
	    drumDef->bank=imapId*SAM_PARAM_BLOCKS+SAM_GET_USER(&prioItem->u.imap.ptr->page);
	    drumDef->paramPtr=SAM_GET_USER(&prioItem->u.imap.ptr->paramPtr)+
	      (paramAddrs[SAM_GET_USER(&prioItem->u.imap.ptr->page)]&0xFFFF);
	    *(char **)&dest+=sizeof(DrumDef);
	    break;
	  case PriorityItemInstrDef:
	    instrDef=(InstrDef *)dest;
	    memcpy(instrDef, prioItem->u.instr.ptr, sizeof(InstrDef));
	    *(char **)&dest+=sizeof(InstrDef);
	    break;
	  case PriorityItemDrumDef:
	    drumDef=(DrumDef *)dest;
	    memcpy(drumDef, prioItem->u.drum.ptr, sizeof(DrumDef));
	    *(char **)&dest+=sizeof(DrumDef);
	    break;
	  default:
	    /* PriorityItemIgnore:
	     * will never be called that way,
	     * but we want to keep the compiler happy
	     */
	  }

	  lastProgram=samPrioProgramNr(prioItem);
	  lastVariation=samPrioVariationNr(prioItem);
	  prioItem->type=PriorityItemIgnore;
	  prioIdx++;
	}
	/* finish instr/drum definition */
	*dest++=~0;
      }
    }
    /* finish drumsets */
    if(!expectInstr)
      *dest++=~0;
  }
}

static int adjustMidiTable(SamCard *card, int id, unsigned char *imapPtr, unsigned imapSize, unsigned *paramAddrs)
{
  int error, nInstrs, nDrums, newMidiTableSize, i;
  MidiTable *oldMidiTable, *newMidiTable;
  PriorityItem *prioItems;

  if((error=loadMidiTable(card, &oldMidiTable))<0)
    return error;

  /* first count how many instruments we have in total */
  if((error=countInstruments(card, imapPtr, imapSize, oldMidiTable, NULL, &nInstrs, &nDrums))<0) {
    if(oldMidiTable)
      kfree(oldMidiTable);
    return error;
  }

  newMidiTableSize=sizeof(MidiTable)+nDrums*sizeof(DrumDef)+6+nInstrs*sizeof(InstrDef)+SAM_INSTR_PROGRAMS*4;
  if(newMidiTableSize>SAM_MAX_MT_WORDS<<1) {
#ifdef DEBUG
    printk(KERN_DEBUG "sam9407: midi mapping table too large (%d bytes)\n", newMidiTableSize);
#endif
    if(oldMidiTable)
      kfree(oldMidiTable);
    return -ENOMEM;
  }

  if(nDrums+nInstrs>0) {
    if(!(prioItems=(PriorityItem *)kmalloc((nDrums+nInstrs)*sizeof(PriorityItem), GFP_KERNEL))) {
      if(oldMidiTable)
	kfree(oldMidiTable);
      return -ENOMEM;
    }

    /* now get the prioItems vector filled */
    if((error=countInstruments(card, imapPtr, imapSize, oldMidiTable, prioItems, NULL, NULL))<0) {
      if(prioItems)
	kfree(prioItems);
      if(oldMidiTable)
	kfree(oldMidiTable);
      return error;
    }

    if(nDrums+nInstrs>1)
      qsortPrioItems(card, id, prioItems, prioItems+nDrums+nInstrs-1);
  } else
    prioItems=NULL;

  if(!(newMidiTable=(MidiTable *)kmalloc(newMidiTableSize, GFP_KERNEL))) {
    if(prioItems)
      kfree(prioItems);
    if(oldMidiTable)
      kfree(oldMidiTable);
    return -ENOMEM;
  }
  memset(newMidiTable, 0, newMidiTableSize);
  if(oldMidiTable)
    memcpy(newMidiTable, oldMidiTable, sizeof(MidiTable));
  
  /* setup MIDI mapping table */
  newMidiTable->length=newMidiTableSize/2;

  if(id>=0) {
    if(paramAddrs)
      for(i=0; i<SAM_PARAM_BLOCKS; i++)
	newMidiTable->paramPages[id*SAM_PARAM_BLOCKS+i]=paramAddrs[i]>>16;
    else
      for(i=0; i<SAM_PARAM_BLOCKS; i++)
	newMidiTable->paramPages[id*SAM_PARAM_BLOCKS+i]=0;
  }

  fillupMidiTable(card, id, newMidiTable, paramAddrs, prioItems, nDrums+nInstrs);

  /* relocate instrument pointers and write midi table back */
  error=samAlloc(card, SAM_MMT_ID_MIDI_MT, newMidiTable->length, NULL, SAM_BOUNDARY_64K,
		 relocateWriteMidiTable, (void *)newMidiTable);

  if(error>=0 && card->firmwareInfo->nSBankRehash)
    error=samTalk(card,
		  SamSendCtrlP, card->firmwareInfo->nSBankRehash, card->firmwareInfo->sbankRehash,
		  SamEndOfScript);

  kfree(newMidiTable);
  if(prioItems)
    kfree(prioItems);
  if(oldMidiTable)
    kfree(oldMidiTable);
  return error;
}

int samRelocateMidiTable(SamCard *card)
{
  int error;
  MidiTable *midiTable;

  if((error=loadMidiTable(card, &midiTable))<0)
    return error;

  if(midiTable) {
    /* relocate and write midi mapping table back */
    error=samAlloc(card,
		   SAM_MMT_ID_MIDI_MT, midiTable->length, NULL, SAM_BOUNDARY_64K,
		   relocateWriteMidiTable, (void *)midiTable);
    
    if(error>=0 && card->firmwareInfo->nSBankRehash)
      error=samTalk(card,
		    SamSendCtrlP, card->firmwareInfo->nSBankRehash, card->firmwareInfo->sbankRehash,
		    SamEndOfScript);
    
    kfree(midiTable);
  }

  return error;
}

static int disposeBank(SamCard *card, int id)
{
  int error;

  if(!*card->mem.sbank[id].name)
    return 0;
  memset(card->mem.sbank+id, 0, sizeof(SamMMTSBank));

  if((error=adjustMidiTable(card, id, NULL, 0, NULL))<0 ||
     (error=samFree(card, SAM_MMT_ID_MIDI_PCM|(id<<8), SAM_MMT_BANK_MASK))<0 ||
     (error=samFree(card, SAM_MMT_ID_MIDI_PAR|(id<<8), SAM_MMT_BANK_MASK))<0)
    return error;
  return 0;
}

int samSBankLoad(SamCard *card, SamSoundBank *bank)
{
  int error, id, n, i, ofs;
  RiffChunk *chunk;
  RelocEntry *reloc;
  unsigned paramAddrs[SAM_PARAM_BLOCKS], sampleAddrs[SAM_SAMPLE_BLOCKS];
  int nParamBlocks, nSampleBlocks, paramBytes;
  unsigned size, nReloc, block, addr, paramStartAddr, imapSize, v;
  unsigned char *cp, *paramBuf, *imapPtr;

  if((error=verify_area(VERIFY_READ, bank, sizeof(SamSoundBank)))<0)
    return error;
  id=SAM_GET_USER(&bank->id);
  if(id>=SAM_NROF_SBANKS) {
#ifdef DEBUG
    printk(KERN_DEBUG "sam9407: sounbank id out of range\n");
#endif
    return -EINVAL;
  }
  if((error=disposeBank(card, id))<0)
    return error;
  /* traverse through soundbank chunks */
  chunk=(RiffChunk *)SAM_GET_USER(&bank->data);
  size=SAM_GET_USER(&bank->size);
  if(!chunk || !size || size>SAM_MAX_BANK_SIZE) {
#ifdef DEBUG
    printk(KERN_DEBUG "sam9407: invalid soundbank size\n");
#endif
    return -EINVAL;
  }
  memset(paramAddrs, 0, sizeof(paramAddrs));
  reloc=NULL; nReloc=0;
  paramStartAddr=0;
  paramBuf=NULL; paramBytes=0;
  nParamBlocks=nSampleBlocks=0;
  imapPtr=NULL; imapSize=0;
  error=0;
  while(error>=0 && size>0) {
    cp=(char *)chunk+sizeof(RiffChunk);
    if((error=verify_area(VERIFY_READ, chunk, sizeof(RiffChunk)))<0 ||
       (error=verify_area(VERIFY_READ, cp, SAM_GET_USER(&chunk->size)))<0)
      break;
    if(SAM_GET_USER(&chunk->size)>SAM_MAX_BANK_SIZE ||
       sizeof(RiffChunk)+SAM_GET_USER(&chunk->size)>size) {
#ifdef DEBUG
      printk(KERN_DEBUG "sam9407: RIFF chunk too large\n");
#endif
      error=-EINVAL;
      break;
    }
    switch(SAM_GET_USER(&chunk->id)) {
    case SAM_CKID_DRCP:
      break;
    case SAM_CKID_PRIO:
      if(SAM_GET_USER(&chunk->size)>=sizeof(SamMMTSBank))
	SAM_MEMCPY_FROMFS(card->mem.sbank+id, cp, sizeof(SamMMTSBank));
      else {
#ifdef DEBUG
	printk(KERN_DEBUG "sam9407: PRIO chunk too small\n");
#endif
	error=-EINVAL;
      }
      break;
    case SAM_CKID_IMAP:
      imapPtr=cp;
      imapSize=SAM_GET_USER(&chunk->size);
      break;
    case SAM_CKID_RLOC:
      reloc=(RelocEntry *)cp;
      nReloc=SAM_GET_USER(&chunk->size)/sizeof(RelocEntry);
      break;
    case SAM_CKID_PARA:
      paramStartAddr=cp-(unsigned char *)SAM_GET_USER(&bank->data);
      paramBytes=SAM_GET_USER(&chunk->size);
      if(paramBuf) {
	vfree(paramBuf);
	paramBuf=NULL;
      }
      if(paramBytes>0) {
	if((paramBuf=vmalloc(paramBytes))) {
	  SAM_MEMCPY_FROMFS(paramBuf, cp, paramBytes);
	  error=allocBlocks(card, SAM_MMT_ID_MIDI_PAR|(id<<8), NULL, paramBytes>>1, SAM_BOUNDARY_64K,
			    SAM_PARAM_PAGE_WORDS, SAM_PARAM_BLOCKS, paramAddrs, &nParamBlocks);
	} else
	  error=-ENOMEM;
      } else
	nParamBlocks=0;
      break;
    case SAM_CKID_SAMP:
      error=allocBlocks(card, SAM_MMT_ID_MIDI_PCM|(id<<8), (unsigned short *)cp, SAM_GET_USER(&chunk->size)>>1, SAM_BOUNDARY_256K,
			SAM_SAMPLE_PAGE_WORDS, SAM_SAMPLE_BLOCKS, sampleAddrs, &nSampleBlocks);
      break;
    default:
#ifdef DEBUG
      printk(KERN_DEBUG "sam9407: unknown RIFF chunk 0x%08x\n", SAM_GET_USER(&chunk->id));
#endif
      error=-EINVAL;
      break;
    }
    if(error<0)
      break;
    size-=sizeof(RiffChunk)+SAM_GET_USER(&chunk->size);
    chunk=(RiffChunk *)((char *)chunk+sizeof(RiffChunk)+SAM_GET_USER(&chunk->size));
  }

  if(error>=0 && reloc && paramBytes>0) {
    /* relocate the param blocks */
    for(i=0; error>=0 && i<nReloc; i++) {
      block=SAM_GET_USER(&reloc[i].block)&0x7F;
      addr=SAM_GET_USER(&reloc[i].addr)-paramStartAddr;
      switch(SAM_GET_USER(&reloc[i].type)) {
      case SAM_RELOC_TYPE_PARAM:
	if(addr+2<=paramBytes) {
	  if(block+1==nParamBlocks)
	    *(unsigned short *)(paramBuf+addr)+=paramAddrs[block]&0xFFFF;
	} else {
#ifdef DEBUG
	  printk(KERN_DEBUG "sam9407: relocation address out of range\n");
#endif
	  error=-EINVAL;
	}
	break;
      case SAM_RELOC_TYPE_MA2_PG:
	if(addr+6<=paramBytes) {
	  if(block<nSampleBlocks)
	    *(unsigned short *)(paramBuf+addr+4)=
	      (*(unsigned short *)(paramBuf+addr+4)&0xE03f)|((sampleAddrs[block]>>12)&~0x3f);
	} else {
#ifdef DEBUG
	  printk(KERN_DEBUG "sam9407: relocation address out of range\n");
#endif
	  error=-EINVAL;
	}
	break;
      case SAM_RELOC_TYPE_MA2_LOOP:
	if(block+1==nSampleBlocks)
	  if(addr+4<=paramBytes) {
	    v=(((unsigned short *)(paramBuf+addr))[0]<<2)|(((unsigned short *)(paramBuf+addr))[1]>>14);
	    v+=sampleAddrs[block]&0x3FFFF;
	    ((unsigned short *)(paramBuf+addr))[0]=v>>2;
	    ((unsigned short *)(paramBuf+addr))[1]&=0x3FFF;
	    ((unsigned short *)(paramBuf+addr))[1]|=(v&3)<<14;
	  } else {
#ifdef DEBUG
	    printk(KERN_DEBUG "sam9407: relocation address out of range\n");
#endif
	    error=-EINVAL;
	  }
	break;
      case SAM_RELOC_TYPE_MB_END:
	if(block+1==nSampleBlocks)
	  if(addr+4<=paramBytes) {
	    v=(((unsigned short *)(paramBuf+addr))[1]<<2)|(((unsigned short *)(paramBuf+addr))[0]>>14);
	    v+=sampleAddrs[block]&0x3FFFF;
	    ((unsigned short *)(paramBuf+addr))[1]=v>>2;
	    ((unsigned short *)(paramBuf+addr))[0]&=0x3FFF;
	    ((unsigned short *)(paramBuf+addr))[0]|=(v&3)<<14;
	  } else {
#ifdef DEBUG
	    printk(KERN_DEBUG "sam9407: relocation address out of range\n");
#endif
	    error=-EINVAL;
	  }
	break;
      default:
#ifdef DEBUG
	printk(KERN_DEBUG "sam9407: unknown relocation type %d\n", SAM_GET_USER(&reloc[i].type));
#endif
	error=-EINVAL;
      }
    }
  }

  if(error>=0 && paramBytes>0) {
    /* transfer the param blocks */
    for(i=0; error>=0 && i<nParamBlocks; i++) {
      ofs=i*SAM_PARAM_PAGE_WORDS;
      n=(paramBytes>>1)-ofs;
      if(n>SAM_PARAM_PAGE_WORDS)
	n=SAM_PARAM_PAGE_WORDS;
      error=samTalk(card, SamWriteMem, paramAddrs[i], (unsigned short *)paramBuf+ofs, n, SamEndOfScript);
    }
  }

  if(paramBuf)
    vfree(paramBuf);

  if(error>=0 && imapPtr && imapSize && *card->mem.sbank[id].name)
    error=adjustMidiTable(card, id, imapPtr, imapSize, paramAddrs);

  if(error<0)
    disposeBank(card, id);

  return error;
}

int samSBankUnload(SamCard *card, unsigned char *id)
{
  int error;
  if((error=verify_area(VERIFY_READ, id, sizeof(unsigned char)))<0)
    return error;
  if(SAM_GET_USER(id)>=SAM_NROF_SBANKS) {
#ifdef DEBUG
    printk(KERN_DEBUG "sam9407: sounbank id out of range\n");
#endif
    return -EINVAL;
  }
  return disposeBank(card, SAM_GET_USER(id));
}

int samSBankPrioChange(SamCard *card, SamSoundBankPrio *prio)
{
  int error;
  SamMMTSBank *sbank;
  if((error=verify_area(VERIFY_READ, prio, sizeof(SamSoundBankPrio)))<0)
    return error;
  if(SAM_GET_USER(&prio->id)>=SAM_NROF_SBANKS) {
#ifdef DEBUG
    printk(KERN_DEBUG "sam9407: sounbank id out of range\n");
#endif
    return -EINVAL;
  }
  sbank=card->mem.sbank+SAM_GET_USER(&prio->id);
  if(!*sbank->name) {
#ifdef DEBUG
    printk(KERN_DEBUG "sam9407: sounbank %d not loaded\n", SAM_GET_USER(&prio->id));
#endif
    return -EINVAL;
  }
  sbank->priority=SAM_GET_USER(&prio->priority);
  return adjustMidiTable(card, -1, NULL, 0, NULL);
}

int samSBankInfo(SamCard *card, SamSoundBankInfo *info)
{
  int error, i;
  if((error=verify_area(VERIFY_WRITE, info, SAM_NROF_SBANKS*sizeof(SamSoundBankInfo)))<0)
    return error;
  for(i=0; i<SAM_NROF_SBANKS; i++) {
    SAM_MEMCPY_TOFS(info[i].name, card->mem.sbank[i].name, SAM_SBANK_NAME_LEN);
    SAM_PUT_USER(0, info[i].name+SAM_SBANK_NAME_LEN);
    SAM_PUT_USER(card->mem.sbank[i].priority, &info[i].priority);
  }
  return 0;
}
