/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/soundcard.h>

#include "driver.h"
#include "io.h"
#include "alloc.h"
#include "midi.h"
#include "seq.h"

//#define SAM_BC_WRITE_DELIVERS_ALL

#define SAM_DEFAULT_TIMEBASE		100
#define SAM_MIN_TIMEBASE		1
#define SAM_MAX_TIMEBASE		1000
#define SAM_DEFAULT_TEMPO		60
#define SAM_MIN_TEMPO			8
#define SAM_MAX_TEMPO			360

typedef struct {
  unsigned char msgs[SAM_STORED_MIDI_EVENTS];
  int nMsgs;
} SamTmpMidiBuffer;

typedef struct {
  struct timer_list timer;
  unsigned references;
} SeqTimerInfo;

static void seqTicker(unsigned long data);

static SeqTimerInfo timerInfo={
  {
    NULL, NULL, 0, 0, seqTicker
  }
};

static void seqTicker(unsigned long data)
{
  unsigned flags;
  int i;
  SamCard *card;

  for(i=0; i<samCardsNr; i++) {
    card=samCards[i];

    if(card->seq.rec.timerRunning)
      atomic_inc(&card->seq.rec.timerTicks);

    if(card->seq.play.timerRunning) {
      atomic_inc(&card->seq.play.timerTicks);
      save_flags(flags); cli();
      if(card->seq.sleeping &&
	 SAM_ATOMIC_READ(&card->seq.play.timerTicks)>=SAM_ATOMIC_READ(&card->seq.wakeupTicks))
	wake_up(&samFeedThreadInfo.feedSam);
      restore_flags(flags);
    }
  }

  save_flags(flags); cli();
  if(timerInfo.references) {
    timerInfo.timer.expires=jiffies+1;
    add_timer(&timerInfo.timer);
  }
  restore_flags(flags);
}

static void timerStop(SamSeqBuffer *seqBuffer)
{
  unsigned flags;

  if(!seqBuffer->timerRunning)
    return;

  seqBuffer->timerRunning=0;

  save_flags(flags); cli();
  if(--timerInfo.references==0)
    del_timer(&timerInfo.timer);
  restore_flags(flags);
}

static void timerStart(SamSeqBuffer *seqBuffer, int reset)
{
  unsigned flags;

  if(!reset && seqBuffer->timerRunning)
    return;

  timerStop(seqBuffer);
  if(reset)
    SAM_ATOMIC_SET(&seqBuffer->timerTicks,0);

  seqBuffer->timerRunning=1;

  save_flags(flags); cli();
  if(timerInfo.references++==0) {
    timerInfo.timer.expires=jiffies+1;
    add_timer(&timerInfo.timer);
  }
  restore_flags(flags);
}

static int putInt(unsigned long arg, int value)
{
  int error;
  if(value<0)
    return value;
  if((error=verify_area(VERIFY_WRITE, (void *)arg, sizeof(int)))<0)
    return error;
  SAM_PUT_USER(value, (int *)arg);
  return 0;
}

static void appendRecEvent(SamCard *card, unsigned char *buf)
{
  unsigned flags;
  int i;

  /* we need to disable IRQ's here
   * as appendRecEvent might be called from within
   * an interrupt handler
   */
  save_flags(flags); cli();
  if(card->seq.rec.usedCount<SAM_STORED_MIDI_EVENTS) {
    i=(card->seq.rec.usedIdx+card->seq.rec.usedCount)%SAM_STORED_MIDI_EVENTS;
    memcpy(card->seq.rec.ringBuffer+i*SAM_SEQ_EVENT_SIZE, buf, *buf&0x80 ? 8 : 4);
    card->seq.rec.usedCount++;
    restore_flags(flags);
  } else
    card->seq.recOverflow=1;

  wake_up(&card->seq.dataAvail);

  restore_flags(flags);
}

static int retrieveRecEventFS(SamCard *card, unsigned char *buf)
{
  unsigned flags;
  unsigned char *event;
  int size;

  save_flags(flags); cli();
  if(card->seq.rec.usedCount) {
    event=card->seq.rec.ringBuffer+card->seq.rec.usedIdx*SAM_SEQ_EVENT_SIZE;
    size=*event&0x80 ? 8 : 4;
    SAM_MEMCPY_TOFS(buf, event, size);
    card->seq.rec.usedIdx=(card->seq.rec.usedIdx+1)%SAM_STORED_MIDI_EVENTS;
    card->seq.rec.usedCount--;
    restore_flags(flags);
    return size;
  }

  restore_flags(flags);
  return 0;
}

static void flushBuffers(SamCard *card, SamTmpMidiBuffer *buffers)
{
  int i;

  for(i=0; i<SamMidiPortSentinel; i++) {
    if(buffers[i].nMsgs>0) {
      if(samTalk(card,
		 SamSendCtrl, 1, samMidiSelCtrls[i],
		 SamSendMidi, buffers[i].msgs, buffers[i].nMsgs,
		 SamEndOfScript)<0)
	printk(KERN_ERR "sam9407: error writing midi data\n");
      buffers[i].nMsgs=0;
    }
  }
}

static int resetSeq(SamCard *card, int recording)
{
  int error=0, tmpError, i;

  timerStop(&card->seq.play);
  timerStop(&card->seq.rec);

  SAM_ATOMIC_SET(&card->seq.play.timerTicks,0);
  SAM_ATOMIC_SET(&card->seq.rec.timerTicks,0);

  card->seq.sleeping=0;
  card->seq.recording=0;

  SAM_ATOMIC_SET(&card->seq.lastWaitTimerTicks,0);
  card->seq.lastRecTimer=0;
  card->seq.timeBase=SAM_DEFAULT_TIMEBASE;
  card->seq.tempo=SAM_DEFAULT_TEMPO;
  
  card->seq.play.usedIdx=card->seq.play.usedCount=0;
  card->seq.rec.usedIdx=card->seq.rec.usedCount=0;
  card->seq.recOverflow=card->seq.followupRecEvent=0;

  if(recording) {
    card->seq.recording=1;
    timerStart(&card->seq.rec, 1);
  } else
    card->seq.recording=0;

  for(i=0; i<SamMidiPortSentinel; i++)
    if((tmpError=samResetMidiPort(card, i))<0 && !error)
      error=tmpError;

  return error;
}

/* asynchronous playback of
 * sequencer events stored at card->seq.ringBuffer
 */
void samCheckPlaybackSeq(SamCard *card)
{
  int i;
  unsigned wakeupTicks, v;
  unsigned char *event;
  SamTmpMidiBuffer buffers[SamMidiPortSentinel], *buffer;

  down(&card->midi.accessLock);
  if(card->seq.sleeping) {
    if(SAM_ATOMIC_READ(&card->seq.play.timerTicks)<SAM_ATOMIC_READ(&card->seq.wakeupTicks)) {
      up(&card->midi.accessLock);
      return;
    }
    card->seq.sleeping=0;
  }
  if(card->seq.play.usedCount) {
    for(i=0; i<SamMidiPortSentinel; i++)
      buffers[i].nMsgs=0;
    while(!card->seq.sleeping && card->seq.play.usedCount) {
      event=card->seq.play.ringBuffer+card->seq.play.usedIdx*SAM_SEQ_EVENT_SIZE;
      /* we've got one event,
       * process it
       */
      switch(event[0]) {
      case SEQ_MIDIPUTC:
	if(event[2]<SamMidiPortSentinel) {
	  buffer=buffers+event[2];
	  if(buffer->nMsgs>=sizeof(buffers->msgs))
	    flushBuffers(card, buffers);
	  buffer->msgs[buffer->nMsgs++]=event[1];
	}
	break;
      case EV_TIMING:
	flushBuffers(card, buffers);
	switch(event[1]) {
	case TMR_START:
	  timerStart(&card->seq.play, 1);
	  SAM_ATOMIC_SET(&card->seq.lastWaitTimerTicks,0);
	  break;
	case TMR_STOP:
	  timerStop(&card->seq.play);
	  break;
	case TMR_CONTINUE:
	  timerStart(&card->seq.play, 0);
	  break;
	case TMR_WAIT_ABS:
	case TMR_WAIT_REL:
	  flushBuffers(card, buffers);
	  wakeupTicks=(*(unsigned *)(event+4))*60*100/card->seq.timeBase/card->seq.tempo;
	  if(event[1]==TMR_WAIT_REL)
	    wakeupTicks+=SAM_ATOMIC_READ(&card->seq.lastWaitTimerTicks);
	  if(wakeupTicks>SAM_ATOMIC_READ(&card->seq.play.timerTicks)) {
	    SAM_ATOMIC_SET(&card->seq.wakeupTicks,wakeupTicks);
	    card->seq.sleeping=1;
	  }
	  card->seq.lastWaitTimerTicks=card->seq.play.timerTicks;
	  break;
	case TMR_ECHO:
	  appendRecEvent(card, event);
	  break;
	case TMR_TEMPO:
	  v=*(unsigned *)(event+4);
	  if(v<SAM_MIN_TEMPO)
	    v=SAM_MIN_TEMPO;
	  if(v>SAM_MAX_TEMPO)
	    v=SAM_MAX_TEMPO;
	  card->seq.tempo=v;
	  break;
	}
	break;
      }
      card->seq.play.usedIdx=(card->seq.play.usedIdx+1)%SAM_STORED_MIDI_EVENTS;
      card->seq.play.usedCount--;
      wake_up(&card->seq.moreSpace);
    }
    flushBuffers(card, buffers);
  }
  up(&card->midi.accessLock);
}

/* handle extern midi events */
void samMidiInputSeq(SamCard *card, unsigned v)
{
  unsigned flags;
  int newRecTimer;
  unsigned char event[8];

  save_flags(flags); cli();
  if(!card->seq.recording) {
    restore_flags(flags);
    return;
  }

  if(!card->seq.followupRecEvent) {
    memset(event, 0, sizeof(event));
    event[0]=EV_TIMING;
    event[1]=TMR_START;
    appendRecEvent(card, event);

    card->seq.followupRecEvent=1;
  }

  newRecTimer=SAM_ATOMIC_READ(&card->seq.rec.timerTicks)*SAM_DEFAULT_TEMPO*SAM_DEFAULT_TIMEBASE/60/100;
  if(newRecTimer>card->seq.lastRecTimer) {
    memset(event, 0, sizeof(event));
    event[0]=EV_TIMING;
    event[1]=TMR_WAIT_ABS;
    *(unsigned *)(event+4)=newRecTimer;
    appendRecEvent(card, event);

    card->seq.lastRecTimer=newRecTimer;
  }

  memset(event, 0, sizeof(event));
  event[0]=SEQ_MIDIPUTC;
  event[1]=v;
  event[2]=SamMidiPortExternal;
  appendRecEvent(card, event);

  restore_flags(flags);
}

int samOpenSeq(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  int error;

  down(&card->midi.accessLock);
  if(card->midi.ports[SamMidiPortExternal].owner!=SamMidiPortOwnerNone ||
     card->midi.ports[SamMidiPortInternal].owner!=SamMidiPortOwnerNone) {
    up(&card->midi.accessLock);
    return -EBUSY;
  }

  card->midi.ports[SamMidiPortExternal].owner=SamMidiPortOwnerSeq;
  card->midi.ports[SamMidiPortInternal].owner=SamMidiPortOwnerSeq;

  if((error=resetSeq(card, file->f_mode & SAM_MODE_READ_MASK))<0) {
    card->midi.ports[SamMidiPortExternal].owner=SamMidiPortOwnerNone;
    card->midi.ports[SamMidiPortInternal].owner=SamMidiPortOwnerNone;
    up(&card->midi.accessLock);
    return error;
  }

  up(&card->midi.accessLock);
  return 0;
}

void samReleaseSeq(struct inode *inode, struct file *file, SamCard *card, mode_t mode, char *closure)
{
  unsigned flags;

  /* linux-2.2.0-pre4 requires current->sigpending to be recalculated
   * whenever a signal was received, that signal wasn't available
   * in release anymore.
   */
#if defined(SAM_LINUX22_SIGPENDING_BUG) && LINUX_VERSION_CODE>=0x20200
  spin_lock_irq(&current->sigmask_lock);
  recalc_sigpending(current);
  spin_unlock_irq(&current->sigmask_lock);
#endif /* SAM_LINUX22_SIGPENDING_BUG */

  save_flags(flags); cli();
  while(card->seq.play.usedCount) {
    interruptible_sleep_on(&card->seq.moreSpace);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current))
      break;
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked)
      break;
#endif /* Linux 2.0 */
  }
  restore_flags(flags);

  down(&card->midi.accessLock);

  if(resetSeq(card, 0)<0)
    printk(KERN_ERR "sam9407: couldn't reset midi channel\n");

  card->midi.ports[SamMidiPortExternal].owner=SamMidiPortOwnerNone;
  card->midi.ports[SamMidiPortInternal].owner=SamMidiPortOwnerNone;

  up(&card->midi.accessLock);
}

int samReadSeq(struct inode *inode, struct file *file, char *buf, int count, SamCard *card, char *closure)
{
  unsigned flags;
  int error, n, size;
  unsigned char *cp;

  if(count<0 || (count>0 && count<8))
    return -EINVAL;

  if((error=verify_area(VERIFY_WRITE, buf, count))<0)
    return error;

  cp=(unsigned char *)buf;
  n=count;

  down(&card->midi.accessLock);

  if(!card->seq.recOverflow && !card->seq.rec.usedCount && (file->f_flags & O_NONBLOCK)) {
    up(&card->midi.accessLock);
    return -EAGAIN;
  }

  if(card->seq.recOverflow) {
    save_flags(flags); cli();
    card->seq.recOverflow=0;
    card->seq.rec.usedIdx=card->seq.rec.usedCount=0;
    restore_flags(flags);
    up(&card->midi.accessLock);
    return -EIO;
  }

  /* wait for incoming data */
  save_flags(flags); cli();
  while(!card->seq.recOverflow && !card->seq.rec.usedCount) {
    up(&card->midi.accessLock);
    interruptible_sleep_on(&card->seq.dataAvail);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current)) {
      restore_flags(flags);
      return -ERESTARTSYS;
    }
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked) {
      restore_flags(flags);
      return -ERESTARTSYS;
    }
#endif /* Linux 2.0 */
    down(&card->midi.accessLock);
    cli();
  }
  restore_flags(flags);

  while(n>=8) {
    if(!(size=retrieveRecEventFS(card, cp)))
      break;
    cp+=size;
    n-=size;
  }

  up(&card->midi.accessLock);

  return cp-(unsigned char *)buf;
}

int samWriteSeq(struct inode *inode, struct file *file, const char *buf, int count, SamCard *card, char *closure)
{
  unsigned flags;
  int error, i, n, size;
  unsigned char *event, *cp;

  if(count<0)
    return -EINVAL;

  if((error=verify_area(VERIFY_READ, buf, count))<0)
    return error;

  cp=(unsigned char *)buf;
  n=count;

  down(&card->midi.accessLock);

  if(card->seq.play.usedCount+SAM_STORED_MIDI_EVENTS/2>SAM_STORED_MIDI_EVENTS &&
     (file->f_flags & O_NONBLOCK)) {
    up(&card->midi.accessLock);
    return -EAGAIN;
  }

#ifdef SAM_BC_WRITE_DELIVERS_ALL
 deliver_again:
#endif

  /* wait until at least half SAM_STORED_MIDI_EVENTS
   * buffer is free
   * to avoid writing small chunks only
   */
  while(card->seq.play.usedCount+SAM_STORED_MIDI_EVENTS/2>SAM_STORED_MIDI_EVENTS) {
    save_flags(flags); cli();
    up(&card->midi.accessLock);
    interruptible_sleep_on(&card->seq.moreSpace);
    restore_flags(flags);
#if LINUX_VERSION_CODE>=0x20200
    if(signal_pending(current))
      return -ERESTARTSYS;
#else /* Linux 2.0 */
    if(current->signal & ~current->blocked)
      return -ERESTARTSYS;
#endif /* Linux 2.0 */
    down(&card->midi.accessLock);
  }

  while(n>0 && card->seq.play.usedCount<SAM_STORED_MIDI_EVENTS) {
    i=(card->seq.play.usedIdx+card->seq.play.usedCount)%SAM_STORED_MIDI_EVENTS;
    event=card->seq.play.ringBuffer+i*SAM_SEQ_EVENT_SIZE;

    size=SAM_GET_USER(cp)&0x80 ? 8 : 4;
    if(n<size) {
      up(&card->midi.accessLock);
      return -EINVAL;
    }

    SAM_MEMCPY_FROMFS(event, cp, size);
    cp+=size;
    n-=size;
      
    card->seq.play.usedCount++;
  }

#ifdef SAM_BC_WRITE_DELIVERS_ALL
  if(n>0)
    goto deliver_again;
#endif

  up(&card->midi.accessLock);
  wake_up(&samFeedThreadInfo.feedSam);

  return cp-(unsigned char *)buf;
}

#if LINUX_VERSION_CODE>=0x20200
unsigned samPollSeq(struct inode *inode, struct file *file, poll_table *wait, SamCard *card, mode_t mode, char *closure)
{
  unsigned mask, flags;

  poll_wait(file, &card->seq.dataAvail, wait);
  poll_wait(file, &card->seq.moreSpace, wait);

  mask=0;
  save_flags(flags); cli();
  if(card->seq.recOverflow || card->seq.rec.usedCount)
    mask|= POLLIN | POLLRDNORM;

  if(card->seq.play.usedCount+SAM_STORED_MIDI_EVENTS/2<=SAM_STORED_MIDI_EVENTS)
    mask|= POLLOUT | POLLWRNORM;

  restore_flags(flags);

  return mask;
}
#else /* Linux 2.0 */
int samSelectSeq(struct inode *inode, struct file *file, int type, select_table *wait, SamCard *card, char *closure)
{
  unsigned flags;

  switch(type) {
  case SEL_IN:
    save_flags(flags); cli();
    if(card->seq.recOverflow || card->seq.rec.usedCount) {
      restore_flags(flags);
      return 1;
    }
    select_wait(&card->seq.dataAvail, wait);
    restore_flags(flags);
    break;
  case SEL_OUT:
    down(&card->midi.accessLock);
    if(card->seq.play.usedCount+SAM_STORED_MIDI_EVENTS/2<=SAM_STORED_MIDI_EVENTS) {
      up(&card->midi.accessLock);
      return 1;
    }
    save_flags(flags); cli();
    up(&card->midi.accessLock);
    select_wait(&card->seq.moreSpace, wait);
    restore_flags(flags);
    break;
  }
  return 0;
}
#endif /* Linux 2.0 */

int samIoctlSeq(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg, SamCard *card, mode_t mode, char *closure, int callNr)
{
  int error, v;

  error=0;
  switch(cmd) {
  case SNDCTL_SEQ_NRSYNTHS:
    return putInt(arg, 0);
  case SNDCTL_SEQ_NRMIDIS:
    return putInt(arg, 2);
  case SNDCTL_SEQ_RESET:
    down(&card->midi.accessLock);
    error=resetSeq(card, file->f_mode & SAM_MODE_READ_MASK);
    up(&card->midi.accessLock);
    break;
  case SNDCTL_TMR_START:
    down(&card->midi.accessLock);
    timerStart(&card->seq.play, 1);
    SAM_ATOMIC_SET(&card->seq.lastWaitTimerTicks,0);
    up(&card->midi.accessLock);
    break;
  case SNDCTL_TMR_STOP:
    down(&card->midi.accessLock);
    timerStop(&card->seq.play);
    up(&card->midi.accessLock);
    break;
  case SNDCTL_TMR_CONTINUE:
    down(&card->midi.accessLock);
    timerStart(&card->seq.play, 0);
    up(&card->midi.accessLock);
    break;
  case SNDCTL_TMR_TIMEBASE:
    if((error=verify_area(VERIFY_READ, (void *)arg, sizeof(int)))<0)
      return error;
    v=SAM_GET_USER((int *)arg);
    if(v<SAM_MIN_TIMEBASE)
      v=SAM_MIN_TIMEBASE;
    if(v>SAM_MAX_TIMEBASE)
      v=SAM_MAX_TIMEBASE;
    down(&card->midi.accessLock);
    card->seq.timeBase=v;
    up(&card->midi.accessLock);
    break;
  case SNDCTL_TMR_TEMPO:
    v=SAM_GET_USER((int *)arg);
    if(v<SAM_MIN_TEMPO)
      v=SAM_MIN_TEMPO;
    if(v>SAM_MAX_TEMPO)
      v=SAM_MAX_TEMPO;
    down(&card->midi.accessLock);
    card->seq.tempo=v;
    up(&card->midi.accessLock);
    break;
  default:
    return -ENOSYS;
  }

  return error;
}
