/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <tcl.h>

#include <sam9407.h>
#include <mixerctl.h>

#define NumberOf(vec) (sizeof(vec)/sizeof(*(vec)))

#ifndef MAP_FAILED
#define MAP_FAILED ((void *)-1)
#endif

static Tcl_HashTable optionsByName, optionsById, doesResetByName;
static int mixerFd;
static Tcl_DString notifyCmd, sbankNotifyCmd;

static void mixerNotify(ClientData clientData, int mask)
{
  Tcl_Interp *interp=(Tcl_Interp *)clientData;
  char dirtyIds[SAM_MIXER_ID_LAST+2];
  Tcl_DString cmd;
  Tcl_HashEntry *entry;
  int n, i;
  if((n=read(mixerFd, dirtyIds, sizeof(dirtyIds)))<0) {
    fprintf(stderr, "fatal: can't read dirty flags from /dev/mixer\n");
    exit(1);
  }
  if(*Tcl_DStringValue(&notifyCmd)) {
    Tcl_DStringInit(&cmd);
    Tcl_DStringAppendElement(&cmd, Tcl_DStringValue(&notifyCmd));
    Tcl_DStringStartSublist(&cmd);
  }
  for(i=0; i<n; i++) {
    if(dirtyIds[i]<0) {
      switch(dirtyIds[i]) {
      case SAM_MIXER_SBANK_CHANGED:
	if(*Tcl_DStringValue(&sbankNotifyCmd)) {
	  if(Tcl_Eval(interp, Tcl_DStringValue(&sbankNotifyCmd))==TCL_ERROR) {
	    fprintf(stderr, "warning: sbank notification \"%s\" failed: %s\n", Tcl_DStringValue(&sbankNotifyCmd), interp->result);
	  }
	}
	break;
      }
    } else
      if(*Tcl_DStringValue(&notifyCmd) && (entry=Tcl_FindHashEntry(&optionsById, (char *)(unsigned)dirtyIds[i])))
	Tcl_DStringAppendElement(&cmd, Tcl_GetHashValue(entry));
  }
  if(*Tcl_DStringValue(&notifyCmd)) {
    Tcl_DStringEndSublist(&cmd);
    if(Tcl_Eval(interp, Tcl_DStringValue(&cmd))==TCL_ERROR) {
      fprintf(stderr, "warning: mixer notification \"%s\" failed: %s\n", Tcl_DStringValue(&cmd), interp->result);
    }
    Tcl_DStringFree(&cmd);
  }
}

static int parseOptions(int set, Tcl_Interp *interp, int argc, char **argv, SamMixerValue *mixerValues, int nMixerValues)
{
  Tcl_HashEntry *entry;
  int n, i;

  n=0;
  for(i=2; i<argc;) {
    if(!(entry=Tcl_FindHashEntry(&optionsByName, argv[i]))) {
      Tcl_AppendResult(interp, "mixer: unknown mixer option: \"", argv[i], "\"", NULL);
      return -1;
    }
    
    i++;
    if(n+1>=nMixerValues) {
      Tcl_AppendResult(interp, "mixer: too many options", NULL);
      return -1;
    }
    mixerValues[n].id=(SamMixerId)Tcl_GetHashValue(entry);
    if(set) {
      mixerValues[n].value=strtol(argv[i++], NULL, 0);
    } else {
      mixerValues[n].value=0;
    }
    n++;
  }
  if(n>=nMixerValues) {
    Tcl_AppendResult(interp, "mixer: too many options", NULL);
    return -1;
  }
  mixerValues[n].id=SAM_MIXER_ID_SENTINEL;
  return n;
}

static int mixerCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
  SamMixerValue mixerValues[SAM_MIXER_ID_LAST+1];
  char buf[32];
  Tcl_HashEntry *entry;
  Tcl_DString dstr;
  int n, i;

  if(argc>1 && strcmp(argv[1], "get")==0) {
    if((n=parseOptions(0, interp, argc, argv, mixerValues, NumberOf(mixerValues)))<0)
      return TCL_ERROR;
    if(ioctl(mixerFd, SAM_IOC_MIXER_GET, mixerValues)<0) {
      Tcl_AppendResult(interp, "mixer: can't retrieve mixer values: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    Tcl_DStringInit(&dstr);
    for(i=0; i<n; i++) {
      sprintf(buf, "%d", mixerValues[i].value);
      Tcl_DStringAppendElement(&dstr, buf);
    }
    Tcl_DStringResult(interp, &dstr);
    Tcl_DStringFree(&dstr);
  } else if(argc>1 && strcmp(argv[1], "set")==0) {
    if((n=parseOptions(1, interp, argc, argv, mixerValues, NumberOf(mixerValues)))<0)
      return TCL_ERROR;
    if(ioctl(mixerFd, SAM_IOC_MIXER_SET, mixerValues)<0) {
      Tcl_AppendResult(interp, "mixer: can't set mixer values: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    Tcl_ResetResult(interp);
  } else if(argc>1 && strcmp(argv[1], "info")==0) {
    for(i=0; samMixerAssoc[i].name; i++) {
      mixerValues[i].id=samMixerAssoc[i].id;
      mixerValues[i].value=0;
    }
    mixerValues[i].id=SAM_MIXER_ID_SENTINEL;
    if(ioctl(mixerFd, SAM_IOC_MIXER_INFO, mixerValues)<0) {
      Tcl_AppendResult(interp, "mixer: can't get mixer info: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    Tcl_DStringInit(&dstr);
    for(i=0; samMixerAssoc[i].name; i++) {
      if(mixerValues[i].value)
	Tcl_DStringAppendElement(&dstr, samMixerAssoc[i].name);
    }
    Tcl_DStringResult(interp, &dstr);
  } else if(argc>1 && strcmp(argv[1], "does_reset")==0) {
    Tcl_DStringInit(&dstr);
    for(i=2; i<argc; i++) {
      if(!(entry=Tcl_FindHashEntry(&doesResetByName, argv[i]))) {
	Tcl_DStringFree(&dstr);
	Tcl_AppendResult(interp, "mixer: unknown mixer option: \"", argv[i], "\"", NULL);
	return TCL_ERROR;
      }
      Tcl_DStringAppendElement(&dstr, Tcl_GetHashValue(entry) ? "1" : "0");
    }
    Tcl_DStringResult(interp, &dstr);
    Tcl_DStringFree(&dstr);
  } else if(argc==3 && strcmp(argv[1], "notify")==0) {
    Tcl_DStringSetLength(&notifyCmd, 0);
    Tcl_DStringAppend(&notifyCmd, argv[2], -1);
  } else {
    Tcl_AppendResult(interp, "invalid subcommand: should be \"", argv[0], " {info|get|set|does_reset|notify} ...\"", NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static int sbankCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
  int fd, i;
  SamSoundBank soundBank;
  SamSoundBankInfo info[SAM_NROF_SBANKS];
  SamSoundBankPrio prio;
  struct stat st;
  Tcl_DString dstr;
  char buf[16];

  if(argc>3 && strcmp(argv[1], "load")==0) {
    soundBank.id=strtol(argv[2], NULL, 0);
    if((fd=open(argv[3], O_RDONLY))<0) {
      Tcl_AppendResult(interp, "sbank: can't open soundbank: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    if(fstat(fd, &st)<0) {
      Tcl_AppendResult(interp, "sbank: can't stat soundbank: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    if((soundBank.data=(char *)mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0))==MAP_FAILED) {
      close(fd);
      Tcl_AppendResult(interp, "sbank: can't mmap soundbank: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    soundBank.size=st.st_size;
    if(ioctl(mixerFd, SAM_IOC_BANK_LOAD, &soundBank)<0) {
      munmap(soundBank.data, st.st_size);
      close(fd);
      Tcl_AppendResult(interp, "sbank: can't load soundbank: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    munmap(soundBank.data, st.st_size);
    close(fd);
  } else if(argc>2 && strcmp(argv[1], "unload")==0) {
    soundBank.id=strtol(argv[2], NULL, 0);
    if(ioctl(mixerFd, SAM_IOC_BANK_UNLOAD, &soundBank.id)<0) {
      Tcl_AppendResult(interp, "sbank: can't unload soundbank: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
  } else if(argc>3 && strcmp(argv[1], "prio")==0) {
    prio.id=strtol(argv[2], NULL, 0);
    prio.priority=strtol(argv[3], NULL, 0);
    if(ioctl(mixerFd, SAM_IOC_BANK_PRIO_CHG, &prio)<0) {
      Tcl_AppendResult(interp, "sbank: can't change soundbank priority: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
  } else if(argc>1 && strcmp(argv[1], "info")==0) {
    if(ioctl(mixerFd, SAM_IOC_BANK_INFO, &info)<0) {
      Tcl_AppendResult(interp, "sbank: can't query soundbank info: ", strerror(errno), NULL);
      return TCL_ERROR;
    }
    Tcl_DStringInit(&dstr);
    for(i=0; i<NumberOf(info); i++) {
      Tcl_DStringStartSublist(&dstr);
      Tcl_DStringAppendElement(&dstr, info[i].name);
      sprintf(buf, "%d", info[i].priority);
      Tcl_DStringAppendElement(&dstr, buf);
      Tcl_DStringEndSublist(&dstr);
    }
    Tcl_DStringResult(interp, &dstr);
    Tcl_DStringFree(&dstr);
  } else if(argc>2 && strcmp(argv[1], "notify")==0) {
    Tcl_DStringSetLength(&sbankNotifyCmd, 0);
    Tcl_DStringAppend(&sbankNotifyCmd, argv[2], -1);
  } else {
    Tcl_AppendResult(interp, "invalid subcommand: should be \"", argv[0], " {info|load id filename|unload id|prio id prio|notify cmd}\"", NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

int Sam_Init(Tcl_Interp *interp)
{
  int i, newEntry;
  Tcl_HashEntry *entry;

  if((mixerFd=open("/dev/sam0_mixer", O_RDWR))<0 &&
     (mixerFd=open("/dev/mixer", O_RDWR))<0) {
    Tcl_AppendResult(interp, "can't open /dev/mixer: ", strerror(errno), NULL);
    return TCL_ERROR;
  }

  Tcl_InitHashTable(&optionsByName, TCL_STRING_KEYS);
  Tcl_InitHashTable(&optionsById, TCL_ONE_WORD_KEYS);
  Tcl_InitHashTable(&doesResetByName, TCL_STRING_KEYS);
  for(i=0; samMixerAssoc[i].name; i++) {
    entry=Tcl_CreateHashEntry(&optionsByName, samMixerAssoc[i].name, &newEntry);
    Tcl_SetHashValue(entry, (ClientData)samMixerAssoc[i].id);

    entry=Tcl_CreateHashEntry(&optionsById, (char *)samMixerAssoc[i].id, &newEntry);
    Tcl_SetHashValue(entry, (ClientData)samMixerAssoc[i].name);

    entry=Tcl_CreateHashEntry(&doesResetByName, (char *)samMixerAssoc[i].name, &newEntry);
    Tcl_SetHashValue(entry, (ClientData)samMixerAssoc[i].doesReset);
  }

  Tcl_DStringInit(&notifyCmd);
  Tcl_DStringInit(&sbankNotifyCmd);
#if TCL_MAJOR_VERSION<8
  Tcl_CreateFileHandler(Tcl_GetFile(mixerFd, TCL_UNIX_FD), TCL_READABLE, mixerNotify, (ClientData)interp);
#else
  Tcl_CreateFileHandler(mixerFd, TCL_READABLE, mixerNotify, (ClientData)interp);
#endif

  Tcl_CreateCommand(interp, "mixer", (Tcl_CmdProc *)mixerCmd, NULL, NULL);
  Tcl_CreateCommand(interp, "sbank", (Tcl_CmdProc *)sbankCmd, NULL, NULL);

  return TCL_OK;
}
