/*
  Copyright (C) 1998, 1999 Gerd Rausch <gerd@alf.gun.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw3d/Label.h>
#include <X11/Xaw3d/Command.h>
#include <X11/Xaw3d/Toggle.h>
#include <X11/Xaw3d/List.h>
#include <X11/Xaw3d/AsciiText.h>

#include "Tab.h"
#include "Dial.h"
#include "Layout.h"

#include "x9407.h"

struct Pane;

typedef struct {
  char *name;
  String (*retrieve)(void);
  Widget w;
  struct Pane *pane;
} LabelEntry;

typedef struct {
  char *name;
  XtCallbackProc callback;
  Widget w;
  struct Pane *pane;
} CommandEntry;

typedef struct {
  char *name;
  int mixerId, offValue, onValue;
  Widget w;
  struct Pane *pane;
  MixerCallback notify;
} ToggleEntry;

typedef struct {
  char *name;
  void (*retrieve)(String **list, int *numberStrings);
  XtCallbackProc callback;
  Widget w;
  struct Pane *pane;
  int selected;
} ListEntry;

typedef struct TextEntry {
  char *name;
  void (*acknowledge)(Widget w, struct TextEntry *entry);
  Widget w;
  struct Pane *pane;
} TextEntry;

typedef struct {
  char *name;
  int mixerId;
  float min, max, step;
  float multiply, add;
  int relatedDial;
  Widget w;
  struct Pane *pane;
  MixerCallback notify;
} DialEntry;

typedef struct Pane {
  char *name;
  int needsOpenDevice;
  struct Pane *subPanes;
  LabelEntry *labels;
  CommandEntry *commands;
  ToggleEntry *toggles;
  ListEntry *lists;
  TextEntry *texts;
  DialEntry *dials;
  Widget tab, w;
} Pane;

static void selectElementCallback(Widget w, XtPointer closure, XtPointer callData);
static void dialValueCallback(Widget w, XtPointer closure, XtPointer callData);
static void dialValueNotify(int id, MixerCallback *info);
static void toggleValueCallback(Widget w, XtPointer closure, XtPointer callData);
static void toggleValueNotify(int id, MixerCallback *info);
static void ackTextCallback(Widget w, XtPointer closure);

static String retrieveCardNr(void);
static String getMemUsedSBank(void);
static String getMemFree(void);
static void retrieveDevices(String **list, int *numberStrings);
static void retrieveSoundbanks(String **list, int *numberStrings);
static void rescanDevicesCallback(Widget w, XtPointer closure, XtPointer callData);
static void useDeviceCallback(Widget w, XtPointer closure, XtPointer callData);
static void releaseDeviceCallback(Widget w, XtPointer closure, XtPointer callData);
static void loadFirmwareCallback(Widget w, XtPointer closure, XtPointer callData);
static void sbankSelectedCallback(Widget w, XtPointer closure, XtPointer callData);
static void loadSoundbankCallback(Widget w, XtPointer closure, XtPointer callData);
static void unloadSoundbankCallback(Widget w, XtPointer closure, XtPointer callData);
static void changeSBankPriorityCallback(Widget w, TextEntry *entry);

static char textTranslations[]="\
<Key>Return:	ackText()\n\
<Key>Linefeed:	ackText()\n\
:<Key>KP_Enter:	ackText()\
";

/***** Devices pane *****/
static LabelEntry devicesLabels[]={
  {
    "inuse_lbl",
  },
  {
    "inuse",
    retrieveCardNr,
  },
  {
    "title",
  },
  {
    NULL		/* sentinel */
  }
};

static ListEntry devicesLists[]={
  {
    "list",
    retrieveDevices,
  },
  {
    NULL		/* sentinel */
  }
};

static CommandEntry devicesCommands[]={
  {
    "rescan",
    rescanDevicesCallback,
  },
  {
    "use",
    useDeviceCallback,
  },
  {
    "release",
    releaseDeviceCallback,
  },
  {
    "firmload",
    loadFirmwareCallback,
  },
  {
    NULL		/* sentinel */
  }
};

/***** Soundbank pane *****/
static LabelEntry soundbankLabels[]={
  {
    "title",
  },
  {
    "prio_lbl",
  },
  {
    "memsbank_lbl",
  },
  {
    "memsbank",
    getMemUsedSBank,
  },
  {
    "memfree_lbl",
  },
  {
    "memfree",
    getMemFree,
  },
  {
    NULL		/* sentinel */
  }
};

static ListEntry soundbankLists[]={
  {
    "list",
    retrieveSoundbanks,
    sbankSelectedCallback,
  },
  {
    NULL		/* sentinel */
  }
};

static CommandEntry soundbankCommands[]={
  {
    "load",
    loadSoundbankCallback,
  },
  {
    "unload",
    unloadSoundbankCallback,
  },
  {
    NULL		/* sentinel */
  }
};

static TextEntry soundbankTexts[]={
  {
    "prio",
    changeSBankPriorityCallback,
  },
  {
    NULL		/* sentinel */
  }
};

static Pane configPanes[]={
  {
    "devices",		/* name */
    0,			/* needsOpenDevice */
    NULL,		/* subPanes */
    devicesLabels,	/* labels */
    devicesCommands,	/* commands */
    NULL,		/* toggles */
    devicesLists,	/* lists */
    NULL,		/* texts */
    NULL,		/* dials */
  },
  {
    "soundbank",	/* name */
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    soundbankLabels,	/* labels */
    soundbankCommands,	/* commands */
    NULL,		/* toggles */
    soundbankLists,	/* lists */
    soundbankTexts,	/* texts */
    NULL,		/* dials */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Main pane *****/
static LabelEntry mainLabels[]={
  {
    "lbl_post",
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry mainToggles[]={
  {
    "gm_post",
    SAM_MIXER_ID_GM_POST,
    0, 1,
  },
  {
    "wave_post",
    SAM_MIXER_ID_WAVE_POST,
    0, 1,
  },
  {
    "mod_post",
    SAM_MIXER_ID_MOD_POST,
    0, 1,
  },
  {
    "audech_post",
    SAM_MIXER_ID_AUDECH_POST,
    0, 1,
  },
  {
    "eff_post",
    SAM_MIXER_ID_EFF_POST,
    0, 1,
  },
  {
    NULL		/* sentinel */
  }
};

static DialEntry mainDials[]={
  {
    "volume",
    SAM_MIXER_ID_MASTER_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Config pane *****/

static LabelEntry configLabels[]={
  {
    "equ_lbl",
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry configToggles[]={
  {
    "rec_cfg",
    SAM_MIXER_ID_REC_CFG,
    0, 1,
  },
  {
    "aud_onoff",
    SAM_MIXER_ID_AUD_ONOFF,
    0, 1,
  },
  {
    "ech_onoff",
    SAM_MIXER_ID_ECH_ONOFF,
    0, 1,
  },
  {
    "rev_onoff",
    SAM_MIXER_ID_REV_ONOFF,
    0, 1,
  },
  {
    "chr_onoff",
    SAM_MIXER_ID_CHR_ONOFF,
    0, 1,
  },
  {
    "sur_onoff",
    SAM_MIXER_ID_SUR_ONOFF,
    0, 1,
  },
  {
    "wave_ass",
    SAM_MIXER_ID_WAVE_ASS,
    0, 1,
  },
  {
    "mod_ass",
    SAM_MIXER_ID_MOD_ASS,
    0, 1,
  },
  {
    "equ_type_4b",
    SAM_MIXER_ID_EQU_TYPE,
    -1, 0,
  },
  {
    "equ_type_2b",
    SAM_MIXER_ID_EQU_TYPE,
    -1, 1,
  },
  {
    "equ_type_none",
    SAM_MIXER_ID_EQU_TYPE,
    -1, 2,
  },
  {
    NULL		/* sentinel */
  }
};

/***** Midi pane *****/

static DialEntry midiDials[]={
  {
    "gm_vol",
    SAM_MIXER_ID_GM_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "gm_pan",
    SAM_MIXER_ID_GM_PAN,
    -64, 63, 1,		/* min, max, step */
    1, 64,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Recording pane *****/

static DialEntry recordDials[]={
  {
    "rec_vol",
    SAM_MIXER_ID_REC_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry recordToggles[]={
  {
    "rec_mode",
    SAM_MIXER_ID_REC_MODE,
    1, 0,
  },
  {
    NULL		/* sentinel */
  }
};

/***** Audio In pane *****/

static LabelEntry audinLabels[]={
  {
    "aud_gain_lbl",
  },
  {
    "aud_vol_lbl",
  },
  {
    "aud_pan_lbl",
  },
  {
    NULL		/* sentinel */
  }
};

static DialEntry audinDials[]={
  {
    "aud_gainl",
    SAM_MIXER_ID_AUD_GAINL,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "aud_gainr",
    SAM_MIXER_ID_AUD_GAINR,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    "audl_vol",
    SAM_MIXER_ID_AUDL_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "audr_vol",
    SAM_MIXER_ID_AUDR_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    "audl_pan",
    SAM_MIXER_ID_AUDL_PAN,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "audr_pan",
    SAM_MIXER_ID_AUDR_PAN,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry audinToggles[]={
  {
    "aud_sel",
    SAM_MIXER_ID_AUD_SEL,
    0, 1,
  },
  {
    "audchr_send",
    SAM_MIXER_ID_AUDCHR_SEND,
    0, 1,
  },
  {
    "audrev_send",
    SAM_MIXER_ID_AUDREV_SEND,
    0, 1,
  },
  {
    NULL		/* sentinel */
  }
};

static Pane generalPanes[]={
  {
    "main",		/* name */
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    mainLabels,		/* labels */
    NULL,		/* commands */
    mainToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    mainDials,		/* dials */
  },
  {
    "config",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    configLabels,	/* labels */
    NULL,		/* commands */
    configToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    NULL,		/* dials */
  },
  {
    "midi",		/* name */
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    midiDials,		/* dials */
  },
  {
    "recording",	/* name */
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    recordToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    recordDials,	/* dials */
  },
  {
    "audio_in",		/* name */
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    audinLabels,	/* labels */
    NULL,		/* commands */
    audinToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    audinDials,		/* dials */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Echo pane *****/
static DialEntry echoDials[]={
  {
    "ech_lev",
    SAM_MIXER_ID_ECH_LEV,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "ech_tim",
    SAM_MIXER_ID_ECH_TIM,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "ech_feed",
    SAM_MIXER_ID_ECH_FEED,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Reverb pane *****/

static LabelEntry reverbLabels[]={
  {
    "type_lbl",
  },
  {
    NULL		/* sentinel */
  }
};

static DialEntry reverbDials[]={
  {
    "rev_vol",
    SAM_MIXER_ID_REV_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "rev_time",
    SAM_MIXER_ID_REV_TIME,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "rev_feed",
    SAM_MIXER_ID_REV_FEED,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "gmrev_send",
    SAM_MIXER_ID_GMREV_SEND,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry reverbToggles[]={
  {
    "rev_type_room1",
    SAM_MIXER_ID_REV_TYPE,
    -1, 0,
  },
  {
    "rev_type_room2",
    SAM_MIXER_ID_REV_TYPE,
    -1, 1,
  },
  {
    "rev_type_room3",
    SAM_MIXER_ID_REV_TYPE,
    -1, 2,
  },
  {
    "rev_type_hall1",
    SAM_MIXER_ID_REV_TYPE,
    -1, 3,
  },
  {
    "rev_type_hall2",
    SAM_MIXER_ID_REV_TYPE,
    -1, 4,
  },
  {
    "rev_type_plate",
    SAM_MIXER_ID_REV_TYPE,
    -1, 5,
  },
  {
    "rev_type_delay",
    SAM_MIXER_ID_REV_TYPE,
    -1, 6,
  },
  {
    "rev_type_pandelay",
    SAM_MIXER_ID_REV_TYPE,
    -1, 7,
  },
  {
    NULL		/* sentinel */
  }
};

/***** Chorus pane *****/

static LabelEntry chorusLabels[]={
  {
    "type_lbl",
  },
  {
    NULL		/* sentinel */
  }
};

static DialEntry chorusDials[]={
  {
    "chr_vol",
    SAM_MIXER_ID_CHR_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "chr_del",
    SAM_MIXER_ID_CHR_DEL,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "chr_feed",
    SAM_MIXER_ID_CHR_FEED,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "chr_rate",
    SAM_MIXER_ID_CHR_RATE,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "chr_depth",
    SAM_MIXER_ID_CHR_DEPTH,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "gmchr_send",
    SAM_MIXER_ID_GMCHR_SEND,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry chorusToggles[]={
  {
    "chr_type_chorus1",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 0,
  },
  {
    "chr_type_chorus2",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 1,
  },
  {
    "chr_type_chorus3",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 2,
  },
  {
    "chr_type_chorus4",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 3,
  },
  {
    "chr_type_fbchorus",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 4,
  },
  {
    "chr_type_flanger",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 5,
  },
  {
    "chr_type_shortdel",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 6,
  },
  {
    "chr_type_fbdel",
    SAM_MIXER_ID_CHR_TYPE,
    -1, 7,
  },
  {
    NULL		/* sentinel */
  }
};

static Pane effectsPanes[]={
  {
    "echo",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    echoDials,		/* dials */
  },
  {
    "reverb",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    reverbLabels,	/* labels */
    NULL,		/* commands */
    reverbToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    reverbDials,	/* dials */
  },
  {
    "chorus",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    chorusLabels,	/* labels */
    NULL,		/* commands */
    chorusToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    chorusDials,	/* dials */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Equalizer Low pane *****/
static DialEntry equLowDials[]={
  {
    "eq_lbl",
    SAM_MIXER_ID_EQ_LBL,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "eq_lbr",
    SAM_MIXER_ID_EQ_LBR,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    "eqf_lb",
    SAM_MIXER_ID_EQF_LB,
    0, 4.7, 0.1,	/* min, max, step */
    127./4.7, 0,	/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};


/***** Equalizer Medium Low pane *****/
static DialEntry equMedLowDials[]={
  {
    "eq_mlbl",
    SAM_MIXER_ID_EQ_MLBL,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "eq_mlbr",
    SAM_MIXER_ID_EQ_MLBR,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    "eqf_mlb",
    SAM_MIXER_ID_EQF_MLB,
    0, 4.2, 0.1,	/* min, max, step */
    127./4.2, 0,	/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Equalizer Medium High pane *****/
static DialEntry equMedHighDials[]={
  {
    "eq_mhbl",
    SAM_MIXER_ID_EQ_MHBL,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "eq_mhbr",
    SAM_MIXER_ID_EQ_MHBR,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    "eqf_mhb",
    SAM_MIXER_ID_EQF_MHB,
    0, 4.2, 0.1,	/* min, max, step */
    127./4.2, 0,	/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Equalizer High pane *****/
static DialEntry equHighDials[]={
  {
    "eq_hbl",
    SAM_MIXER_ID_EQ_HBL,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    +1,			/* relatedDial */
  },
  {
    "eq_hbr",
    SAM_MIXER_ID_EQ_HBR,
    -12, 12, 0.1,	/* min, max, step */
    128./24., 64,	/* multiply, add */
    -1,			/* relatedDial */
  },
  {
    "eqf_hb",
    SAM_MIXER_ID_EQF_HB,
    0, 18.75, 0.1,	/* min, max, step */
    127./18.75, 0,	/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

/***** Surround pane *****/
static DialEntry surroundDials[]={
  {
    "sur_vol",
    SAM_MIXER_ID_SUR_VOL,
    0, 255, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    "sur_del",
    SAM_MIXER_ID_SUR_DEL,
    0, 127, 1,		/* min, max, step */
    1, 0,		/* multiply, add */
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry surroundToggles[]={
  {
    "sur_inp",
    SAM_MIXER_ID_SUR_INP,
    0, 1,
  },
  {
    "sur_24",
    SAM_MIXER_ID_SUR_24,
    0, 1,
  },
  {
    NULL		/* sentinel */
  }
};

static Pane postPanes[]={
  {
    "equ_low",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    equLowDials,	/* dials */
  },
  {
    "equ_medlow",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    equMedLowDials,	/* dials */
  },
  {
    "equ_medhigh",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    equMedHighDials,	/* dials */
  },
  {
    "equ_high",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    equHighDials,	/* dials */
  },
  {
    "surround",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    surroundToggles,	/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    surroundDials,	/* dials */
  },
  {
    NULL		/* sentinel */
  }
};

static ToggleEntry daadToggles[]={
  {
    "sur_inp",
    SAM_MIXER_ID_SUR_INP,
    0, 1,
  },
  {
    "sur_24",
    SAM_MIXER_ID_SUR_24,
    0, 1,
  },
  {
    NULL		/* sentinel */
  }
};

static Pane daadPanes[]={
  {
    "daad",
    1,			/* needsOpenDevice */
    NULL,		/* subPanes */
    NULL,		/* labels */
    NULL,		/* commands */
    daadToggles,	/* toggles */
    NULL,		/* toggles */
    NULL,		/* lists */
    NULL,		/* texts */
    NULL,	/* dials */
  },
  {
	NULL
  }
};

static Pane topPanes[]={
  {
    "config",
    0,			/* needsOpenDevice */
    configPanes,
  },
  {
    "general",
    1,			/* needsOpenDevice */
    generalPanes,
  },
  {
    "effects",
    1,			/* needsOpenDevice */
    effectsPanes,
  },
  {
    "post",
    1,			/* needsOpenDevice */
    postPanes,
  },
  {
    "Daad",
    1,			/* needsOpenDevice */
    daadPanes,
  },
  {
    NULL		/* sentinel */
  }
};

static void refreshPane(Pane *pane)
{
  Boolean cardWorking;
  LabelEntry *labelEntry;
  CommandEntry *commandEntry;
  ToggleEntry *toggleEntry;
  ListEntry *listEntry;
  TextEntry *textEntry;
  DialEntry *dialEntry;

  if(pane->needsOpenDevice) {
    cardWorking=applState.deviceUsed>=0 && *applState.knownDevices[applState.deviceUsed].driverInfo.firmware;
    XtSetSensitive(pane->tab, cardWorking);
    if(!cardWorking)
      return;
  }

  if(pane->subPanes)
    return;

  if(pane->labels)
    for(labelEntry=pane->labels; labelEntry->name; labelEntry++) {
      if(labelEntry->retrieve)
	XtVaSetValues(labelEntry->w,
		      XtNlabel, (*labelEntry->retrieve)(),
		      NULL);
    }
  if(pane->commands)
    for(commandEntry=pane->commands; commandEntry->name; commandEntry++) {
    }
  if(pane->toggles)
    for(toggleEntry=pane->toggles; toggleEntry->name; toggleEntry++) {
      Boolean supported=applState.mixer[toggleEntry->mixerId].supported;
      XtVaSetValues(toggleEntry->w,
		    XtNsensitive, supported,
		    XtNstate, supported ? applState.mixer[toggleEntry->mixerId].value==toggleEntry->onValue : 0,
		    NULL);
    }
  if(pane->lists)
    for(listEntry=pane->lists; listEntry->name; listEntry++) {
      String *list;
      int numberStrings=0;
      if(listEntry->retrieve) {
	XtVaSetValues(listEntry->w,
		      XtNlist, NULL,
		      XtNnumberStrings, 0,
		      NULL);
	(*listEntry->retrieve)(&list, &numberStrings);
	XtVaSetValues(listEntry->w,
		      XtNlist, list,
		      XtNnumberStrings, numberStrings,
		      NULL);
      }
      listEntry->selected=-1;
    }
  if(pane->texts)
    for(textEntry=pane->texts; textEntry->name; textEntry++) {
      XtVaSetValues(textEntry->w,
		    XtNstring, "",
		    NULL);
    }
  if(pane->dials)
    for(dialEntry=pane->dials; dialEntry->name; dialEntry++) {
      Boolean supported=applState.mixer[dialEntry->mixerId].supported;
      float value;

      if(supported)
	value=((float)applState.mixer[dialEntry->mixerId].value-dialEntry->add)/dialEntry->multiply;
      else
	value=0;
      XtVaSetValues(dialEntry->w,
		    XtNsensitive, supported,
		    XtNvalue, *(unsigned *)&value,
		    NULL);
    }
}

static void buildPanes(Widget tab, Pane *aPane)
{
  Pane *pane;
  int i;
  char buf[256];
  LabelEntry *labelEntry;
  CommandEntry *commandEntry;
  ToggleEntry *toggleEntry;
  ListEntry *listEntry;
  TextEntry *textEntry;
  DialEntry *dialEntry;

  for(i=0, pane=aPane; pane->name; i++, pane++) {
    sprintf(buf, "pane_%s", pane->name);
    if(pane->subPanes) {
      pane->w=XtVaCreateManagedWidget(buf, tabWidgetClass, tab,
				      XtNisTab, False,
				      NULL);
      buildPanes(pane->w, pane->subPanes);
    } else {
      pane->w=XtVaCreateManagedWidget(buf, layoutWidgetClass, tab,
				      XtNisTab, False,
				      NULL);
      if(pane->labels)
	for(labelEntry=pane->labels; labelEntry->name; labelEntry++) {
	  labelEntry->w=XtVaCreateManagedWidget(labelEntry->name, labelWidgetClass, pane->w,
						NULL);
	  labelEntry->pane=pane;
	}
      if(pane->commands)
	for(commandEntry=pane->commands; commandEntry->name; commandEntry++) {
	  commandEntry->w=XtVaCreateManagedWidget(commandEntry->name, commandWidgetClass, pane->w,
						  NULL);
	  commandEntry->pane=pane;
	  if(commandEntry->callback)
	    XtAddCallback(commandEntry->w, XtNcallback, commandEntry->callback, commandEntry);
	}
      if(pane->toggles)
	for(toggleEntry=pane->toggles; toggleEntry->name; toggleEntry++) {
	  toggleEntry->w=XtVaCreateManagedWidget(toggleEntry->name, toggleWidgetClass, pane->w,
						 NULL);
	  XtAddCallback(toggleEntry->w, XtNcallback, toggleValueCallback, toggleEntry);

	  toggleEntry->pane=pane;

	  toggleEntry->notify.callback=toggleValueNotify;
	  toggleEntry->notify.next=applState.mixer[toggleEntry->mixerId].callbacks;
	  toggleEntry->notify.closure=toggleEntry;
	  applState.mixer[toggleEntry->mixerId].callbacks=&toggleEntry->notify;
	}
      if(pane->lists)
	for(listEntry=pane->lists; listEntry->name; listEntry++) {
	  listEntry->w=XtVaCreateManagedWidget(listEntry->name, listWidgetClass, pane->w,
					       XtNforceColumns, True,
					       XtNdefaultColumns, 1,
					       NULL);
	  XtAddCallback(listEntry->w, XtNcallback, selectElementCallback, listEntry);

	  listEntry->pane=pane;
	  listEntry->selected=-1;
	}
      if(pane->texts)
	for(textEntry=pane->texts; textEntry->name; textEntry++) {
	  textEntry->w=XtVaCreateManagedWidget(textEntry->name, asciiTextWidgetClass, pane->w,
					       XtNeditType, XawtextEdit,
					       NULL);
	  textEntry->pane=pane;
	  XtOverrideTranslations(textEntry->w,
				 XtParseTranslationTable(textTranslations));
	  XSaveContext(XtDisplay(textEntry->w), (XID)textEntry->w, applState.ackTextContext,
		       (XPointer)ackTextCallback);
	  XSaveContext(XtDisplay(textEntry->w), (XID)textEntry->w, applState.ackTextClosureContext,
		       (XPointer)textEntry);
	}
      if(pane->dials)
	for(dialEntry=pane->dials; dialEntry->name; dialEntry++) {
	  dialEntry->w=XtVaCreateManagedWidget(dialEntry->name, dialWidgetClass, pane->w,
					       XtNduplexMode, True,
					       XtNminimum, *(unsigned *)&dialEntry->min,
					       XtNmaximum, *(unsigned *)&dialEntry->max,
					       XtNstep, *(unsigned *)&dialEntry->step,
					       NULL);
	  XtAddCallback(dialEntry->w, XtNcallback, dialValueCallback, dialEntry);

	  dialEntry->pane=pane;

	  dialEntry->notify.callback=dialValueNotify;
	  dialEntry->notify.next=applState.mixer[dialEntry->mixerId].callbacks;
	  dialEntry->notify.closure=dialEntry;
	  applState.mixer[dialEntry->mixerId].callbacks=&dialEntry->notify;
	}
    }
    sprintf(buf, "tab_%s", pane->name);
    pane->tab=XtVaCreateManagedWidget(buf, labelWidgetClass, tab,
				      XtNisTab, True,
				      XtNrelatedPane, pane->w,
				      XtNtabPosition, i,
				      NULL);
  }
}

static void refreshAllPanes(struct Pane *aPane)
{
  Pane *pane;

  for(pane=aPane; pane->name; pane++) {
    refreshPane(pane);
    if(pane->subPanes)
      refreshAllPanes(pane->subPanes);
  }
}

void addPanes(Widget parent)
{
  buildPanes(parent, topPanes);
  refreshAllPanes(topPanes);
}

void notifySBankChanged(void)
{
  refreshAllPanes(configPanes);
}

static void selectElementCallback(Widget w, XtPointer closure, XtPointer callData)
{
  ListEntry *entry=(ListEntry *)closure;
  XawListReturnStruct *info=(XawListReturnStruct *)callData;

  entry->selected=info->list_index;

  if(entry->callback)
    (*entry->callback)(w, closure, callData);
}


static void dialValueCallback(Widget w, XtPointer closure, XtPointer callData)
{
  DialEntry *entry=(DialEntry *)closure;
  DialCallbackInfo *info=(DialCallbackInfo *)callData;
  unsigned value;

  value=info->value*entry->multiply+entry->add;
  changeMixerValue(entry->mixerId, value);
  if(entry->relatedDial && info->button==resources.adjustRelatedButton)
    changeMixerValue(entry[entry->relatedDial].mixerId, value);
}

static void dialValueNotify(int id, MixerCallback *info)
{
  DialEntry *entry=(DialEntry *)info->closure;
  float value=((double)applState.mixer[id].value-entry->add)/entry->multiply;

  XtVaSetValues(entry->w,
		XtNvalue, *(unsigned *)&value,
		NULL);
}

static void toggleValueCallback(Widget w, XtPointer closure, XtPointer callData)
{
  ToggleEntry *entry=(ToggleEntry *)closure;
  int newValue;

  /* value changes are reported from device
   * so switch back toggle state
   */
  XtVaSetValues(w, XtNstate, !callData, NULL);

  if(callData)
    newValue=entry->onValue;
  else
    newValue=entry->offValue;

  if(newValue>=0)
    changeMixerValue(entry->mixerId, newValue);
}

static void toggleValueNotify(int id, MixerCallback *info)
{
  ToggleEntry *entry=(ToggleEntry *)info->closure;

  XtVaSetValues(entry->w,
		XtNstate, applState.mixer[id].value==entry->onValue,
		NULL);
}

static void ackTextCallback(Widget w, XtPointer closure)
{
  TextEntry *entry=(TextEntry *)closure;
  if(entry->acknowledge)
    (*entry->acknowledge)(w, entry);
}

static String retrieveCardNr(void)
{
  static char result[32];

  if(applState.deviceUsed>=0)
    sprintf(result, "%d", applState.deviceUsed);
  else
    strcpy(result, "None");

  return result;
}

static String getMemUsedSBank(void)
{
  static char result[32];
  KnownDevice *knownDevice;

  if(applState.deviceUsed>=0) {
    knownDevice=applState.knownDevices+applState.deviceUsed;
    sprintf(result, "%dk", knownDevice->driverInfo.memSBanks>>10);
  } else
    strcpy(result, "N/A");

  return result;
}

static String getMemFree(void)
{
  static char result[32];
  KnownDevice *knownDevice;

  if(applState.deviceUsed>=0) {
    knownDevice=applState.knownDevices+applState.deviceUsed;
    sprintf(result, "%dk", knownDevice->driverInfo.memFree>>10);
  } else
    strcpy(result, "N/A");

  return result;
}

static void retrieveDevices(String **list, int *numberStrings)
{
  static String result[MAX_KNOWN_DEVICES];
  char buf[256];
  int cardNr;
  KnownDevice *knownDevice;

  for(cardNr=0; cardNr<MAX_KNOWN_DEVICES; cardNr++) {
    if(!result[cardNr])
      break;
    XtFree(result[cardNr]);
  }

  for(cardNr=0; cardNr<MAX_KNOWN_DEVICES; cardNr++) {
    knownDevice=applState.knownDevices+cardNr;
    if(!knownDevice->device)
      break;
    sprintf(buf, "%d %-15s %s", cardNr, knownDevice->driverInfo.hardware,
	    *knownDevice->driverInfo.firmware ? knownDevice->driverInfo.firmware : "N/A");
    result[cardNr]=XtNewString(buf);
  }

  *list=result;
  *numberStrings=cardNr;
}

static void retrieveSoundbanks(String **list, int *numberStrings)
{
  static String result[SAM_NROF_SBANKS];
  static char buf[SAM_NROF_SBANKS][32];
  int i;

  for(i=0; i<SAM_NROF_SBANKS; i++) {
    result[i]=buf[i];
    if(*applState.sbanks[i].name)
      sprintf(buf[i], "%d %-8s %d", i, applState.sbanks[i].name, applState.sbanks[i].priority);
    else
      sprintf(buf[i], "%d", i);
  }

  *list=result;
  *numberStrings=SAM_NROF_SBANKS;
}

static void rescanDevicesCallback(Widget w, XtPointer closure, XtPointer callData)
{
  scanDevices(-1);
  refreshAllPanes(topPanes);
}

static void useDeviceCallback(Widget w, XtPointer closure, XtPointer callData)
{
  CommandEntry *entry=(CommandEntry *)closure;

  openDevice(entry->pane->lists->selected);
  refreshAllPanes(topPanes);
}

static void releaseDeviceCallback(Widget w, XtPointer closure, XtPointer callData)
{
  closeDevice();
  refreshAllPanes(topPanes);
}

static void loadFirmwareCallback(Widget w, XtPointer closure, XtPointer callData)
{
  char *fileName;

  if(applState.deviceUsed<0) {
    error(0, "%s", resources.errNoCardSel);
    return;
  }

  if((fileName=XsraSelFile(applState.top, resources.fsFirmTitle, NULL, &applState.fsFirmwarePath, 1, NULL))) {
    loadFirmware(fileName);
    refreshAllPanes(topPanes);
    XtFree(fileName);
  }
}

static void sbankSelectedCallback(Widget w, XtPointer closure, XtPointer callData)
{
  ListEntry *entry=(ListEntry *)closure;
  Widget text;
  char buf[32];

  if(!(text=XtNameToWidget(XtParent(w), "*prio")))
    return;

  if(*applState.sbanks[entry->selected].name)
    sprintf(buf, "%d", applState.sbanks[entry->selected].priority);
  else
    *buf=0;

  XtVaSetValues(text, XtNstring, buf, NULL);
}

static void loadSoundbankCallback(Widget w, XtPointer closure, XtPointer callData)
{
  CommandEntry *entry=(CommandEntry *)closure;
  char *fileName;

  if(entry->pane->lists->selected<0)
    return;

  if((fileName=XsraSelFile(applState.top, resources.fsSBankTitle, NULL, &applState.fsSoundbankPath, 1, NULL))) {
    loadSoundbank(entry->pane->lists->selected, fileName);
    XtFree(fileName);
  }
}

static void unloadSoundbankCallback(Widget w, XtPointer closure, XtPointer callData)
{
  CommandEntry *entry=(CommandEntry *)closure;

  if(entry->pane->lists->selected>=0)
    unloadSoundbank(entry->pane->lists->selected);
}

static void changeSBankPriorityCallback(Widget w, TextEntry *entry)
{
  String value;
  XtVaGetValues(w, XtNstring, &value, NULL);
  if(entry->pane->lists->selected>=0)
    changeSBankPriority(entry->pane->lists->selected, atoi(value));
}
